/* File:      builtin.P
** Author(s): Debray, Sagonas, Swift, Warren, Xu
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1999
** Copyright (C) ECRC, Germany, 1990
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: builtin.P,v 1.12 2005/03/04 16:03:18 kostis Exp $
** 
*/


:- compiler_options([xpp_on]).
#include "../emu/builtin_tab.h"
#include "../emu/function_tab.h"

/*======================================================================*/
/* inline predicates							*/
/*======================================================================*/

inline_predicate('=', 2).
inline_predicate('<', 2).
inline_predicate('>', 2).
inline_predicate('=<', 2).
inline_predicate('>=', 2).
inline_predicate('=:=', 2).
inline_predicate('=\=', 2).
inline_predicate(is, 2).
inline_predicate('==', 2).
inline_predicate('\==', 2).
inline_predicate('@<', 2).
inline_predicate('@=<', 2).
inline_predicate('@>', 2).
inline_predicate('@>=', 2).

inline_predicate(var, 1).
inline_predicate(nonvar, 1).
inline_predicate('_$cutto', 1).
inline_predicate('_$savecp', 1).
inline_predicate('_$builtin', 1).

inline_predicate(fail, 0).
inline_predicate(true, 0).
inline_predicate(halt, 0).
inline_predicate('!', 0).

/*======================================================================*/
/* primitive predicates	("hard-wired" builtins)				*/
/*======================================================================*/

inline_builtin(psc_name, 2, PSC_NAME).
inline_builtin(psc_arity, 2, PSC_ARITY).
inline_builtin(psc_type, 2, PSC_TYPE).
inline_builtin(psc_prop, 2, PSC_PROP).
inline_builtin(psc_set_type, 2, PSC_SET_TYPE).
inline_builtin(psc_set_prop, 2, PSC_SET_PROP).
inline_builtin(psc_set_spy, 2, PSC_SET_SPY).
inline_builtin(psc_ep, 2, PSC_EP).

inline_builtin(term_psc, 2, TERM_PSC).
inline_builtin(term_type, 2, TERM_TYPE).
inline_builtin(term_compare, 3, TERM_COMPARE).
inline_builtin(term_new, 2, TERM_NEW).
inline_builtin(term_arg, 3, TERM_ARG).
inline_builtin(term_set_arg, 4, TERM_SET_ARG).
inline_builtin(stat_flag, 2,  STAT_FLAG).
inline_builtin(stat_set_flag, 2, STAT_SET_FLAG).
inline_builtin(buff_alloc, 2, BUFF_ALLOC).
inline_builtin(buff_word, 3, BUFF_WORD).
inline_builtin(buff_set_word, 3, BUFF_SET_WORD).

inline_builtin(code_call, 3, CODE_CALL).

inline_builtin(substring, 4, SUBSTRING).
inline_builtin(str_len, 2, STR_LEN).
inline_builtin(str_cat, 3, STR_CAT).
inline_builtin(str_cmp, 3, STR_CMP).
inline_builtin(string_substitute, 4, STRING_SUBSTITUTE).

inline_builtin(call0, 1, CALL0).
inline_builtin(stat_sta, 1, STAT_STA).
inline_builtin(stat_cputime, 1, STAT_CPUTIME).
inline_builtin(code_load, 3, CODE_LOAD).
inline_builtin(buff_set_var, 3, BUFF_SET_VAR).
inline_builtin(buff_dealloc, 2, BUFF_DEALLOC).
inline_builtin(buff_cell, 2, BUFF_CELL).
inline_builtin(buff_set_cell, 2, BUFF_SET_CELL).
inline_builtin(copy_term, 2, COPY_TERM).
inline_builtin(str_sub, 3, STR_SUB).
inline_builtin(dirname_canonic, 2, DIRNAME_CANONIC).
inline_builtin(psc_insert, 4, PSC_INSERT).
inline_builtin(psc_import, 3, PSC_IMPORT).

inline_builtin(psc_insertmod, 3, PSC_INSERTMOD).

inline_builtin(file_gettoken, 5, FILE_GETTOKEN).
inline_builtin(file_puttoken, 3, FILE_PUTTOKEN).
inline_builtin(term_hash, 3, TERM_HASH).
inline_builtin(unload_seg, 1, UNLOAD_SEG).
inline_builtin(load_obj, 4, LOAD_OBJ).

inline_builtin(getenv, 2, GETENV).
/* inline_builtin(sys_syscall, 4, SYS_SYSCALL). */
inline_builtin(sys_system, 7, SYS_SYSTEM).
inline_builtin(sys_gethost, 2, SYS_GETHOST).
inline_builtin(sys_errno, 1, SYS_ERRNO).

inline_builtin(file_stat, 3, FILE_STAT).
inline_builtin(file_writequoted, 2, FILE_WRITEQUOTED).
inline_builtin(ground, 1, GROUND).
inline_builtin(numbervars, 3, NUMBERVARS).

inline_builtin(intern_string, 2, INTERN_STRING).
inline_builtin(expand_filename, 2, EXPAND_FILENAME).
inline_builtin(tilde_expand_filename, 2, TILDE_EXPAND_FILENAME).
inline_builtin(is_absolute_filename, 1, IS_ABSOLUTE_FILENAME).
inline_builtin(parse_filename, 4, PARSE_FILENAME).
inline_builtin(almost_search_module, 5, ALMOST_SEARCH_MODULE).
inline_builtin(existing_file_extension, 2, EXISTING_FILE_EXTENSION).

inline_builtin(psc_env, 2, PSC_ENV).
inline_builtin(psc_spy, 2, PSC_SPY).
inline_builtin(psc_tabled, 2, PSC_TABLED).

% inline_builtin(is_incomplete, 4, IS_INCOMPLETE). % Should not be inlined -- Kostis !!

inline_builtin(get_ptcp, 1, GET_PTCP).
inline_builtin(get_subgoal_ptr, 2, GET_SUBGOAL_PTR).
inline_builtin(dereference_the_bucket, 2, DEREFERENCE_THE_BUCKET).
inline_builtin(pair_psc, 2, PAIR_PSC).
inline_builtin(pair_next, 2, PAIR_NEXT).
inline_builtin(next_bucket, 2, NEXT_BUCKET).

inline_builtin(slg_not, 1, SLG_NOT).

inline_builtin(is_xwammode, 0, IS_XWAMMODE).
inline_builtin(cleanup_open_tables, 0, CLEANUP_OPEN_TABLES).

inline_builtin(file_function, 7, FILE_FUNCTION).
inline_builtin(slash, 1, SLASH_BUILTIN).

inline_builtin(abolish_table_info, 0, ABOLISH_TABLE_INFO).

inline_builtin(formatted_io, 5, FORMATTED_IO).
inline_builtin(table_status, 3, TABLE_STATUS).
inline_builtin(get_delay_lists, 2, GET_DELAY_LISTS).

inline_builtin(abolish_table_predicate, 1, ABOLISH_TABLE_PREDICATE).

inline_builtin(trie_delete_branch_ret, 3, TRIE_DELETE_TERM).
% inline_builtin(get_returns, 2, TRIE_GET_RETURN). % Cannot be inlined - compiler bug ?
% inline_builtin(get_calls, 1, TRIE_UNIFY_CALL). % Cannot be inlined - compiler bug ?
inline_builtin(get_call, 3, TRIE_GET_CALL).
inline_builtin(get_lastnode_cs_retskel, 4, GET_LASTNODE_CS_RETSKEL).
% inline_builtin(incomplete_retskel_suboal, 2, INCOMPLETE_RETSKEL_SUBGOAL).
% inline_builtin(get_constraints, 1, GET_CONSTRAINTS).

inline_builtin(nl, 0, NL).
% The following two are not declared as inlined, because they are inlined
% in a special way (through WAM instructions) by the XSB compiler anyway.
% inline_builtin(var, 1, VAR).
% inline_builtin(nonvar, 1, NONVAR).
inline_builtin(atom, 1, ATOM).
inline_builtin(integer, 1, INTEGER).
inline_builtin(real, 1, REAL).
inline_builtin(number, 1, NUMBER).
inline_builtin(atomic, 1, ATOMIC).
inline_builtin(compound, 1, COMPOUND).
inline_builtin(callable, 1, CALLABLE).
inline_builtin(is_list, 1, IS_LIST).
inline_builtin(functor, 3, FUNCTOR).
inline_builtin(arg, 3, ARG).
inline_builtin(=.., 2, UNIV).
inline_builtin(is_most_general_term, 1, IS_MOST_GENERAL_TERM).
inline_builtin(hilog_arg, 3, HiLog_ARG).
% inline_builtin(hilog_univ, 2, HiLog_UNIV).
inline_builtin(closetail, 1, CLOSETAIL).
inline_builtin(atom_codes, 2, ATOM_CODES).
inline_builtin(atom_chars, 2, ATOM_CHARS).
inline_builtin(number_chars, 2, NUMBER_CHARS).
inline_builtin(put, 1, PUT).
inline_builtin(tab, 1, TAB).
inline_builtin(number_codes, 2, NUMBER_CODES).
inline_builtin(is_charlist, 2, IS_CHARLIST).
inline_builtin(number_digits, 2, NUMBER_DIGITS).

inline_builtin(sort, 2, SORT).
inline_builtin(keysort, 2, KEYSORT).

inline_builtin('$$set_scope_marker', 0, SET_SCOPE_MARKER).
inline_builtin('$$unwind_stack', 0, UNWIND_STACK).
inline_builtin('$$clean_up_block', 0, CLEAN_UP_BLOCK).

inline_builtin(print_compl, 0, PRINT_COMPL).
inline_builtin(print_chat, 0, PRINT_CHAT).
inline_builtin(print_ls, 0, PRINT_LS).
inline_builtin(print_tr, 0, PRINT_TR).
inline_builtin(print_heap, 0, PRINT_HEAP).
inline_builtin(print_cp, 0, PRINT_CP).
inline_builtin(print_regs, 0, PRINT_REGS).
inline_builtin(print_all_stacks, 0, PRINT_ALL_STACKS).
inline_builtin(expand_heap, 0, EXP_HEAP).
inline_builtin(mark_heap, 1, MARK_HEAP).
inline_builtin(gc_heap, 0, GC_HEAP).
inline_builtin('$$findall_init', 2, FINDALL_INIT).
inline_builtin('$$findall_add', 3, FINDALL_ADD).
inline_builtin('$$findall_get_solutions', 4, FINDALL_GET_SOLS).

inline_builtin(force_truth_value, 2, FORCE_TRUTH_VALUE).

/*======================================================================*/
/* builtin functions							*/
/*======================================================================*/

builtin_function('+', 2, FUN_PLUS).
builtin_function('-', 2, FUN_MINUS).
builtin_function('*', 2, FUN_TIMES).
builtin_function('/', 2, FUN_DIVIDE).
builtin_function('/\', 2, FUN_AND).
builtin_function('\/', 2, FUN_OR).
builtin_function('//', 2, FUN_INTDIV).
builtin_function(sin, 1, FUN_sin).
builtin_function(cos, 1, FUN_cos).
builtin_function(tan, 1, FUN_tan).
builtin_function(float, 1, FUN_float).
builtin_function(floor, 1, FUN_floor).
builtin_function(exp, 1, FUN_exp).
builtin_function(log, 1, FUN_log).
builtin_function(log10, 1, FUN_log10).
builtin_function(sqrt, 1, FUN_sqrt).
builtin_function(asin, 1, FUN_asin).
builtin_function(acos, 1, FUN_acos).
builtin_function(atan, 1, FUN_atan).
builtin_function(abs, 1, FUN_abs).
builtin_function(truncate, 1, FUN_truncate).
builtin_function(round, 1, FUN_round).
builtin_function(ceiling, 1, FUN_ceiling).
% builtin_function(min, 2, FUN_min).
% builtin_function(max, 2, FUN_max).
