/* File:      listutil.P
** Author(s): Saumya Debray, Kostis F. Sagonas
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** Copyright (C) ECRC, Germany, 1990
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: listutil.P,v 1.2 2004/01/26 22:17:13 kostis Exp $
** 
*/



/* --- some Prolog list utilities used in cmplib ---------------------- */

merge([], L1, L1).
merge([H|T], L1, L3) :- (member(H, L1), merge(T, L1, L3)) ;
			(L3 = [H | L4], merge(T, L1, L4)).


%   absmember(+Element, +List)
%   Predicate absmember/2 is similar to the good old member/2, except 
%   that it checks for identity rather than unifiability.

absmember(X, [H|T]) :- '=='(X,H) ; absmember(X, T).


%   absmerge(+List1, +List2, ?Merged_List)
%   Predicate absmerge/3 is similar to merge/3, except that it uses 
%   absmember/2 rather than member/2.

absmerge([], L1, L1).
absmerge([H|T], L1, L3) :- (absmember(H, L1), absmerge(T, L1, L3)) ;
			   (L3 = [H | L4], absmerge(T, L1, L4)).


%   member2(?Element, +List)
%   Predicate member2/2 is used just to check for membership in a list 
%   whose tail is a variable. It fails if the element is not found or
%   if the list is an improper list (including variable).

member2(X,L) :- nonvar(L), L = [Y|Z], (X = Y ; member2(X,Z)).

%   closetail(?List)
%   Predicate closetail/1 closes the tail of an open-ended list.

closetail(L) :- closetail(L).


membercheck(Elem, Table) :- nonvar(Table), Table = [Head|Rest],
			    (Elem = Head -> true ; membercheck(Elem, Rest)).

%   delete_ith(+I, +List, -Ith, -NewList).
%   Deletes the Ith element of List and returns the rest of the list
%   in NewList.

delete_ith(N, [H|T], Ith, Rest) :-
	( N =:= 1 -> Ith = H, Rest = T
	; Rest = [H|Rest1], N1 is N-1, delete_ith(N1, T, Ith, Rest1)
	).

force_listutil_loaded.
