/* File:      peephole.P
** Author(s): Saumya Debray, Kostis Sagonas
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** Copyright (C) ECRC, Germany, 1990
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: peephole.P,v 1.5 2003/11/23 23:15:49 kostis Exp $
** 
*/



/************************************************************************
*									*
*		some code for peephole optimization			*
*								 	*
*************************************************************************/

/*  "peephole_opt" is the top-level optimizer, it calls various others.	*/

compile_peephole_opt(Pil, OptPil) :-
	popt1(Pil, Pil1),
	popt2(Pil1, Pil2),
	popt3(Pil2, Pil3),
	popt4(Pil3,_,Pil4),
	popt_alloc(Pil4, OptPil),
	!.


popt1([], []).
popt1([Inst|Rest], Pil1) :- popt11(Inst, Rest, Pil1).

%:- ti popt11/3.

popt11(deallocate, [proceed|PilRest], [dealloc_proceed|OptPilRest]) :-
	popt1a(PilRest, OptPilRest).
popt11(deallocate, [execute(S)|PilRest], [dealloc_execute(S)|OptPilRest]) :-
	popt1a(PilRest, OptPilRest).
popt11(puttvar(T,R),[getstr(R,S)|PilRest], [putstr(T,S)|OptPilRest]) :-
	popt1a(PilRest, OptPilRest).
popt11(movreg(T,R), [puttbreg(R)|PilRest], [puttbreg(T)|OptPilRest]) :-
	popt1(PilRest, OptPilRest).
popt11(movreg(T,R), [addreg(R,S)|PilRest], [addreg(T,S)|OptPilRest]) :-
	popt1(PilRest, OptPilRest).
popt11(movreg(T,R), [subreg(R,S)|PilRest], [subreg(T,S)|OptPilRest]) :-
	popt1(PilRest, OptPilRest).
popt11(movreg(T,R), [mulreg(R,S)|PilRest], [mulreg(T,S)|OptPilRest]) :-
	popt1(PilRest, OptPilRest).
popt11(movreg(T,R), [divreg(R,S)|PilRest], [divreg(T,S)|OptPilRest]) :-
	popt1(PilRest, OptPilRest).
popt11(movreg(T,R), [idivreg(R,S)|PilRest], [idivreg(T,S)|OptPilRest]) :-
	popt1(PilRest, OptPilRest).
popt11(movreg(T,R), [int_test_eq(R,N,L)|PilRest], [int_test_eq(T,N,L)|OptPilRest]) :-
	popt1(PilRest, OptPilRest).
popt11(movreg(T,R), [int_test_neq(R,N,L)|PilRest], [int_test_neq(T,N,L)|OptPilRest]) :-
	popt1(PilRest, OptPilRest).
popt11(movreg(T,R), [atom_test_eq(R,N,L)|PilRest], [atom_test_eq(T,N,L)|OptPilRest]) :-
	popt1(PilRest, OptPilRest).
popt11(movreg(T,R), [atom_test_neq(R,N,L)|PilRest], [atom_test_neq(T,N,L)|OptPilRest]) :-
	popt1(PilRest, OptPilRest).
popt11(putnumcon(R,N), [addreg(R,R2)|PilRest], [add_byte(N,R2)|OptPilRest]) :-
	integer(N), N > 0, N < 256,
	popt1(PilRest, OptPilRest).
popt11(putnumcon(R,N), [subreg(R,R2)|PilRest], [sub_byte(N,R2)|OptPilRest]) :-
	integer(N), N > 0, N < 256,
	popt1(PilRest, OptPilRest).
popt11(getpvar(V,R), [putpval(V,R)|PilRest], [getpvar(V,R)|OptPilRest]) :-
	popt1(PilRest, OptPilRest).
popt11(putpvar(V,R), [getpval(V,R)|PilRest], [putpvar(V,R)|OptPilRest]) :-
	popt1(PilRest, OptPilRest).
popt11(putpval(V,R), [getstr(R,Str)|PilRest], [getstrv(V,Str)|OptPilRest]) :-
	\+ sym_name(Str,'.',2),		/* to enable list opt */
	popt1(PilRest, OptPilRest).
popt11(putpvar(V,R), [getstr(R,Str)|PilRest], [putstrv(V,Str)|OptPilRest]) :-
	\+ sym_name(Str,'.',2),		/* to enable list opt */
	popt1a(PilRest, OptPilRest).
popt11(puttvar(R,R), PRest, [putvoid(R)|OptPRest]) :- % added
	popt1(PRest, OptPRest).
popt11(gettval(R,R), PRest, OptPRest) :-
	popt1(PRest, OptPRest).
popt11(movreg(R,R), PRest, OptPRest) :- 
	popt1(PRest, OptPRest).
popt11(jump(L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1(PRest,OptPRest).
% Dead-code elimination
popt11(jump(Addr), [Instr|PRest], OptCode) :-
	( Instr = label(_,_) ->
	    OptCode = [jump(Addr),Instr|OptPRest], % label stays
	    popt1(PRest, OptPRest)
	;   
	    popt11(jump(Addr), PRest, OptCode)
	).
% More dead-code elimination
popt11(fail, [Instr|PRest], OptCode) :-
	( Instr = label(_,_) ->
	    OptCode = [fail,Instr|OptPRest], % label stays
	    popt1(PRest, OptPRest)
	;
	    popt11(fail, PRest, OptCode)
	).
popt11(jumpz(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1(PRest, OptPRest).
popt11(jumpnz(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1(PRest, OptPRest).
popt11(jumplt(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1(PRest, OptPRest).
popt11(jumple(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1(PRest, OptPRest).
popt11(jumpgt(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1(PRest, OptPRest).
popt11(jumpge(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1(PRest, OptPRest).
popt11(int_test_eq(_,_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1(PRest, OptPRest).
popt11(int_test_neq(_,_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1(PRest, OptPRest).
popt11(Inst, PilRest, [Inst|OptPilRest]) :- 
	popt1(PilRest, OptPilRest).


popt1a([], []).
popt1a([Inst|PilRest], Pil1) :- 
	popt1a1(Inst, PilRest, Pil1).

popt1a1(univoid, PilRest, [bldvoid|OptPilRest]) :-
	popt1a(PilRest, OptPilRest).
popt1a1(unipvar(X), PilRest, [bldpvar(X)|OptPilRest]) :-
	popt1a(PilRest, OptPilRest).
popt1a1(unipval(X), PilRest, [bldpval(X)|OptPilRest]) :-
	popt1a(PilRest, OptPilRest).
popt1a1(unitvar(X), PilRest, [bldtvar(X)|OptPilRest]) :-
	popt1a(PilRest, OptPilRest).
popt1a1(unitval(X), PilRest, [bldtval(X)|OptPilRest]) :-
	popt1a(PilRest, OptPilRest).
popt1a1(unicon(X), PilRest, [bldcon(X)|OptPilRest]) :-
	popt1a(PilRest, OptPilRest).
popt1a1(uninumcon(X), PilRest, [bldnumcon(X)|OptPilRest]) :-
	popt1a(PilRest, OptPilRest).
popt1a1(unifloat(X), PilRest, [bldfloat(X)|OptPilRest]) :-
	popt1a(PilRest, OptPilRest).
popt1a1(gettval(R,R), PilRest, OptPilRest) :- 
	popt1a(PilRest, OptPilRest).
popt1a1(movreg(R,R), PilRest, OptPilRest) :- 
	popt1a(PilRest, OptPilRest).
popt1a1(jump(Addr), [jump(_)|PRest], [jump(Addr)|OptPRest]) :-
	popt1a(PRest,OptPRest).
popt1a1(jumpz(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1a(PRest,OptPRest).
popt1a1(jumpnz(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1a(PRest,OptPRest).
popt1a1(jumplt(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1a(PRest, OptPRest).
popt1a1(jumple(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1a(PRest, OptPRest).
popt1a1(jumpgt(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1a(PRest,OptPRest).
popt1a1(jumpge(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1a(PRest,OptPRest).
popt1a1(int_test_eq(_,_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1a(PRest,OptPRest).
popt1a1(int_test_neq(_,_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	popt1a(PRest,OptPRest).
popt1a1(Inst, PilRest, OptPilRest) :- 
	popt1([Inst|PilRest], OptPilRest).

/*----------------------------------------------------------------------*/
/*  "popt2" optimizes list instructions and pass throughs
    generated from transformational indexing: (put|get)str '.'/2 are 
    replaced by (put|get)list, and (get|put|uni|bld)con '[]'/0 are
    replaced by (get|put|uni|bld)nil, respectively.  Pass throughs
    are simply forgotten.						*/
/*----------------------------------------------------------------------*/

popt2([], []).
popt2([Inst|PilRest], Pil1) :- popt21(Inst, PilRest, Pil1).

popt21(getstr(R,Str), PilRest, [getlist(R)|OptPilRest]) :-
	sym_name(Str,'.',2),
	popt2(PilRest, OptPilRest).
popt21(putstr(R, Str), PilRest, [putlist(R)|OptPilRest]) :-
	sym_name(Str,'.',2),
	popt2(PilRest, OptPilRest).

popt21(getcon(R,Sym), PilRest, [getnil(R)|OptPilRest]) :-
	sym_name(Sym,[],0),
	popt2(PilRest, OptPilRest).
popt21(putcon(R,Sym), PilRest, OptPil) :-
	sym_name(Sym,N,A),
	( N == [], A =:= 0 -> OptPil = [putnil(R)|OptPilRest]
	; N == pass_thru, A =:= -1 -> OptPil = OptPilRest
	),
	popt2(PilRest, OptPilRest).
popt21(unicon(Sym), PilRest, [uninil|OptPilRest]) :-
	sym_name(Sym,[],0),
	popt2(PilRest, OptPilRest).
popt21(bldcon(Sym), PilRest, [bldnil|OptPilRest]) :-
	sym_name(Sym,[],0),
	popt2(PilRest, OptPilRest).
popt21(Inst, PilRest, [Inst|OptPilRest]) :-
	popt2(PilRest, OptPilRest).


/*----------------------------------------------------------------------*/
/* "popt3" optimizes common cases of list gets and collapses
    some commonly-occurring consecutive unification instructions.       */
/*----------------------------------------------------------------------*/

popt3([], []).
popt3([Inst|Rest], Pil) :- popt31(Inst, Rest, Pil).

popt31(getlist(R0), [Inst|Rest], OptInsts) :-
	popt31_getlist(Inst, Rest, OptInsts, R0).
popt31(getstr(R,Str), [unitvar(R1)|Rest], OptInsts) :-
	popt31_getstr(Rest, OptInsts, R, R1, Str).
popt31(unitvar(R1), [Inst|Rest], OptInsts) :-
	popt31_unitvar(Inst, Rest, OptInsts, R1).
popt31(unitval(R1), [Inst|Rest], OptInsts) :-
	popt31_unitval(Inst, Rest, OptInsts, R1).
popt31(univoid, [Inst|Rest], OptInsts) :-
	popt31_univoid(Inst, Rest, OptInsts).
popt31(bldvoid, [bldvoid|R], [bld_n_voids(N)|OptRest]) :-
	collect_voids(R, bldvoid, 2, N, Rest),
	popt3(Rest, OptRest).
popt31(Inst, Rest, [Inst|OptRest]) :- 
	popt3(Rest, OptRest).

popt31_getlist(unitvar(R1), [Inst|Rest], OptInsts, R0) :-
	popt31_getlist_tvar(Inst, Rest, OptInsts, R0, R1).
popt31_getlist(unitval(R1), [unitvar(R2)|Rest],
	       [getlist_tval_tvar(R0,R1,R2)|OptRest], R0) :- !,
	popt3(Rest, OptRest).
popt31_getlist(unitval(R1), [univoid|Rest],
	       [getlist_tval_void(R0,R1)|OptRest], R0) :-
	popt3(Rest, OptRest).
popt31_getlist(univoid, [unitvar(R2)|Rest],
	       [getlist_void_tvar(R0,R2)|OptRest], R0) :-
	popt3(Rest, OptRest).

popt31_getlist_tvar(univoid, Rest, [NewInst|OptRest], R0, R1) :-
	NewInst = getlist_tvar_void(R0,R1),
	popt3(Rest, OptRest).
popt31_getlist_tvar(unitvar(R2), Rest, [NewInst|OptRest], R0, R1) :-
	NewInst = getlist_tvar_tvar(R0,R1,R2),
	popt3(Rest, OptRest).
popt31_getlist_tvar(unipvar(V), Rest, [NewInst|OptRest], R0, R1) :-
	NewInst = getlist_tvar_pvar(R0,R1,V),
	popt3(Rest, OptRest).

popt31_getstr([unitvar(R2),unitvar(R3)|Rest], OptInsts, R1, R1, Str) :- !,
	%% Note R1, R1 condition before the cut !!
	OptInsts = [getstr_tvar_tvar_tvar(R1,R2,R3,Str)|OptRest],
	popt3(Rest, OptRest).
popt31_getstr([unitvar(R2)|Rest], OptInsts, R, R1, Str) :- !,
	OptInsts = [getstr_tvar_tvar(R,R1,R2,Str)|OptRest],
	popt3(Rest, OptRest).
popt31_getstr(Rest, OptInsts, R, R1, Str) :-
	OptInsts = [getstr_tvar(R,R1,Str)|OptRest],
	popt3(Rest, OptRest).

popt31_unitvar(getlist(R1), Rest, [unilist|OptRest], R1) :-
 	popt3(Rest, OptRest).
popt31_unitvar(univoid, Rest, [uni_tvar_void(R1)|OptRest], R1) :-
	popt3(Rest, OptRest).
popt31_unitvar(unitvar(R2), Rest, [uni_tvar_tvar(R1,R2)|OptRest], R1) :-
	popt3(Rest, OptRest).
popt31_unitvar(unitval(R2), Rest, [uni_tvar_tval(R1,R2)|OptRest], R1) :-
	popt3(Rest, OptRest).

popt31_unitval(unitvar(R2), Rest, [uni_tval_tvar(R1,R2)|OptRest], R1) :-
	popt3(Rest, OptRest).
popt31_unitval(unitval(R2), Rest, [uni_tval_tval(R1,R2)|OptRest], R1) :-
	popt3(Rest, OptRest).

popt31_univoid(univoid, R, [uni_n_voids(N)|OptRest]) :-
	collect_voids(R, univoid, 2, N, Rest),
	popt3(Rest, OptRest).
popt31_univoid(unitvar(R2), Rest, [uni_void_tvar(R2)|OptRest]) :-
	popt3(Rest, OptRest).

/*
 * collects a sequence of specific instructions given as second arg
 */
collect_voids([Inst|Insts], Inst, Nin, Nout, Remaining) :-
	!,
	Nmid is Nin + 1,
	collect_voids(Insts, Inst, Nmid, Nout, Remaining).
collect_voids(Insts, _, N, N, Insts).

/*----------------------------------------------------------------------*/
/* "popt4" eliminates some redundant "execute"s, mainly those arising from
   the generation of new predicates for handling cuts and indexing.     */
/*----------------------------------------------------------------------*/

popt4([], _, []).
popt4([label(T,(Sym,K))|Rest], Seen, [label(T,(Sym,K))|ORest]) :-
	!,
	( Sym = disj(_)
	; memberchk(Sym, Seen)
	),
	popt4(Rest, Seen, ORest).
popt4([execute(Sym), label(T,(Sym,K))|Rest], Seen, OList) :-
	!,
	( Sym = disj, OList = [execute(Sym),label(T,(Sym,K))|ORest]
	; /* Sym = sym(...) */
	  popt_chkmember(Sym,Seen,SFlag),
	  ( SFlag =:= 1,	/* already seen, "execute" can't be dropped */
	    OList = [execute(Sym),label(T,(Sym,K))|ORest]
	  ; SFlag =\= 1,	/* not seen, "execute" can be dropped */
	    OList = [label(T,(Sym,K))|ORest]
	  )
	),
	popt4(Rest, Seen, ORest).
popt4([Inst|Rest], Seen, [Inst|ORest]) :-
	popt4(Rest,Seen,ORest).

popt_chkmember(P, L, Flag) :-
	( var(L) -> L = [P|_], Flag = 0
	; L = [P1|L1], ( P = P1 -> Flag = 1 ; popt_chkmember(P,L1,Flag) )
	).

/*----------------------------------------------------------------------*/
/* "popt_alloc" exchanges the places of the allocate and all the get-type
   instructions so as to make more effective the indexing optimizations
   that the loader performs; i.e., bypass the get instruction.
   This also avoids the cost of the allocate in case that the gets fail.
   Of course, only strictly a-reg instructions can be moved.            */
/*----------------------------------------------------------------------*/

popt_alloc([], []).
popt_alloc([Inst|Rest], Pil) :- popt_alloc_1(Inst, Rest, Pil).

popt_alloc_1(AllocInst, [Inst|Rest], NewInsts) :-
	AllocInst = allocate_gc(_,_),
	popt_alloc_2(Inst, AllocInst, Rest, NewInsts).
popt_alloc_1(Inst, PilRest, [Inst|OptPilRest]) :-
	popt_alloc(PilRest, OptPilRest).

popt_alloc_2(gettval(R1,R2), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [gettval(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(getcon(R,I), AllocInst, Rest, NewInsts) :-
	NewInsts = [getcon(R,I)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(getnil(R), AllocInst, Rest, NewInsts) :-
	NewInsts = [getnil(R)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(getstr(R,I), AllocInst, Rest, NewInsts) :-
	NewInsts = [getstr(R,I)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(getlist(R), AllocInst, Rest, NewInsts) :-
	NewInsts = [getlist(R)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(unitvar(R), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [unitvar(R)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(unitval(R), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [unitval(R)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(unicon(I), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [unicon(I)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(uninil, AllocInst, Rest, NewInsts) :- % added
	NewInsts = [uninil|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(getnumcon(R,N), AllocInst, Rest, NewInsts) :-
	NewInsts = [getnumcon(R,N)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(uninumcon(N), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [uninumcon(N)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(uni_tvar_tvar(R1,R2), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [uni_tvar_tvar(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(uni_tvar_tval(R1,R2), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [uni_tvar_tval(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(uni_tval_tvar(R1,R2), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [uni_tval_tvar(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(uni_tval_tval(R1,R2), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [uni_tval_tval(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(uni_n_voids(N), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [uni_n_voids(N)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(uni_void_tvar(R2), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [uni_void_tvar(R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(uni_tvar_void(R1), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [uni_tvar_void(R1)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(getlist_tval_tvar(R1,R2,R3), AllocInst, Rest, NewInsts) :-
	NewInsts = [getlist_tval_tvar(R1,R2,R3)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(getlist_tvar_tvar(R1,R2,R3), AllocInst, Rest, NewInsts) :-
	NewInsts = [getlist_tvar_tvar(R1,R2,R3)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(getlist_void_tvar(R1,R2), AllocInst, Rest, NewInsts) :-
	NewInsts = [getlist_void_tvar(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(getlist_tvar_void(R1,R2), AllocInst, Rest, NewInsts) :-
	NewInsts = [getlist_tvar_void(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(getlist_tval_void(R1,R2), AllocInst, Rest, NewInsts) :-
	NewInsts = [getlist_tval_void(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(getstr_tvar(R,R1,Str), AllocInst, Rest, NewInsts) :-
	NewInsts = [getstr_tvar(R,R1,Str)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(getstr_tvar_tvar(R,R1,R2,Str), AllocInst, Rest, NewInsts) :-
	NewInsts = [getstr_tvar_tvar(R,R1,R2,Str)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(getstr_tvar_tvar_tvar(R1,R2,R3,Str), AllocInst, Rest, NewInsts) :-
	NewInsts = [getstr_tvar_tvar_tvar(R1,R2,R3,Str)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(getfloat(R,F), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [getfloat(R,F)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(unifloat(F), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [unifloat(F)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(unilist, AllocInst, Rest, NewInsts) :- % added
	NewInsts = [unilist|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(univoid, AllocInst, Rest, NewInsts) :- % added
	NewInsts = [univoid|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(gettbreg(R), AllocInst, Rest, NewInsts) :- % added - necessary?
	NewInsts = [gettbreg(R)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(term_comp(R1,R2,R3), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [term_comp(R1,R2,R3)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(movreg(R1,R2), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [movreg(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(negate(R), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [negate(R)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(and(R1,R2), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [and(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(or(R1,R2), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [or(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(logshiftl(R1,R2), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [logshiftl(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(logshiftr(R1,R2), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [logshiftr(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(addreg(R1,R2), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [addreg(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(subreg(R1,R2), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [subreg(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(mulreg(R1,R2), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [mulreg(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(divreg(R1,R2), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [divreg(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(idivreg(R1,R2), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [idivreg(R1,R2)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
% popt_alloc_2(int_test_eq(R,I,L), AllocInst, Rest, NewInsts) :- % added
% 	NewInsts = [int_test_eq(R,I,L)|NewRest],
% 	popt_alloc_1(AllocInst, Rest, NewRest).
% popt_alloc_2(int_test_neq(R,I,L), AllocInst, Rest, NewInsts) :- % added
% 	NewInsts = [int_test_neq(R,I,L)|NewRest],
% 	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(add_byte(A,R), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [add_byte(A,R)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(sub_byte(A,R), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [sub_byte(A,R)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(jump(L), AllocInst, Rest, NewInsts) :- % added - necessary?
	NewInsts = [jump(L)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(jumpz(R,L), AllocInst, Rest, NewInsts) :- % added - necessary?
	NewInsts = [jumpz(R,L)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(jumpnz(R,L), AllocInst, Rest, NewInsts) :- % added - necessary?
	NewInsts = [jumpnz(R,L)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(jumplt(R,L), AllocInst, Rest, NewInsts) :- % added - necessary?
	NewInsts = [jumplt(R,L)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(jumple(R,L), AllocInst, Rest, NewInsts) :- % added - necessary?
	NewInsts = [jumple(R,L)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(jumpgt(R,L), AllocInst, Rest, NewInsts) :- % added - necessary?
	NewInsts = [jumpgt(R,L)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(jumpge(R,L), AllocInst, Rest, NewInsts) :- % added - necessary?
	NewInsts = [jumpge(R,L)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(fail, AllocInst, Rest, NewInsts) :- % added - hardly useful...
	NewInsts = [fail|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(builtin(W), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [builtin(W)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(unifunc(W,R), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [unifunc(W,R)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).
popt_alloc_2(putvoid(R), AllocInst, Rest, NewInsts) :- % added
	NewInsts = [putvoid(R)|NewRest],
	popt_alloc_1(AllocInst, Rest, NewRest).

/*----------------------------------------------------------------------*/

force_peephole_loaded.

/*----------------------------------------------------------------------*/
