/* File:      tab_CP.P
** Author(s): Terrance Swift, Kostis Sagonas
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: tab_CP.P,v 1.1 2003/04/04 16:38:01 kostis Exp $
** 
*/


tab_CP__transform(module(Symtab,DcList,ClList,QrList),
		  module(Symtab,DcList,ClList1,QrList)) :-
	transform_preds(ClList, Symtab, ClList1).

:- mode transform_preds(+,#,-).

transform_preds([], _Symtab, []).
transform_preds([Pred|Preds], Symtab, OutPreds) :-
	Pred = pred(Sym,_Clauses,_Prag),
	sym_tabled(Sym, Tabled),
	( Tabled =:= 0 -> OutPreds = [Pred|OPs]
	; transform_tab_pred(Pred, Symtab, OutPreds, OPs)
	),
	transform_preds(Preds, Symtab, OPs).

/*======================================================================*
  TransNeeded =\= 0 if the code of the tabled predicate needs
  transformation, that is, if it cannot be optimized. Conceptually,
  the default consists of transforming a tabled predicate

        :- table tabpred.
	tabpred :- body1.
	tabpred :- body2.

  into
        :- table tabpred.
	tabpred :- '_$pred'.
	
	'_$pred' :- body1.
	'_$pred' :- body2.

  where _$pred is NOT tabled.  Optimization in this context consists
  of forgoing the transformation.  

  Currently the transformation is made when
	1) One of the bodies has a hard cut.
  or 
	2) When the indexing will generate more than one try for any
	   entry point (at least if I got my check right).

  As can be seen from corresponding the .H file, I have cannibalized
  other pieces of code to do this. I have documented this code in their
  original source files, where needed.
 *======================================================================*/

:- mode transform_tab_pred(+,#,#,#).

transform_tab_pred(Pred, SymTab, OutPreds, OPs) :-
	Pred = pred(_Sym,Clauses,_Prag),
	have_hardcut(Clauses, Hc),
	must_keep_tabpoint(Pred, Res),
	TransNeeded is Hc + Res,
	( TransNeeded =:= 0 -> OutPreds = [Pred|OPs]
	; transform_tab_pred1(Pred, SymTab, OutPreds, OPs)
	).

:- mode transform_tab_pred1(+,#,#,#).

transform_tab_pred1(pred(Sym,Clauses,Prag), SymTab, OutPreds, OPs) :-
	sym_name(Sym, Name, Arity),
	length(VarList, Arity),
	new_pred_name(Name, NewName),
	Head =.. [Name|VarList],
	Body =.. [NewName|VarList],
	%% The following slightly (ab)uses parse_preddef/4 in order to
	%% modify the SymTab in the process of performing the transformation
	%% STEP 1: Create the tabpred :- '_$pred' tabled clause.
	parse_preddef(Head, Body, ClList, SymTab),
	ClList = [NewPred|_],
	%% STEP 2: Create the '_$pred' :- ... Prolog clauses.
	% parse_preddef(Body, Head, ClList1, SymTab),
	% ClList1 = [pred(NewSym,_Cls1,_Prag1)|_],
	%% sym_insert(NewName, Arity, [pred,comp_gen,defined,used], SymTab, NewSym),
	sym_insert(NewName, Arity, [pred,defined,used], SymTab, NewSym),
	OutPreds = [NewPred,pred(NewSym,Clauses,Prag)|OPs].

:- mode new_pred_name(+,-).

new_pred_name(Name, NewName) :- str_cat('_$', Name, NewName).

:- mode must_keep_tabpoint(+,-).

must_keep_tabpoint(Pred, Res) :-
	Pred = pred(_Sym,Clauses,_Prag),
	( pred_indexarg(Pred, IndxArg),
	  ( IndxArg =:= 0 -> Res = 0 % predicate will not be indexed
	  ; index_argpats(Clauses, IndxArg, Argpat),
	    index_split(Argpat, Groups),
	    ( member(arglabel(g,_,_), Groups) -> Res = 1 ; Res = 0 )
	  )
	).

