/* File:      abs_machine.h
** Author(s): Kostis Sagonas
** Contact:   kostis@csd.uu.se
** 
** Copyright (C) Kostis Sagonas, 2000-2002
** 
** XXX is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XXX is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: abs_machine.h,v 1.33 2007/03/17 07:38:18 kostis Exp $
** 
*/

/*----------------------------------------------------------------------*/

#include "xxx_inst_defs.h"	       /* generated by the do_me script */

#include "jiti.h"
#include "biassert.h"

#undef JITI_DEBUG

#define JIT_INDEXING

#ifdef JIT_INDEXING
#warning COMPILED WITH JIT INDEXING INSTRUCTIONS
#endif


/*----------------------------------------------------------------------*/
/* Macros for accessing arguments of instructions.                      */
/*----------------------------------------------------------------------*/

#define opregaddr(OFFSET)	(rreg+(*(lpcreg+OFFSET)))
#define opvaraddr(OFFSET)	(ereg-(*(lpcreg+OFFSET)))

#define opreg(OFFSET)		cell(opregaddr(OFFSET))
#define opvar(OFFSET)	 	cell(opvaraddr(OFFSET))
#define opbyte(OFFSET)	 	cell(lpcreg+OFFSET)
#define opword(OFFSET)	 	cell((CPtr)(lpcreg+OFFSET))
#define oppsc(OFFSET)	 	cell((Psc *)(lpcreg+OFFSET))
#define oplabel(OFFSET) 	cell((byte **)(lpcreg+OFFSET))
#define optable(OFFSET) 	cell((TIFptr *)(lpcreg+OFFSET))

/*----------------------------------------------------------------------*/
/* The following macros work for all CPs.  Make sure this remains	*/
/* the case...								*/
/*----------------------------------------------------------------------*/

#define Fail1 lpcreg = cp_pcreg(breg)

#define restore_trail_condition_registers(BREG) \
  ebreg = cp_ebreg(BREG); \
  hbreg = cp_hreg(BREG);

/*----------------------------------------------------------------------*/

#define subtryme(OP1, OP2) \
{                                                                    \
  register CPtr cps_top;   /* cps_top only needed for efficiency */  \
                                                                     \
  save_find_locx(ereg);	   /* sets ebreg to the top of the E-stack */\
  check_tcpstack_overflow();                                         \
  cps_top = top_of_cpstack;                                          \
  save_registers(cps_top, OP1, rreg);                                \
  save_choicepoint(cps_top, ereg, (byte *)OP2, breg);                \
  breg = cps_top;                                                    \
  hbreg = hreg;                                                      \
} /* end of subtryme */

/*----------------------------------------------------------------------*/

#define restore_sub(OP1) \
{                                      \
  register CPtr tbreg;                 \
                                       \
  tbreg = breg;                        \
  undo_bindings(tbreg);                \
  ptcpreg = cp_ptcp(tbreg);            \
  delayreg = cp_pdreg(tbreg);          \
  restore_some_wamregs(tbreg, ereg);   \
  restore_registers(tbreg, OP1, rreg); \
} /* end of restore_sub */

#define trust_restore_sub(OP1) \
{                                          \
  restore_sub(OP1);                        \
  breg = cp_prevbreg(breg);                \
  restore_trail_condition_registers(breg); \
}

/*----------------------------------------------------------------------*/

#define table_restore_sub(OP1) \
{                                            \
  register CPtr tbreg;                       \
                                             \
  tbreg = breg;                              \
  undo_bindings(tbreg);                      \
  /* This CP should be used for the dependency graph */\
  ptcpreg = tcp_subgoal_ptr(tbreg);          \
  delayreg = NULL;                           \
  restore_some_wamregs(tbreg, ereg);         \
  table_restore_registers(tbreg, OP1, rreg); \
} /* end of table_restore_sub */

#define trust_table_restore_sub(OP1) \
{                                           \
  table_restore_sub(OP1);                   \
  { CPtr xt1 = tcp_prevbreg(breg);          \
    restore_trail_condition_registers(xt1); \
  }                                         \
}

/*======================================================================*/
/* The main emulator loop below.  A correct, but quite slow version of	*/
/* this function existed for quite some time in XSB, originally written	*/
/* by David S. Warren and Jiyang Xu (mainly), Terry Swift, and myself.	*/
/*									*/
/* Prompted by discussions with Bart Demoen, I completely rewrote this	*/
/* function during the period 10/1999 - 3/2000 optimizing it for speed.	*/
/* A bunch of new instructions were introduced and special attention to */
/* register allocation was paid	throughout.  I think the end-result is  */
/* quite satisfactory.						       	*/
/*									*/
/* There might be a couple of things that look a bit weird below, but	*/
/* make sure that BEFORE you change something look first on how your	*/
/* change looks in assembly.  Also, you really have to have a VERY GOOD	*/
/* REASON to add new variables anywhere and especially in the beginning	*/
/* of the function!							*/
/*						- Kostis Sagonas.	*/
/*======================================================================*/

#ifdef THREADED
void **instr_label_tab;
#endif

static int emuloop(byte *startaddr)
{
  register byte *lpcreg LPCREG;
  register CPtr rreg;
  /* The following is used specially in unification; please use carefully */
  register CPtr sreg;   /* current build or unify field */
/*----------------------------------------------------------------------*/

#include "xxx_inst_labels.h" /* generated by the do_me script */

/*----------------------------------------------------------------------*/

#if (defined(GC) && defined(GC_TEST))
/* Used only in the garbage collection test; does not affect emulator o/w */
#define GC_INFERENCES 233 /* make sure the garbage collection test is hard */
  static int infcounter = 0;
#endif

#ifdef THREADED
  instr_label_tab = linstr_label_tab;  /* needed by the loader & assert */
  if (startaddr == NULL) return 0;
#endif

  rreg = reg;
  sreg = NULL;
  lpcreg = (byte *)&reset_inst;  /* start by initializing abort handler */

#if ((!defined(JUMP_TABLE)) && (!defined(THREADED)))
contcase:     /* the main loop */
#elif defined(THREADED)
  MoreMoreYesGimmeMore;
#endif

#ifdef DEBUG
  if (flags[PIL_TRACE]) debug_inst(lpcreg, ereg);
  xctr++;
#endif
#ifdef PROFILE
  if (flags[PROFFLAG]) {
    inst_table[(int) *(lpcreg)][sizeof(Cell)+1]
      = inst_table[(int) *(lpcreg)][sizeof(Cell)+1] + 1;
    if (flags[PROFFLAG] > 1 && (int) *lpcreg == builtin) 
      builtin_table[(int) *(lpcreg+builtin_arg_1)][1] = 
	builtin_table[(int) *(lpcreg+builtin_arg_1)][1] + 1;
  }
#endif

  switch (*lpcreg) {
  
  case getpvar:  /* PVR */
  getpvar_label:
  {
    register CPtr op1;
    register Cell op2;
      
    op1 = opvaraddr(getpvar_arg_1);
    op2 = opreg(getpvar_arg_2);
    lpcreg += getpvar_sizeof;
    /* trailing is needed here because this instruction can also be
       generated *after* the occurrence of the first call - kostis */
    bind_copy(op1, op2);      /* In WAM bld_copy() */
    MoreMoreYesGimmeMore;
  }

  case getpval: /* PVR */
  getpval_label:
  {
    register Cell op1;
    register Cell op2;

    op1 = opvar(getpval_arg_1);
    op2 = opreg(getpval_arg_2);
    deref(op1);
    deref(op2);
    lpcreg += getpval_sizeof;
    if ((isref(op1)) && (isnonvar(op2))) { bind_copy((CPtr)(op1), op2); }
    else
      if ((isnonvar(op1)) && (isref(op2))) { bind_copy((CPtr)(op2), op1); }
      else if (! unify(op1,op2)) Fail1; /* modifies lpcreg */
    MoreMoreYesGimmeMore;
  }

  case getstrv: /* PPV-S */
  getstrv_label:
  {
    register Cell op1;
    register Cell strct;
    
    op1 = opvar(getstrv_arg_1);
    strct = opword(getstrv_arg_2);
    deref2(op1, goto getstrv_is_free);
    if (isstruct(op1)) {
      op1 = (Cell)(struct_val(op1));
      if (*((Psc *)op1) == (Psc)strct) {
	lpcreg += getstrv_sizeof;
	sreg = (CPtr)op1 + 1;
      }
      else Fail1;
    }
#ifdef ATTR_VARS
    else if (isattv(op1)) {
      attv_dbgmsg(">>>> ATTV getstrv, interrupt needed\n");
      lpcreg += getstrv_sizeof;
      add_interrupt(op1, makestruct(hreg));
      new_heap_functor(hreg, (Psc)strct);
      sreg = NULL;
    }
#endif
    else Fail1;
    MoreMoreYesGimmeMore;

  getstrv_is_free:
    lpcreg += getstrv_sizeof;
    bind_struct((CPtr)(op1), (Pair)hreg);
    new_heap_functor(hreg, (Psc)strct);
    sreg = NULL;
    MoreMoreYesGimmeMore;
  }

  case gettval: /* PRR */
  gettval_label:
  {
    register Cell op1;
    register Cell op2;
    
    op1 = opreg(gettval_arg_1);
    op2 = opreg(gettval_arg_2);
    deref(op1);
    deref(op2);
    lpcreg += gettval_sizeof;
    if ((isref(op1)) && (isnonvar(op2))) { bind_copy((CPtr)(op1), op2); }
    else
      if ((isnonvar(op1)) && (isref(op2))) { bind_copy((CPtr)(op2), op1); }
      else if (! unify(op1,op2)) Fail1; /* modifies lpcreg */
    MoreMoreYesGimmeMore;
  }

  case getcon: /* PPR-C */
  getcon_label:
  {
    register Cell op1;
    register Cell op2;
    
    op1 = opreg(getcon_arg_1);
    op2 = opword(getcon_arg_2);
    deref2(op1, goto getcon_op1_is_free);
    if (op1 == op2) { lpcreg += getcon_sizeof; }
#ifdef ATTR_VARS
    else if (isattv(op1)) {
      attv_dbgmsg(">>>> ATTV getcon, interrupt needed\n");
      lpcreg += getcon_sizeof;
      add_interrupt(op1, op2);
    }
#endif
    else Fail1;
    MoreMoreYesGimmeMore;

  getcon_op1_is_free:
    lpcreg += getcon_sizeof;
    bind_copy((CPtr)(op1), op2);
    MoreMoreYesGimmeMore;
  }

  case getnil: /* PPR */
  getnil_label:
  {
    register Cell op1;

    op1 = opreg(getnil_arg_1);
    deref2(op1, goto getnil_is_free);
    if (isnil(op1)) { lpcreg += getnil_sizeof; }
#ifdef ATTR_VARS
    else if (isattv(op1)) {
      attv_dbgmsg(">>>> ATTV getnil, interrupt needed\n");
      lpcreg += getnil_sizeof;
      add_interrupt(op1, makenil);
    }
#endif
    else Fail1;
    MoreMoreYesGimmeMore;

  getnil_is_free:
    lpcreg += getnil_sizeof;
    bind_nil((CPtr)(op1));
    MoreMoreYesGimmeMore;
  }

  case getstr: /* PPR-S */
  getstr_label:
  {
    register Cell op1;
    register Cell strct;

    op1 = opreg(getstr_arg_1);
    strct = opword(getstr_arg_2);
    deref2(op1, goto getstr_is_free);
    if (isstruct(op1)) {
      op1 = (Cell)(struct_val(op1));
      if (*((Psc *)op1) == (Psc)strct) {
	lpcreg += getstr_sizeof;
	sreg = (CPtr)op1 + 1;
      }
      else Fail1;
    }
#ifdef ATTR_VARS
    else if (isattv(op1)) {
      attv_dbgmsg(">>>> ATTV getstr, interrupt needed\n");
      lpcreg += getstr_sizeof;
      add_interrupt(op1, makestruct(hreg));
      new_heap_functor(hreg, (Psc)strct);
      sreg = NULL;
    }
#endif
    else Fail1;
    MoreMoreYesGimmeMore;

  getstr_is_free:
    lpcreg += getstr_sizeof;
    bind_struct((CPtr)(op1), (Pair)hreg);
    new_heap_functor(hreg, (Psc)strct);
    sreg = NULL;
    MoreMoreYesGimmeMore;
  }

  case getlist: /* PPR */
  getlist_label:
  {
    register Cell op;

    op = opreg(getlist_arg_1);
    deref2(op, goto getlist_op_is_free);
    if (islist(op)) {
      lpcreg += getlist_sizeof;
      sreg = clref_val(op);
    }
#ifdef ATTR_VARS
    else if (isattv(op)) {
      attv_dbgmsg(">>>> ATTV getlist, interrupt needed\n");
      lpcreg += getlist_sizeof;
      add_interrupt(op, makelist(hreg));
      sreg = NULL;
    }
#endif
    else Fail1;
    MoreMoreYesGimmeMore;

  getlist_op_is_free:
    lpcreg += getlist_sizeof;
    bind_list((CPtr)(op), hreg);
    sreg = NULL;
    MoreMoreYesGimmeMore;
  }

  case unipvar: /* PPV */
  unipvar_label:
  {
    register CPtr op1;

    op1 = opvaraddr(unipvar_arg_1);
    lpcreg += unipvar_sizeof;
    if (!sreg) { /* if (flag == WRITE) */
      bind_ref(op1, hreg);  /* tls 12/8/92 */
      new_heap_free(hreg);
    } else {
      /* also introduce trailing here - bmd & kostis
         was: bld_copy(op1, *(sreg++)); */
      bind_copy(op1, *(sreg));
      sreg++;
    }
    MoreMoreYesGimmeMore;
  }

  case unipval: /* PPV */
  unipval_label:
  {
    register Cell op1;

    op1 = opvar(unipval_arg_1);
    if (!sreg) { /* if (flag == WRITE) */
      lpcreg += unipval_sizeof;
      bld_val(op1);
    } 
    else {
      if (! unify(op1, *sreg)) Fail1;
      else { lpcreg += unipval_sizeof; sreg++; }
    } 
    MoreMoreYesGimmeMore;
  }

  case unitvar: /* PPR */
  unitvar_label:
  {
    register CPtr op1;

    op1 = opregaddr(unitvar_arg_1);
    lpcreg += unitvar_sizeof;
    if (!sreg) { /* if (flag == WRITE) */
      bld_ref(op1, hreg);
      new_heap_free(hreg);
    }
    else {
      bld_copy(op1, *(sreg++));
    }
    MoreMoreYesGimmeMore;
  }

  case unitval: /* PPR */
  unitval_label:
  {
    register Cell op1;

    op1 = opreg(unitval_arg_1);
    if (!sreg) { /* if (flag == WRITE) */
      lpcreg += unitval_sizeof;
      bld_val(op1); 
    }
    else {
      if (! unify(op1, *sreg)) Fail1;
      else { lpcreg += unitval_sizeof; sreg++; }
    } 
    MoreMoreYesGimmeMore; 
  }

  case unicon: /* PPP-C */
  unicon_label:
  {
    register Cell op1; /* contains a tagged atom */

    op1 = opword(unicon_arg_1);
    if (!sreg) { /* if (flag == WRITE) */
      lpcreg += unicon_sizeof;
      bld_copy(hreg++, op1);
      MoreMoreYesGimmeMore;
   }
    else {
      register Cell cell = *(sreg++);
      deref2(cell, goto unicon_cell_is_free);
      if (cell == op1) { lpcreg += unicon_sizeof; }
#ifdef ATTR_VARS
      else if (isattv(cell)) {
	attv_dbgmsg(">>>> ATTV unicon, interrupt needed\n");
	lpcreg += unicon_sizeof;
	add_interrupt(cell, op1);
      }
#endif
      else Fail1;
      MoreMoreYesGimmeMore;

    unicon_cell_is_free:
      lpcreg += unicon_sizeof;
      bind_copy((CPtr)(cell), op1);
      MoreMoreYesGimmeMore;
    }
  }

  case uninil: /* PPP */
  uninil_label:
  {
    if (!sreg) { /* if (flag == WRITE) */
      lpcreg += uninil_sizeof;
      new_heap_nil(hreg);
      MoreMoreYesGimmeMore;
    }
    else {
      register Cell cell = *(sreg++);
      deref2(cell, goto uninil_is_free);
      if (isnil(cell)) { lpcreg += uninil_sizeof; }
#ifdef ATTR_VARS
      else if (isattv(cell)) {
	attv_dbgmsg(">>>> ATTV uninil, interrupt needed\n");
	lpcreg += uninil_sizeof;
	add_interrupt(cell, makenil);
      }
#endif
      else Fail1;
      MoreMoreYesGimmeMore;
      
    uninil_is_free:
      lpcreg += uninil_sizeof;
      bind_nil((CPtr)(cell));
      MoreMoreYesGimmeMore;
    }
  }

  case getnumcon: /* PPR-N */
  getnumcon_label:
  {
    register Cell op1;
    register Cell op2;

    op1 = opreg(getnumcon_arg_1);
    op2 = opword(getnumcon_arg_2);
    deref2(op1, goto getnumcon_op1_is_free);
    if (op1 == op2) { lpcreg += getnumcon_sizeof; }
#ifdef ATTR_VARS
    else if (isattv(op1)) {
      attv_dbgmsg(">>>> ATTV getnumcon, interrupt needed\n");
      lpcreg += getnumcon_sizeof;
      add_interrupt(op1, op2);
    }
#endif
    else Fail1;
    MoreMoreYesGimmeMore;

  getnumcon_op1_is_free:
    lpcreg += getnumcon_sizeof;
    bind_copy((CPtr)(op1), op2);
    MoreMoreYesGimmeMore;
  }

  case putnumcon: /* PPR-N */
  putnumcon_label:
  {
    register CPtr op1;
    register Cell op2;

    op1 = opregaddr(putnumcon_arg_1);
    op2 = opword(putnumcon_arg_2);
    lpcreg += putnumcon_sizeof;
    bld_copy(op1, op2);
    MoreMoreYesGimmeMore;
  }

  case putpvar: /* PVR */
  putpvar_label:
  {
    register CPtr op1;
    register CPtr op2;

    op1 = opvaraddr(putpvar_arg_1);
    op2 = opregaddr(putpvar_arg_2);
    lpcreg += putpvar_sizeof;
    bld_free(op1);
    bld_ref(op2, op1);
    MoreMoreYesGimmeMore;
  }

  case putpval: /* PVR */
  putpval_label:
  {
    register Cell op1;
    register CPtr op2;

    op1 = opvar(putpval_arg_1);
    op2 = opregaddr(putpval_arg_2);
    lpcreg += putpval_sizeof;
    bld_copy(op2, op1);
    MoreMoreYesGimmeMore;
  }

  case puttvar: /* PRR */
  puttvar_label:
  {
    register CPtr op1;
    register CPtr op2;

    op1 = opregaddr(puttvar_arg_1);
    op2 = opregaddr(puttvar_arg_2);
    lpcreg += puttvar_sizeof;
    bld_ref(op1, hreg);
    bld_ref(op2, hreg);
    new_heap_free(hreg);
    MoreMoreYesGimmeMore;
  }

  case putstrv: /*  PPV-S */
  putstrv_label:
  {
    register CPtr op1;
    register Psc  op2;

    op1 = opvaraddr(putstrv_arg_1);
    op2 = oppsc(putstrv_arg_2);
    lpcreg += putstrv_sizeof;
    bind_struct(op1, (Pair)hreg);  /* tls 12/8/92 */
    new_heap_functor(hreg, op2);
    MoreMoreYesGimmeMore;
  }

  case putcon: /* PPR-C */
  putcon_label:
  {
    register CPtr op1;
    register Cell op2;

    op1 = opregaddr(putcon_arg_1);
    op2 = opword(putcon_arg_2);
    lpcreg += putcon_sizeof;
    bld_copy(op1, op2);
    MoreMoreYesGimmeMore;
  }
 
  case putnil: /* PPR */
  putnil_label:
  {
    register CPtr op1;

    op1 = opregaddr(putnil_arg_1);
    lpcreg += putnil_sizeof;
    bld_nil(op1);
    MoreMoreYesGimmeMore;
  }
 
  case putstr: /* PPR-S; differs from putstrv since it pulls from a register*/
  putstr_label:
  {
    register CPtr op1;
    register Psc  op2;

    op1 = opregaddr(putstr_arg_1);
    op2 = oppsc(putstr_arg_2);
    lpcreg += putstr_sizeof;
    bld_struct(op1, hreg);
    new_heap_functor(hreg, op2);
    MoreMoreYesGimmeMore;
  }

  case putlist: /* PPR */
  putlist_label:
  {
    register CPtr op1;

    op1 = opregaddr(putlist_arg_1);
    lpcreg += putlist_sizeof;
    bld_list(op1, hreg);
    MoreMoreYesGimmeMore;
  }

  case bldpvar: /* PPV */
  bldpvar_label:
  {
    register CPtr op1;

    op1 = opvaraddr(bldpvar_arg_1);
    lpcreg += bldpvar_sizeof;
    /* tls 12/8/92 */
    bind_ref(op1, hreg); /* trailing is needed: if o/w see ai_tests */
    new_heap_free(hreg);
    MoreMoreYesGimmeMore;
  }
 
  case bldpval: /* PPV */
  bldpval_label:
  {
    register Cell op1;

    op1 = opvar(bldpval_arg_1);
    lpcreg += bldpval_sizeof;
    bld_val(op1);
    MoreMoreYesGimmeMore;
  }
 
  case bldtvar: /* PPR */
  bldtvar_label:
  {
    register CPtr op1;

    op1 = opregaddr(bldtvar_arg_1);
    lpcreg += bldtvar_sizeof;
    bld_ref(op1, hreg);
    new_heap_free(hreg);
    MoreMoreYesGimmeMore;
  }
 
  case bldtval: /* PPR */
  bldtval_label:
  {
    register Cell op1;

    op1 = opreg(bldtval_arg_1);
    lpcreg += bldtval_sizeof;
    bld_val(op1);
    MoreMoreYesGimmeMore;
  }
 
  case bldcon: /* PPP-C */
  bldcon_label:
  {
    register Cell op1;

    op1 = opword(bldcon_arg_1);
    lpcreg += bldcon_sizeof;
    bld_copy(hreg++, op1);
    MoreMoreYesGimmeMore;
  }
 
  case bldnil: /* PPP */
  bldnil_label:
  {
    lpcreg += bldnil_sizeof;
    new_heap_nil(hreg);
    MoreMoreYesGimmeMore;
  }

  case uninumcon: /* PPP-N */
  uninumcon_label:
  {
    register Cell op1;

    op1 = opword(uninumcon_arg_1);
    if (!sreg) { /* if (flag == WRITE) */
      lpcreg += uninumcon_sizeof;
      bld_copy(hreg++, op1);
      MoreMoreYesGimmeMore;
    }
    else {  /* op2 set */
      register Cell cell = *(sreg++);
      deref2(cell, goto uninumcon_cell_is_free);
      if (cell == op1) { lpcreg += uninumcon_sizeof; }
#ifdef ATTR_VARS
      else if (isattv(cell)) {
	attv_dbgmsg(">>>> ATTV uninumcon, interrupt needed\n");
	lpcreg += uninumcon_sizeof;
	add_interrupt(cell, op1);
      }
#endif
      else Fail1;
      MoreMoreYesGimmeMore;
      
    uninumcon_cell_is_free:
      lpcreg += uninumcon_sizeof;
      bind_copy((CPtr)(cell), op1);
      MoreMoreYesGimmeMore;
    }
  }

  case bldnumcon: /* PPP-N */
  bldnumcon_label:
  {
    register Cell op1;

    op1 = opword(bldnumcon_arg_1);
    lpcreg += bldnumcon_sizeof;
    bld_copy(hreg++, op1);
    MoreMoreYesGimmeMore;
  }

  case getfloat: /* PPR-N */
  getfloat_label:
  {
    register Cell op1;
    register Cell op2;

    op1 = opreg(getfloat_arg_1);
    op2 = opword(getfloat_arg_2);
    deref2(op1, goto getfloat_op1_is_free);
    if (isfloat(op1)) {
      if (float_val(op1) == asfloat(op2)) {
	lpcreg += getfloat_sizeof;
      } else Fail1;
    }
#ifdef ATTR_VARS
    else if (isattv(op1)) {
      attv_dbgmsg(">>>> ATTV getfloat, interrupt needed\n");
      lpcreg += getfloat_sizeof;
      add_interrupt(op1, makefloat(asfloat(op2)));
    }
#endif
    else Fail1;
    MoreMoreYesGimmeMore;
    
  getfloat_op1_is_free:
    lpcreg += getfloat_sizeof;
    bind_float((CPtr)(op1), asfloat(op2));
    MoreMoreYesGimmeMore;
  }

  case putfloat: /* PPR-N */
  putfloat_label:
  {
    register CPtr op1;
    register Cell op2;

    op1 = opregaddr(putfloat_arg_1);
    op2 = opword(putfloat_arg_2);
    lpcreg += putfloat_sizeof;
    bld_float(op1, asfloat(op2));
    MoreMoreYesGimmeMore;
  }
 
  case unifloat: /* PPP-F */
  unifloat_label:
  {
    register Cell op2;

    op2 = opword(unifloat_arg_1);
    if (!sreg) { /* if (flag == WRITE) */
      lpcreg += unifloat_sizeof;
      new_heap_float(hreg, asfloat(op2));
      MoreMoreYesGimmeMore;
    }
    else {  /* op2 set */
      register Cell op1 = cell(sreg++);
      deref2(op1, goto unifloat_op1_is_free);
      if (isfloat(op1)) {
	if (float_val(op1) == asfloat(op2)) {
	  lpcreg += unifloat_sizeof;
	} else Fail1;
      }
#ifdef ATTR_VARS
      else if (isattv(op1)) {
	attv_dbgmsg(">>>> ATTV unifloat, interrupt needed\n");
	lpcreg += unifloat_sizeof;
	add_interrupt(op1, makefloat(asfloat(op2)));
      }
#endif
      else Fail1;
      MoreMoreYesGimmeMore;

    unifloat_op1_is_free:
      lpcreg += unifloat_sizeof;
      bind_float((CPtr)(op1), asfloat(op2));
      MoreMoreYesGimmeMore;
    }
  }

  case bldfloat: /* PPP-F */
  bldfloat_label:
  {
    register Cell op1;

    op1 = opword(bldfloat_arg_1);
    lpcreg += bldfloat_sizeof;
    new_heap_float(hreg, asfloat(op1));
    MoreMoreYesGimmeMore;
  }

  case unilist: /* PPP - the compression of unitvar & getlist */
  unilist_label:
  {
    register CPtr list;

    lpcreg += unilist_sizeof;
    if (!sreg) { /* if (flag == WRITE) */
      list = hreg+1;
      bld_list(hreg,list);
      hreg++;
    }
    else {
      list = sreg;
      cptr_deref2(list, { bind_list(list,hreg); sreg = NULL;
 	                  MoreMoreYesGimmeMore; });
      if (islist(list)) {
	sreg = clref_val(list);
      }
#ifdef ATTR_VARS
      else if (isattv(list)) {
	add_interrupt((Cell)list,makelist(hreg));
	sreg = NULL;
      }
#endif
      else Fail1;
    }
    MoreMoreYesGimmeMore;
  }

  case univoid: /* PPP - an optimized version of unitvar */
  univoid_label:
  {
    lpcreg += univoid_sizeof;
    if (sreg) sreg++;
    else { new_heap_free(hreg); }
    MoreMoreYesGimmeMore;
  }

  case bldvoid: /* PPP - an optimized version of bldtvar */
  bldvoid_label:
  {
    lpcreg += bldvoid_sizeof;
    new_heap_free(hreg);
    MoreMoreYesGimmeMore;
  }

  case bld_n_voids: /* PPA */
  bld_n_voids_label:
  {
    register byte i, n;

    n = opbyte(bld_n_voids_arg_1);
    lpcreg += bld_n_voids_sizeof;
    for (i = 1; i <= n; i++) { new_heap_free(hreg); }
    MoreMoreYesGimmeMore;
  }

  case putdval: /* PVR */
  putdval_label:
  {
    register Cell op1;
    register CPtr op2;

    op1 = opvar(putdval_arg_1);
    deref(op1);
    op2 = opregaddr(putdval_arg_2);
    lpcreg += putdval_sizeof;
    bld_copy(op2, op1);
    MoreMoreYesGimmeMore;
  }

  case putuval: /* PVR */
  putuval_label:
  {
    register Cell op1;
    register CPtr op2;

    op1 = opvar(putuval_arg_1);
    op2 = opregaddr(putuval_arg_2);
    lpcreg += putuval_sizeof;
    deref(op1);
    if (isnonvar(op1) || ((CPtr)op1 < hreg) || ((CPtr)op1 >= ereg)) {
      bld_copy(op2, op1);
      MoreMoreYesGimmeMore;
    } else {
      bld_ref(op2, hreg);
      bind_ref((CPtr)op1, hreg);
      new_heap_free(hreg);
      MoreMoreYesGimmeMore;
    }
  }

  case putvoid: /* PPR */
  putvoid_label:
  { /* this is the same as bldtvar --- bldtvar should probably be taken out */
    register CPtr op1;

    op1 = opregaddr(putvoid_arg_1);
    lpcreg += putvoid_sizeof;
    bld_ref(op1, hreg);
    new_heap_free(hreg);
    MoreMoreYesGimmeMore;
  }

  case getpbreg: /* PPV */
  getpbreg_label:
  {
    register CPtr op1;

    op1 = opvaraddr(getpbreg_arg_1);
    lpcreg += getpbreg_sizeof;
    bld_int(op1, ((byte *)tcpstack.high - (byte *)breg));
    MoreMoreYesGimmeMore;
  }

  case gettbreg: /* PPR */
  gettbreg_label:
  {
    register CPtr op1;

    op1 = opregaddr(gettbreg_arg_1);
    lpcreg += gettbreg_sizeof;
    bld_int(op1, ((byte *)tcpstack.high - (byte *)breg));
    MoreMoreYesGimmeMore;
  }

  case putpbreg: /* PPV */
  putpbreg_label:
  {
    register Cell op1;

    op1 = opvar(putpbreg_arg_1);
    lpcreg += putpbreg_sizeof;
    cut_code(op1);
    MoreMoreYesGimmeMore;
  }

  case puttbreg: /* PPR */
  puttbreg_label:
  {
    register Cell op1;

    op1 = opreg(puttbreg_arg_1);
    lpcreg += puttbreg_sizeof;
    cut_code(op1);
    MoreMoreYesGimmeMore;
  }

  case jumptbreg: /* PPR-L */	/* used in assert -- needed ? */
  jumptbreg_label:
  {
    register CPtr op1;

    op1 = opregaddr(jumptbreg_arg_1);
    lpcreg = oplabel(jumptbreg_arg_2);
    bld_int(op1, ((byte *)tcpstack.high - (byte *)breg));
    MoreMoreYesGimmeMore;
  }

  case getVn: /* PPV */
  getVn_label:
  {
    register CPtr op1;

    op1 = opvaraddr(getVn_arg_1);
    lpcreg += getVn_sizeof;
    cell(op1) = (Cell)tcp_subgoal_ptr(breg);
    MoreMoreYesGimmeMore;
  }

/*----------------------------------------------------------------------*\
 * Numeric/arithmetic instructions.                                     *
\*----------------------------------------------------------------------*/

  case movreg: /* PRR */
  movreg_label:
  {
    register Cell op1;
    register CPtr op2;

    op1 = opreg(movreg_arg_1);
    op2 = opregaddr(movreg_arg_2);
    lpcreg += movreg_sizeof;
    bld_copy(op2, op1);
    MoreMoreYesGimmeMore;
  }

  case negate:   /* PPR */
  negate_label:
  {
    register Cell op2;
    register CPtr op3;

    op3 = opregaddr(negate_arg_1);
    lpcreg += negate_sizeof;
    op2 = *(op3);
    deref(op2);
    if (!isinteger(op2)) { xxx_arithmetic_abort1("'\\'", op2); }
    else { bld_int(op3, ~(int_val(op2))); }
    MoreMoreYesGimmeMore; 
  }

  case and:   /* PRR */
  and_label:
  {
    register Cell op1;
    register Cell op2;
    register CPtr op3;

    op1 = opreg(and_arg_1);
    op3 = opregaddr(and_arg_2);
    lpcreg += and_sizeof;
    op2 = *(op3);
    deref(op1); 
    deref(op2);
    if (!isinteger(op1) || !isinteger(op2)) {
      xxx_arithmetic_abort(op2, "'/\\'", op1, "integers");
    }
    else { bld_int(op3, int_val(op2) & int_val(op1)); }
    MoreMoreYesGimmeMore; 
  }

  case or:   /* PRR */
  or_label:
  {
    register Cell op1;
    register Cell op2;
    register CPtr op3;

    op1 = opreg(or_arg_1);
    op3 = opregaddr(or_arg_2);
    lpcreg += or_sizeof;
    op2 = *(op3);
    deref(op1); 
    deref(op2);
    if (!isinteger(op1) || !isinteger(op2)) {
      xxx_arithmetic_abort(op2, "'\\/'", op1, "integers");
    }
    else { bld_int(op3, int_val(op2) | int_val(op1)); }
    MoreMoreYesGimmeMore; 
  }

  case logshiftl:   /* PRR */
  logshiftl_label:
  {
    register Cell op1;
    register Cell op2;
    register CPtr op3;

    op1 = opreg(logshiftl_arg_1);
    op3 = opregaddr(logshiftl_arg_2);
    lpcreg += logshiftl_sizeof;
    op2 = *(op3);
    deref(op1); 
    deref(op2);
    if (!isinteger(op1) || !isinteger(op2)) {
      xxx_arithmetic_abort(op2, "'<<'", op1, "integers");
    }
    else { bld_int(op3, int_val(op2) << int_val(op1)); }
    MoreMoreYesGimmeMore; 
  }

  case logshiftr:  /* PRR */
  logshiftr_label:
  {
    register Cell op1;
    register Cell op2;
    register CPtr op3;

    op1 = opreg(logshiftr_arg_1);
    op3 = opregaddr(logshiftr_arg_2);
    lpcreg += logshiftr_sizeof;
    op2 = *(op3);
    deref(op1); 
    deref(op2);
    if (!isinteger(op1) || !isinteger(op2)) {
      xxx_arithmetic_abort(op2, "'>>'", op1, "integers");
    }
    else { bld_int(op3, int_val(op2) >> int_val(op1)); }
    MoreMoreYesGimmeMore; 
  }

#define ARITHPROC(OP, STROP, ARG_1, ARG_2, INSTR_SIZE) \
    op1 = opreg(ARG_1);							\
    op3 = opregaddr(ARG_2);						\
    op2 = *(op3);							\
    lpcreg += INSTR_SIZE;	/* prefetching */			\
    deref(op1);								\
    deref(op2);								\
    if (isinteger(op1)) {						\
	if (isinteger(op2)) {						\
	    bld_int(op3, int_val(op2) OP int_val(op1));	}		\
	else if (isfloat(op2)) {					\
	    bld_float(op3, float_val(op2) OP (Float)int_val(op1)); }	\
	else { xxx_arithmetic_abort(op2, STROP, op1, "numbers"); }      \
    } else if (isfloat(op1)) {						\
	if (isfloat(op2)) {						\
	    bld_float(op3, float_val(op2) OP float_val(op1)); }		\
	else if (isinteger(op2)) {					\
	    bld_float(op3, (Float)int_val(op2) OP float_val(op1)); }	\
	else { xxx_arithmetic_abort(op2, STROP, op1, "numbers"); }      \
    } else { xxx_arithmetic_abort(op2, STROP, op1, "numbers"); }

  case addreg: /* PRR */
  addreg_label:
  {
    register Cell op1;
    register Cell op2;
    register CPtr op3;

    ARITHPROC(+, "+", addreg_arg_1, addreg_arg_2, addreg_sizeof);
    MoreMoreYesGimmeMore; 
  }

  case subreg: /* PRR */
  subreg_label:
  {
    register Cell op1;
    register Cell op2;
    register CPtr op3;

    op1 = opreg(subreg_arg_1);
    op3 = opregaddr(subreg_arg_2);
    op2 = *(op3);
    lpcreg += subreg_sizeof;
    deref(op1);
    deref(op2);
    if (isinteger(op1)) {						
      if (isinteger(op2)) {
	bld_int(op3, int_val(op2) - int_val(op1)); }
      else if (isfloat(op2)) {
	bld_float(op3, float_val(op2) - (Float)int_val(op1)); }
      else { xxx_arithmetic_abort(op2, "-", op1, "numbers"); }
    } else if (isfloat(op1)) {
      if (isfloat(op2)) {
	bld_float(op3, float_val(op2) - float_val(op1)); }
      else if (isinteger(op2)) {
	bld_float(op3, (Float)int_val(op2) - float_val(op1)); }
      else xxx_arithmetic_abort(op2, "-", op1, "numbers");
    }
    else xxx_arithmetic_abort(op2, "-", op1, "numbers");
    MoreMoreYesGimmeMore; 
  }

  case mulreg: /* PRR */
  mulreg_label:
  {
    register Cell op1;
    register Cell op2;
    register CPtr op3;

    ARITHPROC(*, "*", mulreg_arg_1, mulreg_arg_2, mulreg_sizeof);
    MoreMoreYesGimmeMore; 
  }

  case divreg: /* PRR */
  divreg_label:
  {
    register Cell op1;
    register Cell op2;
    register CPtr op3;

    op1 = opreg(divreg_arg_1);
    op3 = opregaddr(divreg_arg_2);
    op2 = *(op3);
    lpcreg += divreg_sizeof;   
    deref(op1);
    deref(op2);
    if (isinteger(op1)) {
      if (isinteger(op2)) {
	bld_float(op3, (Float)int_val(op2)/(Float)int_val(op1)); }
      else if (isfloat(op2)) {
	bld_float(op3, float_val(op2)/(Float)int_val(op1)); }
      else { xxx_arithmetic_abort(op2, "/", op1, "numbers"); }
    } else if (isfloat(op1)) {
      if (isfloat(op2)) {
	bld_float(op3, float_val(op2)/float_val(op1)); }
      else if (isinteger(op2)) {
	bld_float(op3, (Float)int_val(op2)/float_val(op1)); }
      else { xxx_arithmetic_abort(op2, "/", op1, "numbers"); }
    } else { xxx_arithmetic_abort(op2, "/", op1, "numbers"); }
    MoreMoreYesGimmeMore; 
  }

  case idivreg: /* PRR */
  idivreg_label:
  {
    register Cell op1;
    register Cell op2;
    register CPtr op3;

    op1 = opreg(idivreg_arg_1);
    op3 = opregaddr(idivreg_arg_2);
    op2 = *(op3);
    lpcreg += idivreg_sizeof;   
    deref(op1);
    deref(op2);
    if (isinteger(op1) && isinteger(op2)) {
      if (op1 != makeint(0)) { bld_int(op3, int_val(op2) / int_val(op1)); }
      else {
	xxx_err_handle(ZERO_DIVIDE_ERROR, 2,
		       "arithmetic expression involving is/2 or eval/0",
		       2, "non-zero integer", 0);
	lpcreg = pcreg;
      }
    }
    else { xxx_arithmetic_abort(op2, "//", op1, "integers"); }
    MoreMoreYesGimmeMore; 
  }

  case remreg: /* PRR */
  remreg_label:
  {
    register Cell op1;
    register Cell op2;
    register CPtr op3;

    op1 = opreg(remreg_arg_1);
    op3 = opregaddr(remreg_arg_2);
    op2 = *(op3);
    lpcreg += remreg_sizeof;   
    deref(op1);
    deref(op2);
    if (isinteger(op1) && isinteger(op2)) {
      if (op1 != makeint(0)) { bld_int(op3, int_val(op2) % int_val(op1)); }
      else {
	xxx_err_handle(DOMAIN_ERROR, 2, "arithmetic expression involving rem",
		       2, "non-zero integer", 0);
	lpcreg = pcreg;
      }
    }
    else { xxx_arithmetic_abort(op2, "rem", op1, "integers"); }
    MoreMoreYesGimmeMore;
  }

  case modreg: /* PRR */
  modreg_label:
  {
    register Cell op1;
    register Cell op2;
    register CPtr op3;
    
    op1 = opreg(modreg_arg_1);
    op3 = opregaddr(modreg_arg_2);
    op2 = *(op3);
    lpcreg += modreg_sizeof;   
    deref(op1);
    deref(op2);
    if (isinteger(op1) && isinteger(op2)) {
      int x = int_val(op2);
      int y = int_val(op1);
      if (y != 0) { x = x - ((int)floor((double)x/y) * y); bld_int(op3, x); }
      else {
	xxx_err_handle(DOMAIN_ERROR, 2, "arithmetic expression involving mod",
		       2, "non-zero integer", 0);
	lpcreg = pcreg;
      }
    }
    else { xxx_arithmetic_abort(op2, "mod", op1, "integers"); }
    MoreMoreYesGimmeMore;
  }

  case add_byte: /* P-A-R */
  add_byte_label:
  {
    register byte op1;
    register Cell op2;
    register CPtr op3;

    op1 = opbyte(add_byte_arg_1);
    op3 = opregaddr(add_byte_arg_2);
    op2 = *(op3);
    lpcreg += add_byte_sizeof;   
    deref(op2);
    if (isinteger(op2)) {
      bld_int(op3, int_val(op2) + op1);
      MoreMoreYesGimmeMore;
    }
    else if (isfloat(op2)) {
      bld_float(op3, float_val(op2) + op1);
      MoreMoreYesGimmeMore;
    }
    else {
      xxx_arithmetic_abort(op2, "+", makeint(op1), "numbers");
      MoreMoreYesGimmeMore;
    }
  }

  case sub_byte: /* P-A-R */
  sub_byte_label:
  {
    register byte op1;
    register Cell op2;
    register CPtr op3;

    op1 = opbyte(sub_byte_arg_1);
    op3 = opregaddr(sub_byte_arg_2);
    op2 = *(op3);
    lpcreg += sub_byte_sizeof;   
    deref(op2);
    if (isinteger(op2)) {
      bld_int(op3, int_val(op2) - op1); 
      MoreMoreYesGimmeMore;
    }
    else if (isfloat(op2)) {
      bld_float(op3, float_val(op2) - op1);
      MoreMoreYesGimmeMore;
    }
    else {
      xxx_arithmetic_abort(op2, "-", makeint(op1), "numbers");
      MoreMoreYesGimmeMore;
    }
  }

/*----------------------------------------------------------------------*\
 * Instructions produced through compression in peephole optimization.  *
\*----------------------------------------------------------------------*/

  case uni_tvar_tvar: /* PRR */
  uni_tvar_tvar_label:
  {
    register CPtr op1, op2;

    op1 = opregaddr(uni_tvar_tvar_arg_1);
    op2 = opregaddr(uni_tvar_tvar_arg_2);
    lpcreg += uni_tvar_tvar_sizeof;
    if (!sreg) { /* if (flag == WRITE) */
      bld_ref(op1, hreg);
      new_heap_free(hreg);
      bld_ref(op2, hreg);
      new_heap_free(hreg);
    }
    else {
      bld_copy(op1, *(sreg));
      bld_copy(op2, *(sreg+1));
      sreg += 2;
    }
    MoreMoreYesGimmeMore;
  }

  case uni_tvar_tval: /* PRR */
  uni_tvar_tval_label:
  {
    register CPtr op1;
    register Cell op2;

    op1 = opregaddr(uni_tvar_tval_arg_1);
    if (!sreg) { /* if (flag == WRITE) */
      bld_ref(op1, hreg);
      new_heap_free(hreg);
      op2 = opreg(uni_tvar_tval_arg_2);
      lpcreg += uni_tvar_tval_sizeof;
      bld_val(op2); 
    }
    else {
      bld_copy(op1, *sreg);
      op2 = opreg(uni_tvar_tval_arg_2);
      if (! unify(op2, *(sreg+1))) Fail1;
      else { lpcreg += uni_tvar_tval_sizeof; sreg += 2; }
    }
    MoreMoreYesGimmeMore;
  }

  case uni_tval_tvar: /* PRR */
  uni_tval_tvar_label:
  {
    register Cell op1;
    register CPtr op2;

    op1 = opreg(uni_tval_tvar_arg_1);
    if (!sreg) { /* if (flag == WRITE) */
      bld_val(op1);
      op2 = opregaddr(uni_tval_tvar_arg_2);
      lpcreg += uni_tval_tvar_sizeof;
      bld_ref(op2, hreg);
      new_heap_free(hreg);
    }
    else {
      if (! unify(op1, *sreg)) Fail1;
      else {
	op2 = opregaddr(uni_tval_tvar_arg_2);
	lpcreg += uni_tval_tvar_sizeof;
	bld_copy(op2, *(sreg+1));
	sreg += 2;
      }
    }
    MoreMoreYesGimmeMore; 
  }

  case uni_tval_tval: /* PRR */
  uni_tval_tval_label:
  {
    register Cell op12; /* same reg used for both arguments */

    op12 = opreg(uni_tval_tval_arg_1);
    if (!sreg) { /* if (flag == WRITE) */
      bld_val(op12); 
      op12 = opreg(uni_tval_tval_arg_2);
      lpcreg += uni_tval_tval_sizeof;
      bld_val(op12); 
    }
    else {
      if (! unify(op12, *sreg)) Fail1;
      else {
	op12 = opreg(uni_tval_tval_arg_2);
	if (! unify(op12, *(sreg+1))) Fail1;
	else { lpcreg += uni_tval_tval_sizeof; sreg += 2; }
      }
    } 
    MoreMoreYesGimmeMore; 
  }

  case uni_n_voids: /* PPA */
  uni_n_voids_label:
  {
    register byte n;

    n = opbyte(uni_n_voids_arg_1);
    lpcreg += uni_n_voids_sizeof;
    if (sreg) sreg += n;
    else {
      register byte i;
      for (i = 1; i <= n; i++) { new_heap_free(hreg); }
    }
    MoreMoreYesGimmeMore;
  }

  case uni_void_tvar: /* PPR */
  uni_void_tvar_label:
  {
    register CPtr op1;

    op1 = opregaddr(uni_void_tvar_arg_1);
    lpcreg += uni_void_tvar_sizeof;
    if (!sreg) { /* if (flag == WRITE) */
      new_heap_free(hreg);
      bld_ref(op1, hreg);
      new_heap_free(hreg);
    }
    else { bld_copy(op1, *(sreg+1)); sreg += 2; }
    MoreMoreYesGimmeMore;
  }

  case uni_tvar_void: /* PPR */
  uni_tvar_void_label:
  {
    register CPtr op1;

    op1 = opregaddr(uni_tvar_void_arg_1);
    lpcreg += uni_tvar_void_sizeof;
    if (!sreg) { /* if (flag == WRITE) */
      bld_ref(op1, hreg);
      new_heap_free(hreg);
      new_heap_free(hreg);
    }
    else { bld_copy(op1, *(sreg)); sreg += 2; }
    MoreMoreYesGimmeMore;
  }

  case getlist_tval_tvar: /* RRR */
  getlist_tval_tvar_label:
  {
    register Cell op1;
    register CPtr op3;

    op1 = opreg(getlist_tval_tvar_arg_1);
    deref2(op1, goto getlist_tval_tvar_op1ref);
    if (islist(op1)) {
      sreg = clref_val(op1);
      op1 = opreg(getlist_tval_tvar_arg_2);
      if (! unify(op1, *sreg)) Fail1;
      else {
	sreg++;
	op3 = opregaddr(getlist_tval_tvar_arg_3);
	lpcreg += getlist_tval_tvar_sizeof;
	bld_ref(op3, *sreg);
      }
    } else if (isref(op1)) {
    getlist_tval_tvar_op1ref:
      bind_list((CPtr)(op1), hreg);
      op1 = opreg(getlist_tval_tvar_arg_2);
      bld_val(op1);
      op3 = opregaddr(getlist_tval_tvar_arg_3);
      lpcreg += getlist_tval_tvar_sizeof;
      bld_ref(op3, hreg);
      new_heap_free(hreg);
    } else Fail1;
    MoreMoreYesGimmeMore;
  }

  case getlist_tvar_tvar: /* RRR */
  getlist_tvar_tvar_label:
  {
    register Cell op1;
    register CPtr op2;

    op1 = opreg(getlist_tvar_tvar_arg_1);
    deref2(op1, goto getlist_tvar_tvar_op1ref);
    if (islist(op1)) {
      sreg = clref_val(op1);
      op2 = opregaddr(getlist_tvar_tvar_arg_2);
      bld_ref(op2, *(sreg));
      op2 = opregaddr(getlist_tvar_tvar_arg_3);
      lpcreg += getlist_tvar_tvar_sizeof;
      bld_ref(op2, *(sreg+1));
    } else if (isref(op1)) {
    getlist_tvar_tvar_op1ref:
      bind_list((CPtr)(op1), hreg);
      op2 = opregaddr(getlist_tvar_tvar_arg_2);
      bld_ref(op2, hreg);
      new_heap_free(hreg);
      op2 = opregaddr(getlist_tvar_tvar_arg_3);
      lpcreg += getlist_tvar_tvar_sizeof;
      bld_ref(op2, hreg);
      new_heap_free(hreg);
    } 
#ifdef ATTR_VARS
      else if (isattv(op1)) {
        fprintf(stderr, ">>>> ATTV getlist_tvar_tvar, interrupt needed\n");
        /* attv_dbgmsg(">>>> ATTV getlist_tvar_tvar, interrupt needed\n"); */
	add_interrupt(op1, makelist(hreg));
	op2 = opregaddr(getlist_tvar_tvar_arg_2);
        bld_ref(op2, hreg);
        new_heap_free(hreg);
        op2 = opregaddr(getlist_tvar_tvar_arg_3);
	lpcreg += getlist_tvar_tvar_sizeof;
        bld_ref(op2, hreg);
        new_heap_free(hreg);
    }
#endif
    else Fail1;
    MoreMoreYesGimmeMore;
  }

  case getlist_void_tvar: /* PRR */
  getlist_void_tvar_label:
  {
    register Cell op1;
    register CPtr op2;

    op1 = opreg(getlist_void_tvar_arg_1);
    deref2(op1, goto getlist_void_tvar_op1ref);
    if (islist(op1)) {
      sreg = clref_val(op1);
      op2 = opregaddr(getlist_void_tvar_arg_2);
      lpcreg += getlist_void_tvar_sizeof;
      bld_ref(op2, *(sreg+1));
    } else if (isref(op1)) {
    getlist_void_tvar_op1ref:
      bind_list((CPtr)(op1), hreg);
      new_heap_free(hreg);
      op2 = opregaddr(getlist_void_tvar_arg_2);
      lpcreg += getlist_void_tvar_sizeof;
      bld_ref(op2, hreg);
      new_heap_free(hreg);
    } else Fail1;
    MoreMoreYesGimmeMore;
  }

  case getlist_tvar_void: /* PRR */
  getlist_tvar_void_label:
  {
    register Cell op1;
    register CPtr op2;

    op1 = opreg(getlist_tvar_void_arg_1);
    deref2(op1, goto getlist_tvar_void_op1ref);
    if (islist(op1)) {
      sreg = clref_val(op1);
      op2 = opregaddr(getlist_tvar_void_arg_2);
      lpcreg += getlist_tvar_void_sizeof;
      bld_ref(op2, *(sreg));
    } else if (isref(op1)) {
    getlist_tvar_void_op1ref:
      bind_list((CPtr)(op1), hreg);
      op2 = opregaddr(getlist_tvar_void_arg_2);
      lpcreg += getlist_tvar_void_sizeof;
      bld_ref(op2, hreg);
      new_heap_free(hreg);
      new_heap_free(hreg);
    } else Fail1;
    MoreMoreYesGimmeMore;
  }

  case getlist_tval_void: /* PRR */
  getlist_tval_void_label:
  {
    register Cell op1;

    op1 = opreg(getlist_tval_void_arg_1);
    deref2(op1, goto getlist_tval_void_op1ref);
    if (islist(op1)) {
      sreg = clref_val(op1);
      op1 = opreg(getlist_tval_void_arg_2);
      if (! unify(op1, *sreg)) Fail1;
      else { lpcreg += getlist_tval_void_sizeof; sreg++; }
    } else if (isref(op1)) {
    getlist_tval_void_op1ref:
      bind_list((CPtr)(op1), hreg);
      op1 = opreg(getlist_tval_void_arg_2);
      lpcreg += getlist_tval_void_sizeof;
      bld_val(op1);
      new_heap_free(hreg);
    } else Fail1;
    MoreMoreYesGimmeMore;
  }

  case getlist_tvar_pvar: /* RRV */
  getlist_tvar_pvar_label:
  {
    register Cell op1;
    register CPtr op2;

    op1 = opreg(getlist_tvar_pvar_arg_1);
    deref2(op1, goto getlist_tvar_pvar_op1ref);
    if (islist(op1)) {
      sreg = clref_val(op1);
      op2 = opregaddr(getlist_tvar_pvar_arg_2);
      bld_ref(op2, *(sreg));
      op2 = opvaraddr(getlist_tvar_pvar_arg_3);
      lpcreg += getlist_tvar_pvar_sizeof;
      bld_copy(op2, *(sreg+1));
    } else if (isref(op1)) {
    getlist_tvar_pvar_op1ref:
      bind_list((CPtr)(op1), hreg);
      op2 = opregaddr(getlist_tvar_pvar_arg_2);
      bld_ref(op2, hreg);
      new_heap_free(hreg);
      op2 = opvaraddr(getlist_tvar_pvar_arg_3);
      lpcreg += getlist_tvar_pvar_sizeof;
      bld_ref(op2, hreg);
      new_heap_free(hreg);
    } else Fail1;
    MoreMoreYesGimmeMore;
  }

  case getstr_tvar: /* PRR-S */
  getstr_tvar_label:
  {
    register Cell op1;
    register Cell strct1;
    register CPtr op_areg, p;

    op1 = opreg(getstr_tvar_arg_1);
    strct1 = opword(getstr_tvar_arg_3);
    deref2(op1, goto getstr_tvar_is_free);
    if (isstruct(op1)) {
      op1 = (Cell)(struct_val(op1));
      if (*((Psc *)op1) == (Psc)strct1) {
	op_areg = opregaddr(getstr_tvar_arg_2);
	lpcreg += getstr_tvar_sizeof;
	bld_copy(op_areg, *((CPtr)op1+1));
	sreg = (CPtr)op1 + 2;
      }
      else Fail1;
    }
#ifdef ATTR_VARS
    /* needs modification - currently does not work */
    else if (isattv(op1)) {
      attv_dbgmsg(">>>> ATTV getstr_tvar, interrupt needed\n");
      lpcreg += getstr_tvar_sizeof;
      add_interrupt(op1, makestruct(hreg));
      new_heap_functor(hreg, (Psc)strct1);
      sreg = NULL;
    }
#endif
    else Fail1;
    MoreMoreYesGimmeMore;

  getstr_tvar_is_free:
    bind_struct((CPtr)(op1), (Pair)hreg);
    p = hreg;
    hreg = p+2;
    bld_functor(p, (Psc)strct1);
    op_areg = opregaddr(getstr_tvar_arg_2);
    bld_ref(op_areg, (++p));
    bld_free(p);
    lpcreg += getstr_tvar_sizeof;
    sreg = NULL;
    MoreMoreYesGimmeMore;
  }

  case getstr_tvar_tvar: /* RRR-S */
  getstr_tvar_tvar_label:
  {
    register Cell op1;
    register Cell strct2;
    register CPtr op_areg, p;

    op1 = opreg(getstr_tvar_tvar_arg_1);
    strct2 = opword(getstr_tvar_tvar_arg_4);
    deref2(op1, goto getstr_tvar_tvar_is_free);
    if (isstruct(op1)) {
      /* fprintf(stderr, "In getstr_tvar_tvar bound case\n"); */
      op1 = (Cell)(struct_val(op1));
      if (*((Psc *)op1) == (Psc)strct2) {
	op_areg = opregaddr(getstr_tvar_tvar_arg_2);
	bld_copy(op_areg, *((CPtr)op1+1));
	op_areg = opregaddr(getstr_tvar_tvar_arg_3);
	lpcreg += getstr_tvar_tvar_sizeof;
	bld_copy(op_areg, *((CPtr)op1+2));
	sreg = (CPtr)op1 + 3;
      }
      else Fail1;
    }
#ifdef ATTR_VARS
    /* needs modification - currently does not work */
    else if (isattv(op1)) {
      attv_dbgmsg(">>>> ATTV getstr_tvar_tvar, interrupt needed\n");
      lpcreg += getstr_tvar_tvar_sizeof;
      add_interrupt(op1, makestruct(hreg));
      new_heap_functor(hreg, (Psc)strct2);
      sreg = NULL;
    }
#endif
    else Fail1;
    MoreMoreYesGimmeMore;

  getstr_tvar_tvar_is_free:
    /* fprintf(stderr, "In getstr_tvar_tvar_is_free case\n"); */
    bind_struct((CPtr)(op1), (Pair)hreg);
    p = hreg;
    hreg = p+3;
    bld_functor(p, (Psc)strct2);
    op_areg = opregaddr(getstr_tvar_tvar_arg_2);
    bld_ref(op_areg, (++p));
    bld_free(p);
    op_areg = opregaddr(getstr_tvar_tvar_arg_3);
    bld_ref(op_areg, (++p));
    bld_free(p);
    lpcreg += getstr_tvar_tvar_sizeof;
    sreg = NULL;
    MoreMoreYesGimmeMore;
  }

  case getstr_tvar_tvar_tvar: /* RRR-S */
  getstr_tvar_tvar_tvar_label:
  { /* used when the functor and the first argument are mapped to the
     * *SAME* A-register */
    register Cell op1;
    register Cell strct3;
    register CPtr op_areg, p;

    op1 = opreg(getstr_tvar_tvar_tvar_arg_1);
    strct3 = opword(getstr_tvar_tvar_tvar_arg_4);
    deref2(op1, goto getstr_tvar_tvar_tvar_is_free);
    if (isstruct(op1)) {
      /* fprintf(stderr, "In getstr_tvar_tvar_tvar bound case\n"); */
      op1 = (Cell)(struct_val(op1));
      if (*((Psc *)op1) == (Psc)strct3) {
	op_areg = opregaddr(getstr_tvar_tvar_tvar_arg_1);
	bld_copy(op_areg, *((CPtr)op1+1));
	op_areg = opregaddr(getstr_tvar_tvar_tvar_arg_2);
	bld_copy(op_areg, *((CPtr)op1+2));
	op_areg = opregaddr(getstr_tvar_tvar_tvar_arg_3);
	lpcreg += getstr_tvar_tvar_tvar_sizeof;
	bld_copy(op_areg, *((CPtr)op1+3));
	sreg = (CPtr)op1 + 4;
      }
      else Fail1;
    }
#ifdef ATTR_VARS
    /* needs modification - currently does not work */
    else if (isattv(op1)) {
      attv_dbgmsg(">>>> ATTV getstr_tvar_tvar_tvar, interrupt needed\n");
      lpcreg += getstr_tvar_tvar_tvar_sizeof;
      add_interrupt(op1, makestruct(hreg));
      new_heap_functor(hreg, (Psc)strct3);
      sreg = NULL;
    }
#endif
    else Fail1;
    MoreMoreYesGimmeMore;

  getstr_tvar_tvar_tvar_is_free:
    /* fprintf(stderr, "In getstr_tvar_tvar_tvar_is_free case\n"); */
    bind_struct((CPtr)(op1), (Pair)hreg);
    p = hreg;
    hreg = p+4;
    bld_functor(p, (Psc)strct3);
    op_areg = opregaddr(getstr_tvar_tvar_tvar_arg_1);
    bld_ref(op_areg, (++p));
    bld_free(p);
    op_areg = opregaddr(getstr_tvar_tvar_tvar_arg_2);
    bld_ref(op_areg, (++p));
    bld_free(p);
    op_areg = opregaddr(getstr_tvar_tvar_tvar_arg_3);
    bld_ref(op_areg, (++p));
    bld_free(p);
    lpcreg += getstr_tvar_tvar_tvar_sizeof;
    sreg = NULL;
    MoreMoreYesGimmeMore;
  }

/*----------------------------------------------------------------------*\
 * Branching instructions.                                              *
\*----------------------------------------------------------------------*/

  case jump:   /* PPP-L */
  jump_label:
  {
    lpcreg = oplabel(jump_arg_1);
    MoreMoreYesGimmeMore;
  }

  case jumpz:   /* PPR-L */
  jumpz_label:
  {
    register Cell op1;

    op1 = opreg(jumpz_arg_1);
    if (int_val(op1) == 0)
      lpcreg = oplabel(jumpz_arg_2);
    else lpcreg += jumpz_sizeof;
    MoreMoreYesGimmeMore;
  }

  case jumpnz:    /* PPR-L */
  jumpnz_label:
  {
    register Cell op1;

    op1 = opreg(jumpnz_arg_1);
    if (int_val(op1) != 0)
      lpcreg = oplabel(jumpnz_arg_2);
    else lpcreg += jumpnz_sizeof;
    MoreMoreYesGimmeMore;
  }

  case jumplt:    /* PPR-L */
  jumplt_label:
  {
    register Cell op1;

    op1 = opreg(jumplt_arg_1);
    if ((isinteger(op1) && int_val(op1) < 0) ||
	(isfloat(op1) && float_val(op1) < 0.0))
      lpcreg = oplabel(jumplt_arg_2);
    else lpcreg += jumplt_sizeof;
    MoreMoreYesGimmeMore; 
  }

  case jumple:    /* PPR-L */
  jumple_label:
  {
    register Cell op1;

    op1 = opreg(jumple_arg_1);
    if ((isinteger(op1) && int_val(op1) <= 0) ||
	(isfloat(op1) && float_val(op1) <= 0.0))
      lpcreg = oplabel(jumple_arg_2);
    else lpcreg += jumple_sizeof;
    MoreMoreYesGimmeMore; 
  }

  case jumpgt:    /* PPR-L */
  jumpgt_label:
  {
    register Cell op1;

    op1 = opreg(jumpgt_arg_1);
    if ((isinteger(op1) && int_val(op1) > 0) ||
	(isfloat(op1) && float_val(op1) > 0.0))
      lpcreg = oplabel(jumpgt_arg_2);
    else lpcreg += jumpgt_sizeof;
    MoreMoreYesGimmeMore;
  }

  case jumpge:    /* PPR-L */
  jumpge_label:
  {
    register Cell op1;

    op1 = opreg(jumpge_arg_1);
    if ((isinteger(op1) && int_val(op1) >= 0) ||
	(isfloat(op1) && float_val(op1) >= 0.0))
      lpcreg = oplabel(jumpge_arg_2);
    else lpcreg += jumpge_sizeof;
    MoreMoreYesGimmeMore; 
  }

/*----------------------------------------------------------------------*\
 * Procedure instructions.                                              *
\*----------------------------------------------------------------------*/

  case allocate_gc: /* PAA */
  allocate_gc_label:
  {
    register Cell op1;
    register byte op2, op3;

    op2 = opbyte(allocate_gc_arg_1);
    op3 = opbyte(allocate_gc_arg_2);
    lpcreg += allocate_gc_sizeof; /* prefetching */
    if (ereg_on_top(ereg)) 
      op1 = (Cell)(ereg - *(cpreg-call_sizeof+call_arg_1));
    else op1 = (Cell)(ebreg-1);
    *(CPtr *)((CPtr) op1) = ereg;
    *((byte **) (CPtr)op1-1) = cpreg;
    ereg = (CPtr)op1; 
    {/* initialize all permanent variables not in the first chunk to unbound */
      register CPtr p = ((CPtr)op1) - op2;
      op3 -= op2;
      while (op3--) {
	bld_free(p);
        p--;
      }
    }
    MoreMoreYesGimmeMore;
  }

  case call: /* PPA-S */
  call_label:
  {
    register Psc psc;

    /* the first arg is used later by alloc */
    psc = (Psc)opword(call_arg_2);
    lpcreg += call_sizeof;
    cpreg = lpcreg;
    call_sub(psc);
    MoreMoreYesGimmeMore;
  }

  case calld:   /* PPA-L */
  calld_label:
  {
    cpreg = lpcreg + calld_sizeof;
    lpcreg = oplabel(calld_arg_2);
    MoreMoreYesGimmeMore;
  }

  case execute:  /* PPP-S */
  execute_label:
  {
    register Psc psc;

    psc = (Psc)opword(execute_arg_1);
    lpcreg += execute_sizeof;
    call_sub(psc);
    MoreMoreYesGimmeMore;
  }

  case proceed:  /* PPP */
  proceed_label:
  {
    lpcreg = cpreg;
    MoreMoreYesGimmeMore;
  }

  case deallocate: /* PPP */
  deallocate_label:
  {
    lpcreg += deallocate_sizeof;
    cpreg = *((byte **)ereg-1);
    ereg = *(CPtr *)ereg;
    MoreMoreYesGimmeMore;
  }

  case dealloc_executed: /* PPP-L */
  dealloc_executed_label:
  {
    lpcreg = oplabel(dealloc_executed_arg_1);
    cpreg = *((byte **)ereg-1);
    ereg = *(CPtr *)ereg;
    MoreMoreYesGimmeMore;
  }

  case dealloc_execute:  /* PPP-S */
  dealloc_execute_label:
  {
    register Psc psc;

    psc = (Psc)opword(dealloc_execute_arg_1);
    lpcreg += dealloc_execute_sizeof;
    cpreg = *((byte **)ereg-1);
    ereg = *(CPtr *)ereg;
    call_sub(psc);
    MoreMoreYesGimmeMore;
  }

  case dealloc_proceed: /* PPP */
  dealloc_proceed_label:
  {
    lpcreg = *((byte **)ereg-1);
    ereg = *(CPtr *)ereg;
    MoreMoreYesGimmeMore;
  }

/*----------------------------------------------------------------------*\
 * Non-determinism instructions.                                        *
\*----------------------------------------------------------------------*/

  case trymeelse: /* PPA-L */
  trymeelse_label:
  {
    register byte op1;
    register Cell op2;

    op1 = opbyte(trymeelse_arg_1);
    op2 = opword(trymeelse_arg_2);
    lpcreg += trymeelse_sizeof;
    subtryme(op1, op2);
    MoreMoreYesGimmeMore;
  }

  case retrymeelse: /* PPA-L */
  retrymeelse_label:
  {
    register Cell op1;

    op1 = opbyte(retrymeelse_arg_1);
    cp_pcreg(breg) = oplabel(retrymeelse_arg_2);
    lpcreg += retrymeelse_sizeof;
    restore_sub(op1);
    MoreMoreYesGimmeMore;
  }

  case trustmeelsefail: /* PPA */
  trustmeelsefail_label:
  {
    register Cell op1;

    op1 = opbyte(trustmeelsefail_arg_1);
    lpcreg += trustmeelsefail_sizeof;
    trust_restore_sub(op1);
    MoreMoreYesGimmeMore;
  }

  case try: /* PPA-L */
  try_label:
  {
    register byte op1;
    register Cell op2;

    op1 = opbyte(try_arg_1);
    op2 = (Cell)lpcreg + try_sizeof;
    lpcreg = oplabel(try_arg_2);
    subtryme(op1, op2);
    MoreMoreYesGimmeMore;
  }

  case retry: /* PPA-L */
  retry_label:
  {
    register Cell op1;

    op1 = opbyte(retry_arg_1);
    cp_pcreg(breg) = lpcreg + retry_sizeof;
    lpcreg = oplabel(retry_arg_2);
    restore_sub(op1);
    MoreMoreYesGimmeMore;
  }

  case trust: /* PPA-L */
  trust_label:
  {
    register Cell op1;

    op1 = opbyte(trust_arg_1);
    lpcreg = oplabel(trust_arg_2);
    trust_restore_sub(op1);
    MoreMoreYesGimmeMore;
  }

  case trymeorelse: /* PPA-L */
  trymeorelse_label:
  {
    register byte op1;
    register Cell op2;

    op1 = 0;
    op2 = (Cell)oplabel(trymeorelse_arg_2);
    lpcreg += trymeorelse_sizeof;
    cpreg = lpcreg;
    subtryme(op1, op2);
    MoreMoreYesGimmeMore;
  }

  case retrymeorelse: /* PPA-L */
  retrymeorelse_label:
  {
    register Cell op1;

    op1 = 0;
    cp_pcreg(breg) = oplabel(retrymeorelse_arg_2);
    lpcreg += retrymeorelse_sizeof;
    cpreg = lpcreg;
    restore_sub(op1);
    MoreMoreYesGimmeMore;
  }

  case trustmeorelsefail: /* PPA */
  trustmeorelsefail_label:
  {
    register Cell op1;

    op1 = 0;
    lpcreg += trustmeorelsefail_sizeof;
    cpreg = lpcreg;
    trust_restore_sub(op1);
    MoreMoreYesGimmeMore;
  }

/*----------------------------------------------------------------------*\
 * Indexing instructions.                                               *
\*----------------------------------------------------------------------*/

  case switchonterm: /* PPR-L-L */
  switchonterm_label:
  {
    register Cell op1;

    op1 = opreg(switchonterm_arg_1);
    deref2(op1, { lpcreg += switchonterm_sizeof; MoreMoreYesGimmeMore; });
    switch (cell_tag(op1)) {
    case ATTV:
      lpcreg += switchonterm_sizeof;
      break;
    case INT:
    case ATM:
    case FLOAT:
      lpcreg = oplabel(switchonterm_arg_2);
      break;
    case STRUCT:
    case LIST:
      lpcreg = oplabel(switchonterm_arg_3);
      break;
    }
    MoreMoreYesGimmeMore;
  }

  case switchonlist: /* PPR-L-L */
  switchonlist_label:
  {
    register Cell op1;

    op1 = opreg(switchonlist_arg_1);
    /* there is a performance reason for this code placement; do NOT modify */
    if (islist(op1))
           { lpcreg = oplabel(switchonlist_arg_3); MoreMoreYesGimmeMore; }
    else {
      deref2(op1, { lpcreg += switchonlist_sizeof; MoreMoreYesGimmeMore; });
      if (islist(op1))
	   { lpcreg = oplabel(switchonlist_arg_3); MoreMoreYesGimmeMore; }
      else { lpcreg = oplabel(switchonlist_arg_2); MoreMoreYesGimmeMore; }
    }
  }

  case hash_on_bound: /* PPR-L-UN */
  hash_on_bound_label:
  { /* op1 is register, op2 is hash table offset, op3 is modulus */
    register Cell op1;
    register byte *op2;
    register Cell op3;

    op1 = opreg(hash_on_bound_arg_1);
    deref2(op1, { lpcreg += hash_on_bound_sizeof; MoreMoreYesGimmeMore; });
    if (isstruct(op1)) op1 = (Cell)get_str_psc(op1);
    else
      if (islist(op1)) op1 = (Cell)list_str;
#if (defined(ATTR_VARS))
      else
	if (isattv(op1)) { lpcreg += hash_on_bound_sizeof; MoreMoreYesGimmeMore;}
#endif
    op3 = opword(hash_on_bound_arg_3);
    op1 = ihash(op1, op3);
    op2 = oplabel(hash_on_bound_arg_2);
    lpcreg = *(byte **)(op2 + op1 * sizeof(Cell));
    MoreMoreYesGimmeMore;
  }
   
  case hash_on_atomic: /* PPR-L-UN */
  hash_on_atomic_label:
  { /* op1 is register, op2 is hash table offset, op3 is modulus */
    register Cell op1;
    register byte *op2;
    register Cell op3;

    op1 = opreg(hash_on_atomic_arg_1);
    deref2(op1, { lpcreg += hash_on_atomic_sizeof; MoreMoreYesGimmeMore; });
#if (defined(ATTR_VARS))
    if (isattv(op1)) { lpcreg += hash_on_atomic_sizeof; MoreMoreYesGimmeMore; }
#endif
    op3 = opword(hash_on_atomic_arg_3);
    op1 = ihash(op1, op3);
    op2 = oplabel(hash_on_atomic_arg_2);
    lpcreg = *(byte **)(op2 + op1 * sizeof(Cell));
    MoreMoreYesGimmeMore;
  }

  case hash_on_struct: /* PPR-L-UN */
  hash_on_struct_label:
  { /* op1 is register, op2 is hash table offset, op3 is modulus */
    register Cell op1;
    register byte *op2;
    register Cell op3;

    op1 = opreg(hash_on_struct_arg_1);
    deref2(op1, { lpcreg += hash_on_struct_sizeof; MoreMoreYesGimmeMore; });
    if (isstruct(op1)) { op1 = (Cell)get_str_psc(op1); }
    else { lpcreg += hash_on_struct_sizeof; MoreMoreYesGimmeMore; }
    op3 = opword(hash_on_struct_arg_3);
    op1 = ihash(op1, op3);
    op2 = oplabel(hash_on_struct_arg_2);
    lpcreg = *(byte **)(op2 + op1 * sizeof(Cell));
    MoreMoreYesGimmeMore;
  }

/*----------------------------------------------------------------------*\
 * Miscellaneous instructions.                                          *
\*----------------------------------------------------------------------*/

  case test_heap: /* PPA-UN */
  test_heap_label:
  {
    register Integer margin;

    margin = opword(test_heap_arg_2);
    if (((ereg - hreg) < margin)
#ifdef GC_TEST
	|| (infcounter++ > GC_INFERENCES)
#endif
	                                 )
      {
	register byte arity;

#ifdef GC_TEST
	infcounter = 0;
        fprintf(stddbg, ".");
#endif
	arity = opbyte(test_heap_arg_1); /* the arity of the procedure */
	lpcreg += test_heap_sizeof; /* needs to be here because of exception */
        if (gen_gc_heap(arity)) {  /* GC potentially modifies hreg, so this */
	  if ((ereg - hreg) < margin) { /* seemingly repeated test is valid */
	    if (flags[STACK_REALLOC]) {
	      if (glstack_realloc(resize_stack0(glstack.size),arity) != 0) {
		local_global_exception(lpcreg);
	      }
	    } else {
	      xxx_warn("Reallocation is turned OFF !");
	      local_global_exception(lpcreg);
	    }
	  }
	}
	/* are there any locally cached quantities that must be reinstalled? */
      }
    else lpcreg += test_heap_sizeof;
    MoreMoreYesGimmeMore;
  }

  case builtin:
  builtin_label:
  {
    register byte op1;

    op1 = opbyte(builtin_arg_1);
    pcreg = lpcreg + builtin_sizeof;
    if (builtin_call(op1)) { lpcreg = pcreg; }
    else Fail1;
    MoreMoreYesGimmeMore;
   }

  case unifunc:   /* PAR */
  unifunc_label:
  {
    register byte op1;
    register CPtr op2;

    op1 = opbyte(unifunc_arg_1);
    op2 = opregaddr(unifunc_arg_2);
    if (unifunc_call(op1, op2) == 0) {
      xxx_error("Error in unary function call");
      Fail1;
    } else lpcreg += unifunc_sizeof;
    MoreMoreYesGimmeMore;
  }

  case fail:    /* PPP */
  fail_label:
  {
    Fail1; 
    MoreMoreYesGimmeMore;
  }

  case halt:  /* PPP */
  halt_label:
  {
    lpcreg += halt_sizeof;
    pcreg = lpcreg; 
    inst_begin = lpcreg;  /* hack for the moment to make this a ``creturn'' */
    return 0;	/* not "MoreMoreYesGimmeMore"! */
  }

/*----------------------------------------------------------------------*\
 * Term comparison instructions.                                        *
\*----------------------------------------------------------------------*/

  case term_comp: /* RRR */
  term_comp_label:
  {
    register Cell op1;
    register Cell op2;
    register CPtr op3;

    op1 = opreg(term_comp_arg_1);
    op2 = opreg(term_comp_arg_2);
    op3 = opregaddr(term_comp_arg_3);
    lpcreg += term_comp_sizeof;
    bld_int(op3, compare(op1, op2));
    MoreMoreYesGimmeMore;
  }

  case int_test_eq:   /* PPR-UN-L */
  int_test_eq_label:
  {
    register Cell op1;
    register Cell op2;

    op1 = opreg(int_test_eq_arg_1);
    op2 = opword(int_test_eq_arg_2);
    deref(op1);
    if (isnumber(op1)) {
      if ((int_val(op1) - (Integer)op2) != 0)
        lpcreg = oplabel(int_test_eq_arg_3);
      else lpcreg += int_test_eq_sizeof;
      MoreMoreYesGimmeMore;
    }
    else {
      lpcreg += int_test_eq_sizeof;
      xxx_arithmetic_comp_abort(op1, "=:=", op2);
      MoreMoreYesGimmeMore;
    }
  }

  case int_test_neq:   /* PPR-UN-L */
  int_test_neq_label:
  {
    register Cell op1;
    register Cell op2;

    op1 = opreg(int_test_neq_arg_1);
    op2 = opword(int_test_neq_arg_2);
    deref(op1);
    if (isnumber(op1)) {
      if ((int_val(op1) - (Integer)op2) == 0)
        lpcreg = oplabel(int_test_neq_arg_3);
      else lpcreg += int_test_neq_sizeof;
      MoreMoreYesGimmeMore;
    }
    else {
      lpcreg += int_test_neq_sizeof;
      xxx_arithmetic_comp_abort(op1, "=\\=", op2);
      MoreMoreYesGimmeMore;
    }
  }

  case atom_test_eq:    /* PPR-C-L */
  atom_test_eq_label:
  {
    register Cell op1;
    register Cell op2;

    op1 = opreg(atom_test_eq_arg_1);
    deref(op1);
    op2 = opword(atom_test_eq_arg_2);
    if (op1 == op2) lpcreg += atom_test_eq_sizeof;
    else lpcreg = oplabel(atom_test_eq_arg_3);
    MoreMoreYesGimmeMore;
  }

  case atom_test_neq:  /* PPR-C-L */
  atom_test_neq_label:
  {
    register Cell op1;
    register Cell op2;

    op1 = opreg(atom_test_neq_arg_1);
    deref(op1);
    op2 = opword(atom_test_neq_arg_2);
    if (op1 != op2) lpcreg += atom_test_neq_sizeof;
    else lpcreg = oplabel(atom_test_neq_arg_3);
    MoreMoreYesGimmeMore;
  }

/*----------------------------------------------------------------------*\
 * Just-In-Time Indexing instructions.                                  *
\*----------------------------------------------------------------------*/

  case jiti_on_bound:  /* PPR-L-UN */
  jiti_on_bound_label:
  { /* op1 is register, op2 is arity, op3 is predicate address */

#ifdef JIT_INDEXING
    register Cell op1;
    register byte *op2;
    register Cell op3;
    Cell uClauseNo;

    op1 = opreg(jiti_on_bound_arg_1);
    deref2(op1, { lpcreg += jiti_on_bound_sizeof; MoreMoreYesGimmeMore; });
#if (defined(ATTR_VARS))
    if (isattv(op1)) { lpcreg += jiti_on_bound_sizeof; MoreMoreYesGimmeMore; }
#endif

    op3 = opword(jiti_on_bound_arg_3);
    op2 = oplabel(jiti_on_bound_arg_2);
    op1 = opbyte(jiti_on_bound_arg_1);

#ifdef JITI_DEBUG
    fprintf(stderr,"JITI_ON_BOUND %d 0x%x %d\n",
	    (int)op1,(unsigned int)op2,(unsigned int)op3);
#endif 
    uClauseNo = get_clauseno(op2);

    /* create hash table */
    op2 = jiti_create_htable_asserts(op2,op1,op3);

    /* if the hash table isn't created, make the instruction a 'noop' */
    if (op2 == 0)
    {
#ifdef THREADED
      *((CPtr)lpcreg) = (Cell)&&noop_label;
      *(lpcreg+sizeof(CPtr)) = noop;
#else
      *(lpcreg) = noop;
#endif
      
      /* bytes to skip */
      *((byte *)(lpcreg+noop_arg_1)) = ((jiti_on_bound_sizeof-noop_sizeof)/2);
      lpcreg += jiti_on_bound_sizeof;
    }
    else
    {
      /* get number of clauses -> get the modulus */
      op3 = hsize(uClauseNo);
  
      /* operation code */
#ifdef THREADED
      *((CPtr)lpcreg) = (Cell)&&hash_on_bound_label;
      *(lpcreg+sizeof(CPtr)) = hash_on_bound;
#else
      *(lpcreg) = hash_on_bound;
#endif
      /* arguments of hash_on_bound */
      *((CPtr)(lpcreg+hash_on_bound_arg_2)) = (Cell)op2;
      *((CPtr)(lpcreg+hash_on_bound_arg_3)) = (Cell)op3;
  
      /* op1 is register, op2 is hash table offset, op3 is modulus */
      op1 = opreg(hash_on_bound_arg_1);
      deref2(op1, { lpcreg += jiti_on_bound_sizeof; MoreMoreYesGimmeMore; });
      if (isstruct(op1)) op1 = (Cell)get_str_psc(op1);
      else
        if (islist(op1)) op1 = (Cell)list_str;
#if (defined(ATTR_VARS))
        else
          if (isattv(op1)) { lpcreg += jiti_on_bound_sizeof; MoreMoreYesGimmeMore;}
#endif
      op1 = ihash(op1, op3);
      
#ifdef JITI_DEBUG
      fprintf(stderr,"HASH_ON_BOUND %d 0x%x %d\n",
	      (int)op1,(unsigned int)op2,(unsigned int)op3);
      fprint_inst(stderr,*(byte **)(op2 + op1 * sizeof(Cell)));
#endif
      lpcreg = *(byte **)(op2 + op1 * sizeof(Cell));
    }
#else /* NO - JIT_INDEXING */
    lpcreg += jiti_on_bound_sizeof;
#endif
    MoreMoreYesGimmeMore;
  }

  case jiti_on_atomic: /* PPR-L-UN */
  jiti_on_atomic_label:
  { /* op1 is register, op2 is arity, op3 is clauses no */
#ifdef JIT_INDEXING
    register Cell op1;
    register byte *op2; 
    register Cell op3;
    byte reg_num;
	  
    op1 = opreg(jiti_on_atomic_arg_1);
    deref2(op1, { lpcreg += jiti_on_atomic_sizeof; MoreMoreYesGimmeMore; });
#if (defined(ATTR_VARS))
    if (isattv(op1)) { lpcreg += jiti_on_atomic_sizeof; MoreMoreYesGimmeMore; }
#endif

    op3 = opword(jiti_on_atomic_arg_3);
    /* xxx_mesg("JITI on atomic opportunity (NumClauses = %d, Hsz = %d)", op3, hsize(op3)); */
    reg_num = opbyte(jiti_on_atomic_arg_1);
    op2 = (byte *)opword(jiti_on_atomic_arg_2);
    /* jiti_create_hash_table creates the hash table is pointed by indextab */
    op2 = jiti_create_hash_table((CPtr)lpcreg, reg_num, (Integer)op2, op3);
 
    /* patch_instruction:
     * hash_on_atomic op1 is register, op2 is placeholder, op3 is modulus
     */
    op3 = hsize(op3);
    /* operation code */
#ifdef THREADED
    *((CPtr)lpcreg) = (Cell)&&hash_on_atomic_label;
    *(lpcreg+sizeof(CPtr)) = hash_on_atomic;
#else
    *(lpcreg) = hash_on_atomic;
#endif
    /* arguments of hash_on_atomic */
    *((CPtr)(lpcreg+hash_on_atomic_arg_2)) = (Cell)op2;
    *((CPtr)(lpcreg+hash_on_atomic_arg_3)) = (Cell)op3;

    /* execution of hash instruction */
    lpcreg = *(byte**)(op2 + ihash(op1, op3) * sizeof(Cell));
#else /* NO - JIT_INDEXING */
    lpcreg += jiti_on_atomic_sizeof;
#endif
    MoreMoreYesGimmeMore;
  }

/*----------------------------------------------------------------------*\
 * Tabling instructions.                                                *
\*----------------------------------------------------------------------*/

  /*
   * Instruction `check_interrupt' is used before `new_answer' to
   * handle the pending attv interrupts.  It is similar to `call' but
   * the second argument (S) is not used currently.
   */
  case check_interrupt: /* PPA-S */
  check_interrupt_label:
  {
#ifdef ATTR_VARS
    /* register Cell op2; */

    /* op2 = opword(check_interrupt_arg_2); */
    lpcreg += check_interrupt_sizeof;
    if (int_val(cell(interrupt_reg)) > 0) {
      cpreg = lpcreg;
      bld_struct(reg + 2, hreg);	/* see subp.c: build_call() */
      new_heap_functor(hreg, true_psc);
      bld_copy(reg + 1, build_interrupt_chain());
      lpcreg = get_ep((Psc) flags[MYSIG_ATTV + INT_HANDLERS_FLAGS_START]);
    }
#else
    lpcreg += check_interrupt_sizeof;
#endif
    MoreMoreYesGimmeMore;
  }

/*----------------------------------------------------------------------*/

#include "slginsts_xsb_i.h"

/*----------------------------------------------------------------------*/

#include "tc_insts_xsb_i.h"

/*----------------------------------------------------------------------*/

  case areg_equal: /* PPR-C-L */
  areg_equal_label:
  { /* op1 is register, op2 contains a (tagged) term, op3 is label to go to */
    register Cell op1;
    register Cell op2;

    op1 = opreg(areg_equal_arg_1);
    deref(op1);
    if (isstruct(op1)) { op1 = (Cell)get_str_psc(op1); }
    op2 = opword(areg_equal_arg_2);
    if (op1 == op2) lpcreg = oplabel(areg_equal_arg_3);
    else lpcreg += areg_equal_sizeof;
    MoreMoreYesGimmeMore;
  }

  case load_pred:  /* PPP-S */
  load_pred_label:
  {
    register Psc psc;

    psc = (Psc)opword(load_pred_arg_1);
#ifdef DEBUG
    /* check env or type to give (better) error msgs? */
    switch (get_type(psc))
      {
      case T_PRED:
      case T_DYNA:
	xxx_abort("System Error: trying to load an already loaded predicate");
	break;
      default:
	/* printf("loading module %s for %s/%d\n",
	   get_name(get_data(psc)),get_name(psc),get_arity(psc)); */
#endif
	bld_struct(reg+1, build_call(psc)); /* put call-term in r1 */
	psc = (Psc)flags[MYSIG_UNDEF+32];   /* get psc of undef handler */
	lpcreg = get_ep(psc);               /* ep of undef handler */
	bld_int(reg+2, MYSIG_UNDEF);        /* undef-pred code */
#ifdef DEBUG
      }
#endif
    MoreMoreYesGimmeMore;
  }

  case call_forn: /* PPP-L -- generated dynamically */
  call_forn_label:
  {
    register Cell op1;

    op1 = opword(call_forn_arg_1);
    if (((PFI)op1)())  /* call foreign function */
      lpcreg = cpreg;
    else Fail1;
    MoreMoreYesGimmeMore;
  }

  case switchon3bound: /* RRR-L-L */
  switchon3bound_label:
  { /* op1 is register, op2 is hash table offset, op3 is modulus */
    register Cell op3;
    int  i, j = 0;
    Cell opa[3];

    if (opbyte(switchon3bound_arg_1) == 0) { opa[0] = 0; }
    else opa[0] = opreg(switchon3bound_arg_1);
    opa[1] = opreg(switchon3bound_arg_2);
    opa[2] = opreg(switchon3bound_arg_3);
    op3 = opword(switchon3bound_arg_5);
    /* This is not a good way to do this, but until we put retract into C,
       or add new builtins, it will have to do. */
    for (i = 0; i <= 2; i++) {
      register Cell op1;
      if (opa[i] != 0) {
	op1 = opa[i];
	deref2(op1, { lpcreg += switchon3bound_sizeof;
	              goto out_of_switchon3bound; });
	if (isstruct(op1)) op1 = (Cell)get_str_psc(op1);
	else
	  if (islist(op1)) op1 = (Cell)list_str;
#if (defined(ATTR_VARS))
	  else
	    if (isattv(op1))
	      { lpcreg += switchon3bound_sizeof; goto out_of_switchon3bound; }
#endif
	j = (j<<1) + ihash(op1, op3);
      }
    }
    {
      register byte *op2 = oplabel(switchon3bound_arg_4);
      lpcreg = *(byte **)(op2 + ((j % op3) * sizeof(Cell)));  /* THREADED OK ? */
    }
  out_of_switchon3bound:
    MoreMoreYesGimmeMore;
  }

  case dyntrustmeelsefail: /* PPA-L, second word ignored */
  dyntrustmeelsefail_label:
  {
    register Cell op1;

    op1 = opbyte(dyntrustmeelsefail_arg_1);
    lpcreg += dyntrustmeelsefail_sizeof;
    trust_restore_sub(op1);
    MoreMoreYesGimmeMore;
  }

  case noop:  /* PPA */
  noop_label:
  {
    register byte op1;

    op1 = opbyte(noop_arg_1);
    lpcreg += noop_sizeof + op1 + op1;
    MoreMoreYesGimmeMore;
  }

/*----------------------------------------------------------------------*/

#ifdef ATTR_VARS
  case getattv: /* PPR */
  getattv_label:
  {
    register Cell op1;

    op1 = opreg(getattv_arg_1);
    lpcreg += getattv_sizeof;
    nunify_with_attv(op1);
    MoreMoreYesGimmeMore;
  }

  case putattv: /* PPR */
  putattv_label:
  {
    register CPtr op1;

    op1 = opregaddr(putattv_arg_1);
    lpcreg += putattv_sizeof;
    bld_attv(op1, hreg);
    new_heap_free(hreg);
    MoreMoreYesGimmeMore;
  }
#endif

#ifdef GGC
  case backtrack: /* PPP --- temporarily to approximate gen_gc */
  backtrack_label:
  {
    breg = cp_prevbreg(breg);
    restore_trail_condition_registers(breg);
    MoreMoreYesGimmeMore;
  }
#endif

  case reset_bytecode:  /* PPP */
  reset_label:
  { /*
     * This instruction is added just to provide a fall back point for
     * xxx_abort() or xxx_segfault_catcher().  It is called only once, and 
     * then control goes to startaddr.
     */
    if ((lpcreg = (byte *) (long)setjmp(xxx_abort_fallback_environment))) {
      /* Short-circuit untrailing */
      trreg = cp_trreg(breg);
      /* Restore the default signal handling */
      signal(SIGSEGV, xxx_default_segfault_handler);
    }
    else lpcreg = startaddr;  /* first instruction of entire engine */
    MoreMoreYesGimmeMore;
  }

  default:
  unused_instruction_label:
  {
    xxx_exit("Illegal opcode hex %x", get_opcode(*(void **)lpcreg));
  }
} /* end of switch */

  return 0;

} /* end of emuloop() */

/*======================================================================*/
