/* File:      auxlry.c
** Author(s): Warren, Sagonas, Xu
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: auxlry.c,v 1.7 2005/02/15 15:42:19 kostis Exp $
** 
*/


#include "xxx_config.h"

#include <stdio.h>

#ifdef WIN_NT
#include <time.h>

#else

/* time.h is not automatically entailed by sys/time.h under RedHat 7.2 */
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>

#ifdef HP700
#include <sys/syscall.h>
#define getrusage(T, USAGE)	syscall(SYS_getrusage, T, USAGE);
#endif

#endif

#include "auxlry.h"

/*----------------------------------------------------------------------*/

long cpu_time(void)
{
  long time_sec;

#if defined(WIN_NT)

  time_sec = ((long) clock() / CLOCKS_PER_SEC) * 1000;

#else
  struct rusage usage;

  getrusage(RUSAGE_SELF, &usage);
  time_sec = (long)(usage.ru_utime.tv_sec * 1000) +
	     (long)(usage.ru_utime.tv_usec / 1000.0);
#endif

  return time_sec;
}

/*----------------------------------------------------------------------*/

void get_date(int *year, int *month, int *day,
	      int *hour, int *minute, int *second)
{
#ifdef WIN_NT
    SYSTEMTIME SystemTime;
    GetSystemTime(&SystemTime);
    *year = SystemTime.wYear;
    *month = SystemTime.wMonth;
    *day = SystemTime.wDay;
    *hour = SystemTime.wHour;
    *minute = SystemTime.wMinute;
    *second = SystemTime.wSecond;
#else
#ifdef HAVE_GETTIMEOFDAY
    struct timeval tv;
    struct tm *tm;

    gettimeofday(&tv,NULL);
    tm = gmtime(&tv.tv_sec);
    *year = tm->tm_year;
    if (*year < 1900)
      *year += 1900;
    *month = tm->tm_mon + 1;
    *day = tm->tm_mday;
    *hour = tm->tm_hour;
    *minute = tm->tm_min;
    *second = tm->tm_sec;
#endif
#endif
}

/*----------------------------------------------------------------------*/

double real_time(void)
{
#if defined(WIN_NT)
  double value = ((float) clock() / CLOCKS_PER_SEC);
#else
  double value;
  struct timeval tvs;

  gettimeofday(&tvs, 0);
  value = tvs.tv_sec + 0.000001 * tvs.tv_usec;
#endif
  return value;
}

/*----------------------------------------------------------------------*/
