/* File:      binding.h
** Author(s): Jiyang Xu, Terrance Swift, Kostis Sagonas, Baoqiu Cui
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** Copyright (C) ECRC, Germany, 1990
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: binding.h,v 1.4 2003/09/24 17:12:52 kostis Exp $
** 
*/

#ifndef __BINDING_H__
#define __BINDING_H__

/*
 * Structures Of Different Trail Frames:
 *
 * In CHAT, trreg always points to the address field of a trail frame, and
 * the trail stack may contain both 1-word trail frames or 2-word ones.
 * After untrailing a trail frame, trreg is set to trreg-1 or trreg-2
 * (depending on the lowest bit of the address field of the trail frame).
 *
 * 1) Regular WAM-TRAIL frame in CHAT (1-word trail frame)
 * -------------------------------------------------------
 *     +-----+
 *     |     | Address of the trailed variable
 *     +-----+
 *
 * 2) Pre-image trail frame in CHAT (2-word trail frame)
 * -----------------------------------------------------
 *     +-----+
 *     |     | Old value of the trailed variable (cell)
 *     +---+-+
 *     |   |1| Address of the trailed variable (with PRE_IMAGE_MARK)
 *     +---+-+
 */

#ifdef ATTR_VARS
#define PRE_IMAGE_TRAIL

#define PRE_IMAGE_MARK   1
#endif

/* --- trailing ------------------------------------------------------- */

#define conditional(a)	( ((a) >= ebreg) || ((a) < hbreg) )

#define pushtrail(a,v)	if (conditional(a)) { pushtrail0(a,v); }
#define dpushtrail(a,v) pushtrail0(a,v)

#define pushtrail0(addr,val) {						\
  if ((char *)(top_of_trail) > ((char *)(top_of_cpstack) - 10)) {	\
    handle_tcpstack_overflow();						\
  }									\
  *(trreg++) = addr;							\
}

#ifdef PRE_IMAGE_TRAIL
#define push_pre_image_trail(a, new_v)			\
  if (conditional(a)) {push_pre_image_trail0(a, new_v)}

#define push_pre_image_trail0(addr, new_value) {			\
  if ((char *)(top_of_trail) > ((char *)(top_of_cpstack) - 20)) {	\
    handle_tcpstack_overflow();						\
  }									\
  *(trreg++) = (CPtr) (cell(addr));					\
  *(trreg++) = (CPtr) ((Cell) (addr) | PRE_IMAGE_MARK);			\
}
#endif /* PRE_IMAGE_TRAIL */

/* --- binding -------------------------------------------------------- */

#define bind_int(addr, val)	pushtrail(addr, makeint(val)); \
				bld_int(addr, val)

#define bind_float(addr, val)	pushtrail(addr, (Cell)makefloat(val)); \
				bld_float(addr, val)

#define bind_ref(addr, val)	pushtrail(addr, val); \
				bld_ref(addr, val)

#define dbind_ref(addr, val)	dpushtrail(addr, val); \
				bld_ref(addr, val)

#define bind_struct(addr, str)	pushtrail(addr, makestruct(str)); \
				bld_struct(addr, str)

#define bind_atom(addr, atm)	pushtrail(addr, makeatom(atm)); \
				bld_atom(addr, atm)

#define bind_nil(addr)		pushtrail(addr, makenil); \
				bld_nil(addr)

#define bind_list(addr, list)	pushtrail(addr, makelist(list)); \
				bld_list(addr, list)

#ifdef ATTR_VARS
#define bind_attv(addr, attv)	pushtrail(addr, makeattv(attv)); \
				bld_attv(addr, attv)
#endif

#define bind_copy(addr, val)	pushtrail(addr, val); *(addr) = val

/* value trail MUST be used because first CP cell has cp_trreg = 0 !!!! */

#define untrail(addr) bld_free(addr)

#ifdef PRE_IMAGE_TRAIL	/* untrail2 is for pre_image trail. */
#define untrail2(trail_ptr, addr) {		\
  if ((addr) & PRE_IMAGE_MARK) {		\
    bld_copy((CPtr)((addr) - PRE_IMAGE_MARK),	\
             cell((CPtr)trail_ptr - 1));	\
    trreg--;					\
  }						\
  else						\
    bld_free((CPtr)(addr));			\
}
#endif /* PRE_IMAGE_TRAIL */


/* --- testing location of variable ----------------------------------- */

#define ereg_on_top(t_ereg)	t_ereg < ebreg

/* -------------------------------------------------------------------- */

#endif /* __BINDING_H__ */
