/* File:      dis.c
** Author(s): Warren, Xu, Swift, Sagonas
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** Copyright (C) ECRC, Germany, 1990
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: dis.c,v 1.11 2006/08/23 15:52:11 kostis Exp $
** 
*/


#include "xxx_config.h"

#include <stdio.h>
#include <string.h>

#include "auxlry.h"
#include "psc_xsb.h"
#include "hash_xsb.h"
#include "cell_xsb.h"
#include "inst_xsb.h"
#include "flags_xsb.h"
#include "builtin.h"
#include "macro_xsb.h"
#include "loader_xsb.h"
#include "memory_xsb.h"  /* for fail_inst */
#include "dis.h"

/*-----------------------------------------------------------------------*/

static void dis_data_sub(FILE *fd, Pair *chain_ptr)
{
  while (*chain_ptr) {
    Psc psc = (*chain_ptr)->psc_ptr;
    fprintf(fd, "%p: %s/%d,\t", (void*)psc, get_name(psc), get_arity(psc));
    fflush(fd);
    switch(get_type(psc)) 
      {
      case T_PRED: fprintf(fd, "PRED, "); break;
      case T_DYNA: fprintf(fd, "DYNA, "); break;
      case T_ORDI: fprintf(fd, "ORDI, "); break;
      case T_MODU: fprintf(fd, "MODU, "); break;
      case T_FORN: fprintf(fd, "FORN, "); break;
      case T_UDEF: fprintf(fd, "UDEF, "); break;
      default:	   fprintf(fd, "????"); break;
      }
    switch (get_env(psc))
      {
      case T_VISIBLE:  fprintf(fd, "VISIBLE, "); break;
      case T_HIDDEN:   fprintf(fd, "HIDDEN, "); break;
      case T_UNLOADED: fprintf(fd, "UNLOADED, "); break;
      default:	       fprintf(fd, "(error env), "); break;
      }
    if (get_type(psc) == T_PRED) {
      if (get_tip(psc) != NULL) fprintf(fd, "TABLED, ");
    }
    fprintf(fd, "%p\n", get_ep(psc));   /* dsw ???? */
    chain_ptr = &((*chain_ptr)->next);
   } /* while */
}

/*-----------------------------------------------------------------------*/

static void dis_data(FILE *fd)
{
  unsigned long i;
  Pair *temp_ptr;
  char *modname;
  
  temp_ptr = (Pair *)(&flags[MOD_LIST]);
  while (*temp_ptr) {
    Psc psc = (*temp_ptr)->psc_ptr;
    modname = get_name(psc);
    if (strcmp(modname,"usermod") != 0) {
      fprintf(fd, "/*===== module: '%s' is %s =====*/\n\n", modname,
	      (get_type(psc) ? "LOADED" : "UNLOADED"));
      if (!strcmp(modname,"global"))
	for (i=0; i < symbol_table.size; i++) {
	  if (symbol_table.table[i] != 0) {
	    fprintf(fd, "... BUCKET no. %ld\n", i);
	    dis_data_sub(fd, (Pair *)(symbol_table.table + i));
	  }
	}
      else dis_data_sub(fd, (void *)&get_data(psc));
      fprintf(fd, "\n");
    }
    temp_ptr = &((*temp_ptr)->next);
  }
}

/*-----------------------------------------------------------------------*/

byte *fprint_inst(FILE *fd, const byte *inst_ptr)
{
  byte opcode;
  CPtr instr;
  CPtr word_instr_ptr;
  int i, a;
  
  word_instr_ptr = (CPtr)inst_ptr;
  fprintf(fd, "%p:  ", (void*)word_instr_ptr);
  instr = word_instr_ptr;
  opcode = get_opcode(instr);
  fprintf(fd, "%s", inst_name(opcode));
  /* word_instr_ptr points to the first word after word containing the opcode */
  word_instr_ptr += MIN_INSTR_LEN;
  a = 1; /* current operand */
  for (i=1; i<=4; i++) {
    switch (inst_table[opcode][i])
      {
      case A:
	if (opcode == builtin) {
	  a++;
	  fprintf(fd, "\t%d\t\t%% ", inst_operand3(instr));
	  fprintf(fd, "%s", (char *)builtin_table[inst_operand3(instr)][0]);
	} else fprintf(fd, "\t%d", inst_operandn(instr,a++));
	break;
      case V:
	fprintf(fd, "\tv%d", inst_operandn(instr,a++));
	break;
      case R:
	fprintf(fd, "\tr%d", inst_operandn(instr,a++));
	break;
      case T:
	fprintf(fd, "\t%lx", *(word_instr_ptr++));
	break;
      case P:
	a++;
	break;
      case S:
	{
	  Psc psc = (Psc) *(word_instr_ptr++);
	  fprintf(fd,"\t%p (%s/%d)", (void*)psc, get_name(psc), get_arity(psc));
	}
	break;
      case C:
	{
	  Cell c = *(word_instr_ptr++);
	  fprintf(fd, "\t0x%lx (%s)", c, atom_val(c));
	}
	break;
      case L:
      case E:
      case F:
	fprintf(fd, "\t0x%lx", *(word_instr_ptr++));
	break;
      case N:
	{
	  Cell arg = *(word_instr_ptr);
	  fprintf(fd, "\t%ld (%ld)", arg, (long)int_val(arg));
	  word_instr_ptr++;
	  break;
	}
      case I:
	if (i == 2)  /* second operand of switchonbound */
	  fprintf(fd, "\t0x%lx", *(word_instr_ptr++));
	else         /* third operand of switchonbound */
	  fprintf(fd, "\t%ld", *(word_instr_ptr++));
	break;
      case UN:
	fprintf(fd, "\t%ld", *(word_instr_ptr++));
	break;
      case PP:
	a += 2;
	break;
      case PPR:
	fprintf(fd, "\tr%d", inst_operand3(instr));
	if (opcode == areg_equal)
	  {
	    Cell c = *(word_instr_ptr++);
	    if (isinteger(c)) fprintf(fd, "\t%ld (%ld)", c, (long)int_val(c));
	    else
	    if (isatom(c)) fprintf(fd, "\t0x%lx (%s)", c, atom_val(c));
	    else {
	      Psc psc = (Psc)c;
	      fprintf(fd,"\t%p (%s/%d)", (void*)psc, get_name(psc), get_arity(psc));
	    }
	    fprintf(fd, "\t0x%lx", *(word_instr_ptr++));
	    i += 2;
	  }
	break;
      case RRR:
	fprintf(fd, "\tr%d", inst_operand1(instr));
	fprintf(fd, "\tr%d", inst_operand2(instr));
	fprintf(fd, "\tr%d", inst_operand3(instr));
	break;
      case RRV:
	fprintf(fd, "\tr%d", inst_operand1(instr));
	fprintf(fd, "\tr%d", inst_operand2(instr));
	fprintf(fd, "\tv%d", inst_operand3(instr));
	break;
      case X:
      case PPP:
	break;
      default:
	break;
      }  /* switch */
    if (opcode == noop)
      word_instr_ptr += inst_operand3(instr)/2; /* ?!@% */
  } /* for */
  fprintf(fd, "\n");
  fflush(fd);
  return (byte *)word_instr_ptr;
} /* fprint_inst */

/*-----------------------------------------------------------------------*/

static void dis_text(FILE *fd, const byte *inst_begin)
{
  pseg   this_seg;
  pindex index_seg;
  CPtr   endaddr, inst_addr;
  
  fprintf(fd, "/*===================== text below =====================*/\n");
  this_seg = (pseg) inst_begin;
  while (this_seg) {		/* repeat for all text segments */
    fprintf(fd, "\n/*----- New segment below (size = %ld, realsize = %ld bytes) -----*/\n\n", 
	    seg_size(this_seg), seg_size(this_seg)-SIZE_SEG_HDR);
    endaddr = (CPtr)((byte *)seg_hdr(this_seg) + seg_size(this_seg));
    inst_addr = seg_text(this_seg);
    while (inst_addr < endaddr)
      inst_addr = (CPtr)fprint_inst(fd, (byte *)inst_addr);
    index_seg = seg_index(this_seg);
    while (index_seg) {
      inst_addr = i_block(index_seg);
      endaddr = (CPtr)((byte *)index_seg + i_size(index_seg));
      if (get_opcode(inst_addr) == try ||
	  get_opcode(inst_addr) == areg_equal) {
	/* is try/retry/trust */
	while (inst_addr < endaddr) 
	  inst_addr = (CPtr)fprint_inst(fd, (byte *)inst_addr);
      } else {					/* is hash table */
	fprintf(fd, "hash table....\n");
	while (inst_addr < endaddr) {
	  if (*(inst_addr) == (Cell)&fail_inst)
	    fprintf(fd, "  %p:    &fail\n", (void*)inst_addr);
	  else
	    fprintf(fd, "  %p:  0x%lx\n", (void*)inst_addr, *(inst_addr));
	  inst_addr++;
	}
	fprintf(fd, "       end.... %ld buckets total\n",
		(i_size(index_seg) - SIZE_IDX_HDR) / sizeof(Cell));
      }
      index_seg = i_next(index_seg);
    }
    this_seg = seg_next(this_seg);
  }  
}

/*-----------------------------------------------------------------------*/

void dis(bool distext, byte *inst_begin)
{  
   dis_data(stdout);
   if (distext) dis_text(stdout, inst_begin);
}

/*-----------------------------------------------------------------------*/
