/* File:      emudef.h
** Author(s): Warren, Xu, Swift, Sagonas
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) Uppsala University, 2000
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** Copyright (C) ECRC, Germany, 1990
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: emudef.h,v 1.6 2003/09/04 14:45:14 kostis Exp $
** 
*/


/* Argument Registers
   ------------------ */
Cell reg[MAX_REGS];


/* Special Registers
   ----------------- */
CPtr ereg;		/* last activation record       */
CPtr breg;		/* last choice point            */
CPtr hreg;		/* top of heap                  */
CPtr *trreg;		/* top of trail stack           */
CPtr hbreg;		/* heap back track point        */
byte *cpreg;		/* return point register        */
byte *pcreg;		/* program counter              */
CPtr ebreg;		/* breg into environment stack	*/

CPtr pdlreg;
ComplStackFrame openreg;

CPtr ptcpreg;
CPtr delayreg;

char *nil_sym, *true_sym;
Pair list_str;

Psc list_psc;
Psc tnot_psc, delay_psc, true_psc, if_psc;

/*
 * Ret PSCs are used to store substitution factors for subgoal calls
 * or answers.  A psc with a new arity will be created when needed,
 * except that ret_psc[0] stores the pointer to the atom "ret" and is
 * initialized when the system is started.
 */
Psc ret_psc[MAX_ARITY];

char *list_dot;

/*======================================================================*/
/* Include stuff added by Kostis to control the emulator generation...  */
/*======================================================================*/

#include "emu_ctrl.h"

/*======================================================================*/

#if (defined(ATTR_VARS) && defined(DEBUG_ATTV))
#define attv_dbgmsg(String) xsb_dbgmsg(String)
#else
#define attv_dbgmsg(String)
#endif

/*======================================================================*/

#ifdef ATTR_VARS
/*
 * In getattv, the mode will always be WRITE.  The unification will be
 * done by the attv unification handlers.
 */
#define nunify_with_attv(OP1) {					\
  deref(OP1);							\
  if (isref(OP1)) {						\
    bind_attv((CPtr)(OP1), hreg);				\
  }								\
  else {							\
    attv_dbgmsg(">>>> nunify_with_attv, interrupt needed\n");	\
    add_interrupt(makeattv(hreg), OP1);				\
  }								\
  new_heap_free(hreg);		/* the VAR part of the attv */	\
  sreg = NULL;                  /* mode = WRITE */		\
}
#endif

/*======================================================================*/

#ifdef ATTR_VARS
#define call_sub(PSC) {							\
  if ( (asynint_val | int_val(cell(interrupt_reg))) == 0)		\
    /* no interrupt detected */						\
    lpcreg = (byte *)get_ep(PSC);					\
  else {								\
    if (asynint_val & KEYINT_MARK) {					\
      synint_proc(PSC, MYSIG_KEYB);					\
      lpcreg = pcreg;							\
      asynint_val = asynint_val & ~KEYINT_MARK;				\
      asynint_code = 0;							\
    } else {								\
      if (int_val(cell(interrupt_reg))) { /* attv interrupt */		\
        synint_proc(PSC, MYSIG_ATTV);					\
        lpcreg = pcreg;							\
        /* Set PSC to '_$attv_int'/2, so that the later call of */	\
        /* intercept(PSC) will set the return point, pcreg, to  */	\
        /* '_$attv_int'/2.				        */	\
        PSC = (Psc) flags[MYSIG_ATTV+INT_HANDLERS_FLAGS_START];		\
      } else {								\
        lpcreg = (byte *)get_ep(PSC);					\
        if (asynint_val & MSGINT_MARK) { /* for debug or statistics */	\
          pcreg = lpcreg;						\
          intercept(PSC);						\
          lpcreg = pcreg;						\
        } else asynint_code = 0;					\
      }									\
    }									\
  }									\
}

#else

#define call_sub(PSC) {							\
  if (asynint_val == 0) /* no interrupt detected */			\
    lpcreg = (byte *)get_ep(PSC);					\
  else {								\
    if (asynint_val & KEYINT_MARK) {					\
      synint_proc(PSC, MYSIG_KEYB);					\
      lpcreg = pcreg;							\
      asynint_val = asynint_val & ~KEYINT_MARK;				\
      asynint_code = 0;							\
    } else {								\
      lpcreg = (byte *)get_ep(PSC);					\
      if (asynint_val & MSGINT_MARK) { /* for debug or statistics */	\
        pcreg = lpcreg;							\
        intercept(PSC);							\
        lpcreg = pcreg;							\
      } else asynint_code = 0;						\
    }									\
  }									\
}
#endif

/*======================================================================*/
