/* File:      emuloop.c
** Author(s): Warren, Xu, Swift, Sagonas
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** Copyright (C) ECRC, Germany, 1990
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: emuloop.c,v 1.17 2005/10/13 17:10:25 kostis Exp $
** 
*/

#include "xxx_config.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#if (defined(FOREIGN) && !defined(FOREIGN_ELF) && !defined(FOREIGN_WIN32))
#include <sys/un.h>
#endif

#include "auxlry.h"
#include "cell_xsb.h"
#include "error_xsb.h"
#include "inst_xsb.h"
#include "psc_xsb.h"
#include "deref.h"
#include "memory_xsb.h"
#include "sig_xsb.h"
#include "heap_xsb.h"
#include "flags_xsb.h"
#include "macro_xsb.h"
#include "choice.h"
#include "register.h"
#include "sw_envs.h"
#include "emudef.h"
#include "bldval.h"
#include "loader_xsb.h"
#include "tables.h"
#include "subp.h"
#include "tr_utils.h"
#include "chat.h"
#include "chatsched.h"
#include "export.h"
#include "dis.h"
#include "orient_xsb.h"
#include "debug.h"
#include "builtin.h"
#include "function.h"
#include "init_xxx.h"
#include "io_builtins_xsb.h"
#include "emuloop.h"

/*----------------------------------------------------------------------*/

/*
 * Variable ans_var_pos_reg is a pointer to substitution factor of an
 * answer in the heap.  It is used and set in function
 * variant_answer_search().  The name of this variable is from VarPosReg, a
 * variable used in variant_call_search() to save the substitution factor
 * of the call.
 */
CPtr	ans_var_pos_reg;

/*----------------------------------------------------------------------*/

#include "tr_delay.h"
#include "tr_code_xsb_i.h"

/*----------------------------------------------------------------------*/

static byte *inst_begin = NULL;   /* Ptr to the beginning of instr. array */

bool neg_delay;
int  level_num;

#ifdef DEBUG
int  xctr;
#endif

/*======================================================================*/
/*  Unification routine.						*/
/*======================================================================*/

bool unify(Cell rop1, Cell rop2)
{ /* begin unify */
  register Cell op1, op2;

  op1 = rop1; op2 = rop2;
/*-------------------------*/
#include "unify_inc.h"
/*-------------------------*/
}  /* end of unify */


/*======================================================================*/
/* Code for handling cuts appears below.  It was written in December	*/
/* 1997 by Kostis Sagonas based on the code written by Terry Swift and	*/
/* Rui Marques for unwinding the trail.  The input of Bart Demoen is	*/
/* also acknowledged.							*/
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/* Takes a pointer to the choice point frame we are cutting back to.	*/
/*----------------------------------------------------------------------*/

#define cut_restore_trail_condition_registers(CUTB) \
    ebreg = cp_ebreg(CUTB); \
    hbreg = cp_hreg(CUTB);

/*----------------------------------------------------------------------*/
/* Deletes all trail frames that are no longer conditional.		*/
/* Most probably, it does *NOT* work for cuts over tables!!		*/
/*----------------------------------------------------------------------*/

#include "trail_defs.h"

#define unwind_trail(tbreg) {	\
    register CPtr from, to;					\
    from = to = (CPtr) cp_trreg(tbreg);				\
    while (from < (CPtr)trreg) {				\
      if (good_trail_register(from)) { *to = *from; to++; }	\
      from++;							\
    }								\
    trreg = (CPtr *)to;						\
  }

/*----------------------------------------------------------------------*/
/* The main problem is to find which part of the trail to clean up.	*/
/* Cleaning the trail means deleting trail frames for bindings that are	*/
/* no longer conditional, i.e. unbinding these variables.  In general,	*/
/* one needs to tidy/unwind the part of the trail that lies between the	*/
/* current choice point (breg) and the choice point that is immediately	*/
/* _after_ the choice point that we are really cut back to (cut_breg).	*/
/* Of course, if we are cutting to the current choice point which often	*/
/* happens, or if there is nothing in between, there is no need to	*/
/* check whether there is trail to unwind.				*/
/* This is the general scheme and the code that pretty much implements  */
/* this, is as follows:                                         	*/
/*									*/
/*	void cut_code(OP1) {						*/
/*	    deref(OP1);							*/
/*	    cut_breg = (CPtr)(tcpstack.high - int_val(OP1));		*/
/*	    cut_restore_trail_condition_registers(cut_breg);		*/
/*	    if (breg != cut_breg) {					*/
/*		while (cp_prevbreg(breg) != cut_breg)			*/
/*		    breg = cp_prevbreg(breg);				*/
/*		unwind_trail(breg);					*/
/*		breg = cut_breg;					*/
/*	    }            						*/
/*	}								*/
/*----------------------------------------------------------------------*/

inline static void cut_code(Cell offset)
{
  register CPtr cut_breg;

  cut_breg = (CPtr)(tcpstack.high - int_val(offset));
  cut_restore_trail_condition_registers(cut_breg);
  if (breg != cut_breg) { /* not cutting back to the current CP */
  /* xsb_dbgmsg("Tidying trail (cutbreg = %p, breg = %p)", cut_breg, breg); */
    while (cp_prevbreg(breg) != cut_breg) {
      if (openreg != COMPLSTACKBOTTOM) check_and_clean_table_cut(breg);
      breg = cp_prevbreg(breg);
    }
    if (openreg != COMPLSTACKBOTTOM) check_and_clean_table_cut(breg);
    unwind_trail(breg);
    breg = cut_breg;
  }
}


/*======================================================================*/
/* the main emulator loop.						*/
/*======================================================================*/

/*
 * The WAM instructions are aligned with word (4 bytes on 32-bit machines,
 * or 8-byte on 64-bit machines), the shortest instructions (like fail)
 * take one word, and the longest ones take three words (like
 * switchon3bound).  If an instruction takes more than one word, then the
 * 2nd (or 3rd) word always contains an operand that takes one word.  The
 * one-word operands can be (see file emu/inst_xsb.h):
 *
 * 	L - label
 * 	S - structure symbol
 * 	C - constant symbol
 * 	N - number
 * 	G - string
 * 	I - 2nd & 3rd arguments of switchonbound
 * 	F - floating point number
 *
 * The opcode of all instructions takes the first byte in the first word.
 * The rest 3 bytes contain operands that needs only one byte.  These
 * one-byte operands can be:
 *
 * 	P - pad, not used
 * 	A - one byte number
 * 	V - variable offset
 * 	R - register number
 *
 * (In 64-bit machines there are 4 bytes of extra padding space for each 
 *  instruction)
 */

/*======================================================================*/

#include "abs_machine.h"

/*======================================================================*/

/* place for a meaningful message when segfault is detected */
char *xxx_default_segfault_msg =
     "\n++Memory violation occurred during evaluation.\n++Please report this bug to kostis@csd.uu.se.\n++Make sure you supply all information necessary to reproduce the bug.\n";
char *xxx_segfault_message;

/*======================================================================*/

DllExport int call_conv xxx(int calling_mode, int argc, char *argv[])
{ 
   char *startup_file;
   FILE *fd;
   unsigned int magic_num;
   static double realtime;    /* To retain its value across invocations */
#if (defined(FOREIGN) && !defined(FOREIGN_ELF) && !defined(FOREIGN_WIN32))
   char tfile[];
#endif

   xxx_segfault_message = xxx_default_segfault_msg;

   if (calling_mode == 0) {  /* initialize XXX */

#ifdef THREADED
     /* The following dummy call to emuloop sets up the inst_label_tab
	which is needed by function load_text of the loader. */
     emuloop(NULL);
#endif
     /* Set the name of the executable to the real name.
	The name of the executable could have been set in cinterf.c:xsb_init
	if XSB is called from C. In this case, we don't want `executable'
	to be overwritten, so we check if it is initialized. */
     if (executable[0] == '\0')
       xxx_executable_full_path(argv[0]);

     /* set install_dir, xsb_config_file, and user_home */
     xxx_orient();

     realtime = real_time();
     setbuf(stdout, NULL);
     startup_file = init_para(argc, argv);	/* init parameters */
     init_machine();		/* init space, regs, stacks */
     init_inst_table();		/* init table of instruction types */
     init_builtin_table();
     init_symbols();		/* preset a few symbols in PSC table */
     init_interrupt();		/* catch ^C interrupt signal */

     /* "b" does nothing, but POSIX allows it */
     fd = fopen(startup_file, "rb");

     if (!fd) {
       char message[256];
       sprintf(message, "The startup file, %s, could not be found!",
	       startup_file);
       xxx_exit(message);
     }
     magic_num = read_magic(fd);
     fclose(fd);
     if (magic_num == 0x11121301 || magic_num == 0x11121300)
       inst_begin = loader(startup_file,0);
     else
       xxx_exit("Incorrect startup file format");

     if (!inst_begin)
       xxx_exit("Error in loading startup file");

     if (xxx_mode == DISASSEMBLE) {
       dis(1, inst_begin);
       exit(0);
     }

     /* do it after initialization, so that typing 
	xxx -v or xxx -h won't create the ~/.xxx directory */
     set_xxxinfo_dir();

     return 0;

   } else if (calling_mode == 1) {  /* continue execution */

     return emuloop(inst_begin);

   } else if (calling_mode == 2) {  /* shutdown xsb */

#if (defined(FOREIGN) && !defined(FOREIGN_ELF) && !defined(FOREIGN_WIN32))
     if (fopen(tfile, "r")) unlink(tfile);
#endif

     if (xxx_mode != C_CALLING_XSB) {
       long cputime = cpu_time();
       realtime = real_time() - realtime;
       if (cputime < 1000.0) 
	 fprintf(stdmsg, "\nEnd XXX (cputime %ld milliseconds, elapsetime ",
		 cputime);
       else
	 fprintf(stdmsg, "\nEnd XXX (cputime %.2f seconds, elapsetime ",
		 ((float)cpu_time() / 1000));
       if (realtime < 600.0)
	 fprintf(stdmsg, "%.2f seconds)\n", realtime);
       else
	 fprintf(stdmsg, "%.2f minutes)\n", realtime/60.0);
     }
     return 0;
   }
   return 1;
}  /* end of xxx() */

/*======================================================================*/
