/* File:      error_xsb.h
** Author(s): Kostis F. Sagonas
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: error_xsb.h,v 1.7 2005/10/12 13:05:19 kostis Exp $
** 
*/

#include <setjmp.h>

/*----------------------------------------------------------------------*/
/* The following is a list of errors as defined by the Prolog ISO	*/
/* standard.  It is clear that today (October 1993), XXX does NOT	*/
/* catch all of them.  In fact it does not even report half of them.	*/
/* However, one fine day it might!  The following list tries to become	*/
/* a first step towards the uniform treatment of errors by XXX.		*/
/*----------------------------------------------------------------------*/

#define CALCULATION_ERROR	 0
#define DATABASE_ERROR		 1
#define EVALUATION_ERROR	 2
#define IMPLEMENTATION_ERROR	 3
#define INSTANTIATION_ERROR	 4
#define IO_CONTROL_ERROR	 5
#define IO_END_OF_FILE_ERROR	 6
#define IO_FORMATTING_ERROR	 7
#define OPERATOR_ERROR		 8
#define OVERFLOW_ERROR		 9 
#define RANGE_ERROR		10
#define SYNTAX_ERROR		11
#define TYPE_ERROR		12
#define UNDEFINED_PRED_ERROR	13
#define UNDEFINED_VAL_ERROR	14
#define UNDERFLOW_ERROR		15
#define ZERO_DIVIDE_ERROR	16 
#define DOMAIN_ERROR		17

/* TLS: used for determing the offset of a putpvar + call so that
   the pc register can be saved so that the proper choice point can be
   (in a necessarily roundabout manner) determined. 
*/
#define THROWPAD (putpvar_sizeof + call_sizeof)

extern int unwind_stack(void);
extern int clean_up_block(void);
extern int set_scope_marker(void);

extern void check_and_clean_table_cut(CPtr);

extern void xxx_exit(char *, ...);
extern void xxx_abort(char *, ...);
extern void xxx_bug(char *, ...);
extern void xxx_warn(char *, ...);
extern void xxx_mesg(char *, ...);
extern void xxx_error(char *, ...);
extern void xxx_dbgmsg(char *, ...);
extern void xxx_arithmetic_abort1(char *, Cell);
extern void xxx_arithmetic_abort(Cell, char *, Cell, char *);
extern void xxx_arithmetic_comp_abort(Cell, char *, int);
extern void xxx_err_handle(int, int, const char *, int, const char *, Cell);

extern FILE *stdmsg;	    	/* Stream for messages     	    */
extern FILE *stdwarn;	    	/* Stream for warnings     	    */
extern FILE *stddbg;	    	/* Stream for debugging msgs        */
extern FILE *stdfdbk;	    	/* Stream for feedback msgs         */

#define xxx_err(d, a, f, ar)	xxx_err_handle(d, a, f, ar, NULL, (Cell)NULL)

extern char *xxx_default_segfault_msg;
extern char *xxx_segfault_message; /* Put your segfault message here prior to
				      executing the command that might
				      segfault. Then restore it to
				      xxx_default_segfault_message */ 

extern void (*xxx_default_segfault_handler)(int); /* where the previous value
						     of the SIGSEGV handler is
						     saved */ 
extern jmp_buf xxx_abort_fallback_environment; /* Environment for abort
						  longjump saved here */

/* SIGSEGV handler that catches segfaults; used unless configured with DEBUG */
extern void xxx_segfault_catcher(int);

