/* File:      function.c
** Author(s): Jiyang Xu, Kostis Sagonas
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** 
** XXX is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XXX is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XXX; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: function.c,v 1.12 2003/11/23 23:35:00 kostis Exp $
** 
*/

#include "xxx_config.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "auxlry.h"
#include "cell_xsb.h"
#include "memory_xsb.h"
#include "deref.h"
#include "function.h"

/* --- returns 1 when succeeds, and returns 0 when there is an error --	*/

int unifunc_call(int funcnum, CPtr regaddr)
{
  Cell value;
  float fvalue; 

  value = cell(regaddr);
  deref(value);
  switch (funcnum) {
  case FUN_float:
    if (isinteger(value)) fvalue = (float) int_val(value);
    else if (isfloat(value)) fvalue = float_val(value);
    else return 0;
    bld_float(regaddr, fvalue);
    break;
  case FUN_floor:
    if (isinteger(value)) fvalue = (float) int_val(value);
    else if (isfloat(value)) fvalue = float_val(value);
    else return 0;
    bld_int(regaddr, floor(fvalue));
    break;
  case FUN_PLUS:
  case FUN_MINUS:
  case FUN_TIMES:
  case FUN_DIVIDE:
  case FUN_AND:
  case FUN_OR:
  case FUN_INTDIV:
    return 0;		/* should not come here */
  case FUN_sin:
    if (isinteger(value)) fvalue = (float) int_val(value);
    else if (isfloat(value)) fvalue = float_val(value);
    else return 0;
    bld_float(regaddr, sin(fvalue));
    break;
  case FUN_cos:
    if (isinteger(value)) fvalue =  (float) int_val(value);
    else if (isfloat(value)) fvalue = float_val(value);
    else return 0;
    bld_float(regaddr, cos(fvalue));
    break;
  case FUN_tan:
    if (isinteger(value)) fvalue = (float) int_val(value);
    else if (isfloat(value)) fvalue = float_val(value);
    else return 0;
    bld_float(regaddr, tan(fvalue));
    break;
  case FUN_exp:
    if (isinteger(value)) fvalue = (float) int_val(value);
    else if (isfloat(value)) fvalue = float_val(value);
    else return 0;
    bld_float(regaddr, exp(fvalue));
    break;
  case FUN_log:
    if (isinteger(value)) fvalue = (float) int_val(value);
    else if (isfloat(value)) fvalue = float_val(value);
    else return 0;
    bld_float(regaddr, log(fvalue));
    break;
  case FUN_log10:
    if (isinteger(value)) fvalue = (float) int_val(value);
    else if (isfloat(value)) fvalue = float_val(value);
    else return 0;
    bld_float(regaddr, log10(fvalue));
    break;
  case FUN_sqrt:
    if (isinteger(value)) fvalue = (float) int_val(value);
    else if (isfloat(value)) fvalue = float_val(value);
    else return 0;
    bld_float(regaddr, sqrt(fvalue));
    break;
  case FUN_asin:
    if (isinteger(value)) fvalue = (float) int_val(value);
    else if (isfloat(value)) fvalue = float_val(value);
    else return 0;
    bld_float(regaddr, asin(fvalue));
    break;
  case FUN_acos:
    if (isinteger(value)) fvalue = (float) int_val(value);
    else if (isfloat(value)) fvalue = float_val(value);
    else return 0;
    bld_float(regaddr, asin(fvalue));
    break;
  case FUN_atan:
    if (isinteger(value)) fvalue = (float) int_val(value);
    else if (isfloat(value)) fvalue = float_val(value);
    else return 0;
    bld_float(regaddr, atan(fvalue));
    break;
  case FUN_abs:
    if (isinteger(value)) {
      value = int_val(value);
      bld_int(regaddr, abs(value));
    }
    else if (isfloat(value)) {
      fvalue = float_val(value);
      bld_float(regaddr, fabs(fvalue));
    }
    else return 0;
    break;
  case FUN_truncate:
    if (isfloat(value)) {
      int i;
      fvalue = float_val(value);
#ifdef ISO_PROLOG
#ifdef _ISOC99_SOURCE
      i = trunc(fvalue);
#else
      i = (int)(fvalue >= 0 ? floor(fvalue) : ceil(fvalue));
#endif
      bld_int(regaddr, i);
#else /* not ISO_PROLOG */
#ifdef _ISOC99_SOURCE
      fvalue = (float) trunc(fvalue);
#else
      fvalue = (float) (fvalue >= 0 ? floor(fvalue) : ceil(fvalue));
#endif
      bld_float(regaddr, fvalue);
#endif
    } else {
      if (isinteger(value)) {
#ifdef ISO_PROLOG
	bld_copy(regaddr, value);
#else
        fvalue = (float) int_val(value);
        bld_float(regaddr, fvalue);
#endif
      }
    } /* else { DOMAIN_ERROR } */
    break;
  case FUN_round:
    if (isfloat(value)) {
      fvalue = float_val(value);
#ifdef ISO_PROLOG
      bld_int(regaddr, rint(fvalue));
#else
      fvalue = (float) rint(fvalue);
      bld_float(regaddr, fvalue);
#endif
    } else {
      if (isinteger(value)) {
#ifdef ISO_PROLOG
	bld_copy(regaddr, value);
#else
        fvalue = (float) int_val(value);
        bld_float(regaddr, fvalue);
#endif
      }
    } /* else { DOMAIN_ERROR } */
    break;
  case FUN_ceiling:
    if (isfloat(value)) {
      fvalue = float_val(value);
#ifdef ISO_PROLOG
      bld_int(regaddr, ceil(fvalue));
#else
      fvalue = (float) ceil(fvalue);
      bld_float(regaddr, fvalue);
#endif
    } else {
      if (isinteger(value)) {
#ifdef ISO_PROLOG
	bld_copy(regaddr, value);
#else
        fvalue = (float) int_val(value);
	bld_float(regaddr, fvalue);
#endif
      }
    } /* else { DOMAIN_ERROR } */
    break;
  default:
    return 0;
  }
  return 1;
}
