/* File:    gen_inst_defs.c
** Author:  Kostis Sagonas
** Contact: kostis@csd.uu.se
** 
** Copyright (C) Kostis Sagonas, 1999-2000
** 
** XXX is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XXX is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XXX; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: gen_inst_defs.c,v 1.14 2006/02/16 12:44:31 kostis Exp $
** 
*/


#include "xxx_config.h"

#include <stdio.h>
#include <string.h>

#include "auxlry.h"
#include "cell_xsb.h"
#include "inst_xsb.h"

/*----------------------------------------------------------------------*/

static void fgen_header(FILE *fd, char *file_name)
{
  fprintf(fd, "/* File:\t %s\n", file_name);
  fprintf(fd, "** Author:\t a program written by Kostis Sagonas\n");
  fprintf(fd, "** Warning:\t automatically generated;");
  fprintf(fd, " edit & modify at your own risk!\n*/\n\n");
}

/*----------------------------------------------------------------------*/

static int sizeof_inst(byte inst)
{ /* returns the size of an instruction in bytes */
  int i, size;

  if (inst_table[inst][1] == XXX)
    /* no bytecode instruction is present in this case */
    size = sizeof(void *);
  else {
#ifdef THREADED
    size = sizeof(void *) + 1;
#else
    size = 1;
#endif
    switch (inst_table[inst][1])
      {
      case P:
      case R:
	size += 1;
	break;
      case PP:
	size += 2;
	break;
      case PPP:
      case PPR:
      case RRR:
      case RRV:
	size += 3;
	break;
      case X:
	break;
      default:
	fprintf(stderr,
		"! First field is %ld in instruction %d... cannot handle\n",
		inst_table[inst][1], inst);
	break;
      }
  }
  for (i = 2; i <= INST_LAST_OP_NUM; i++) {
    switch (inst_table[inst][i])
      {
      case A:
      case R:
      case V:
	size++;
	break;
      case C:
      case F:
      case I:
      case L:
      case N:
      case S:
      case T:
      case UN:
      case E:
	size += sizeof(Cell);
	break;
      case X:
	break;
      default:
	fprintf(stderr,
		"! Instruction type %ld in instruction %d... cannot handle\n",
		inst_table[inst][i], inst);
	break;
      }
  }
  return size;
}

/*----------------------------------------------------------------------*/

static void fgen_insts_sizes(FILE *fd)
{
  int  inst;
  char *inst_name;

  fprintf(fd,"/* ========== Sizes of instructions in words ========== */\n");
  fprintf(fd,"\n");
  for (inst = 0; inst < INST_TABLE_SIZE; inst++) {
    if ((inst_name = (char *)inst_table[inst][0]) != NULL) {
      int size = (sizeof_inst(inst) / sizeof(Cell));
      fprintf(fd, "#define %s_wordsof  %d\n", inst_name, size);
    }
  }
  fprintf(fd,"\n");

  fprintf(fd,"/* ========== Sizes of instructions in bytes ========== */\n");
  fprintf(fd,"\n");
  for (inst = 0; inst < INST_TABLE_SIZE; inst++) {
    if ((inst_name = (char *)inst_table[inst][0]) != NULL) {
      fprintf(fd, "#define %s_sizeof  %d\n", inst_name, sizeof_inst(inst));
    }
  }
  fprintf(fd,"\n");
}

/*----------------------------------------------------------------------*/

static void handle_byte_arg(FILE *fd, int inst, const char *inst_name,
			    int i, int argno, int offset)
{
  switch (inst_table[inst][i])
    {
    case A:
    case V:
    case R:
      fprintf(fd, "#define %s_arg_%d  %d\n", inst_name, argno, offset);
      break;
    default:
      fprintf(stderr,
	      "! Byte-type %ld in instruction %s... cannot handle\n",
	      inst_table[inst][i], inst_name);
      break;
    }
}

/*----------------------------------------------------------------------*/

static void fgen_inst_arg_byte_offset(FILE *fd, int inst)
{
  int i, j, argno, offset;
  char *inst_name;
 
  if ((inst_name = (char *)inst_table[inst][0]) != NULL) {
    argno = 0; i = 1;
    if (inst_table[inst][i] == XXX)
      offset = sizeof(void *); /* offsets start immediately */
    else {
#ifdef THREADED
      offset = sizeof(void *) + 1;
#else
      offset = 1;
#endif
      switch (inst_table[inst][i])
	{
	case P:
	  offset++;
	  /* generate definitions for next two byte-size arguments */
	  for ( ; i < 3 ; ) {
	    i++;
	    argno++;
	    handle_byte_arg(fd, inst, inst_name, i, argno, offset);
	    offset++;
	  }
	  break;
	case PP:
	  offset += 2;
	  /* generate definitions for next byte-size argument */
	  i++;
	  argno++;
	  handle_byte_arg(fd, inst, inst_name, i, argno, offset);
	  offset++;
	  break;
	case PPP:
	  offset += 3;
	  break;
	case PPR:
	  offset += 2;
	  /* generate definitions for R byte-size argument */
	  argno++;
	  fprintf(fd, "#define %s_arg_%d  %d\n", inst_name, argno, offset);
	  offset++;
	  break;
	case RRR:
	case RRV:
	  /* generate definitions for all three byte-size arguments */
	  for (j = 1; j <= 3; j++) {
	    argno++;
	    fprintf(fd, "#define %s_arg_%d  %d\n", inst_name, argno, offset);
	    offset++;
	  }	
	  break;
	case X:
	  return; /* no need to continue this */
	default:
	  fprintf(stderr,
		  "! First type %ld in instruction %s... cannot handle\n",
		  inst_table[inst][i], inst_name);
	  break;
	}
    }
#ifdef BITS64
    offset += 4; /* pad64 */
#endif
    for (i++; i <= INST_LAST_OP_NUM; i++) {
      switch (inst_table[inst][i])
	{
	case V:
	case R:
	  argno++;
	  fprintf(fd, "#define %s_arg_%d  %d\n", inst_name, argno, offset);
	  offset++;
	  break;
	case C:
	case F:
	case I:
	case N:
	case L:
	case S:
	case T:
	case UN:
	case E:
	  argno++;
	  fprintf(fd, "#define %s_arg_%d  %d\n", inst_name, argno, offset);
	  offset += sizeof(Cell);
	  break;
	case X:
	  return; /* no need to continue this */
	default:
	  fprintf(stderr,
		  "! Type %ld in instruction %s... cannot handle\n",
		  inst_table[inst][i], inst_name);
	  break;
	}
    }
  }  
}

/*----------------------------------------------------------------------*/

static void fgen_inst_arg_offsets(FILE *fd)
{
  int inst;

  fprintf(fd,"/* ========== Offsets of instruction's arguments ========== */");
  fprintf(fd,"\n\n");
  for (inst = 0; inst < INST_TABLE_SIZE; inst++) {
    fgen_inst_arg_byte_offset(fd, inst);
  }
  fprintf(fd,"\n\n");
}

/*----------------------------------------------------------------------*/

static void fgen_inst_lab_tab(FILE *fd)
{
  int  inst;
  char *inst_name;

  fprintf(fd, "#if (defined(JUMP_TABLE) || defined(THREADED))\n\n");
  fprintf(fd, "static void *linstr_label_tab[] = \n{\n");
  for (inst = 0; inst < INST_TABLE_SIZE; inst++) {
    if ((inst_name = (char *)inst_table[inst][0]) != NULL)
      fprintf(fd, "\t&&%s_label,\n", inst_name);
    else fprintf(fd, "\t&&unused_instruction_label,\n");
  }
  fprintf(fd, "0/*trick*/};\n");
  fprintf(fd, "\n#endif\n");
}

/*----------------------------------------------------------------------*/

#define TMP_STR_LENGTH 	30 /* a value that should be safe */
static char inst_args[TMP_STR_LENGTH];
static char arg_list[TMP_STR_LENGTH];

static void construct_inst_args(int inst)
{
  short a, i, al;
  char  aa, r, v, s, c, l, f, n, ui, t, un, e;

  for (a = 0; a < TMP_STR_LENGTH; a++)
    inst_args[a] = arg_list[a] = '\0';
  a = al = 0;
  aa = r = v = s = c = l = f = n = ui = t = un = e = '0';
  for (i = 1; i <= INST_LAST_OP_NUM; i++) {
    switch (inst_table[inst][i])
      {
      case A:
       	inst_args[a++] = 'A';
	arg_list[al++] = 'a', arg_list[al++] = '(', arg_list[al++] = 'A';
	if (aa != '0') { inst_args[a++] = arg_list[al++] = aa; }
	arg_list[al++] = ')';
        aa++;
	break;
      case V:
	inst_args[a++] = 'V';
	arg_list[al++] = 'v', arg_list[al++] = '(', arg_list[al++] = 'V';
	if (v != '0') { inst_args[a++] = arg_list[al++] = v; }
	arg_list[al++] = ')';
	v++;
	break;
      case PPR:
	arg_list[al++] = 'p', arg_list[al++] = 'p';
	arg_list[al++] = ',', arg_list[al++] = ' ';
      case R:
	inst_args[a++] = 'R';
	arg_list[al++] = 'r', arg_list[al++] = '(', arg_list[al++] = 'R';
	if (r != '0') { inst_args[a++] = arg_list[al++] = r; }
	arg_list[al++] = ')';
	r++;
	break;
      case S:
       	inst_args[a++] = 'S';
	arg_list[al++] = 's', arg_list[al++] = '(', arg_list[al++] = 'S';
	if (s != '0') { inst_args[a++] = arg_list[al++] = s; }
	arg_list[al++] = ')';
	s++;
	break;
      case C:
       	inst_args[a++] = 'C';
	arg_list[al++] = 'c', arg_list[al++] = '(', arg_list[al++] = 'C';
	if (c != '0') { inst_args[a++] = arg_list[al++] = c; }
	arg_list[al++] = ')';
	c++;
	break;
      case L:
       	inst_args[a++] = 'L';
	arg_list[al++] = 'l', arg_list[al++] = '(', arg_list[al++] = 'L';
	if (l != '0') { inst_args[a++] = arg_list[al++] = l; }
	arg_list[al++] = ')';
	l++;
	break;
      case F:
       	inst_args[a++] = 'F';
	arg_list[al++] = 'n', arg_list[al++] = '(', arg_list[al++] = 'F';
	if (f != '0') { inst_args[a++] = arg_list[al++] = f; }
	arg_list[al++] = ')';
	f++;
	break;
      case N:
       	inst_args[a++] = 'N';
	arg_list[al++] = 'n', arg_list[al++] = '(', arg_list[al++] = 'N';
	if (n != '0') { inst_args[a++] = arg_list[al++] = n; }
	arg_list[al++] = ')';
	n++;
	break;
      case I:
       	inst_args[a++] = 'I';
	arg_list[al++] = 'i', arg_list[al++] = '(', arg_list[al++] = 'I';
	if (ui != '0') { inst_args[a++] = arg_list[al++] = ui; }
	arg_list[al++] = ')';
	ui++;
	break;
      case X: /* should never happen */
	break;
      case P:
	arg_list[al++] = 'p';
	break;
      case PP:
	arg_list[al++] = 'p', arg_list[al++] = 'p';
	break;
      case PPP:
	arg_list[al++] = 'p', arg_list[al++] = 'p', arg_list[al++] = 'p';
	break;
      case T:
       	inst_args[a++] = 'T';
	arg_list[al++] = 'i', arg_list[al++] = '(', arg_list[al++] = 'T';
	if (t != '0') { inst_args[a++] = arg_list[al++] = t; }
	arg_list[al++] = ')';
	t++;
	break;
      case RRR: {
	int j;
	for (j = 1; j <= 3; j++) {
	  inst_args[a++] = 'R';
	  arg_list[al++] = 'r', arg_list[al++] = '(', arg_list[al++] = 'R';
	  if (r != '0') { inst_args[a++] = arg_list[al++] = r; }
	  arg_list[al++] = ')';
	  if (j != 3) {
	    inst_args[a++] = arg_list[al++] = ','; arg_list[al++] = ' ';
	  }
	  r++;
	}
	break;
      }
      case RRV: {
	int j;
	for (j = 1; j <= 2; j++) {
	  inst_args[a++] = 'R';
	  arg_list[al++] = 'r', arg_list[al++] = '(', arg_list[al++] = 'R';
	  if (r != '0') { inst_args[a++] = arg_list[al++] = r; }
	  arg_list[al++] = ')';
	  inst_args[a++] = arg_list[al++] = ','; arg_list[al++] = ' ';
	  r++;
	}
	inst_args[a++] = 'V';
	arg_list[al++] = 'v', arg_list[al++] = '(', arg_list[al++] = 'V';
	if (v != '0') { inst_args[a++] = arg_list[al++] = v; }
	arg_list[al++] = ')';
	v++;
	break;
      }
      case UN:
       	inst_args[a++] = 'I';
	arg_list[al++] = 'n', arg_list[al++] = '(', arg_list[al++] = 'I';
	if (un != '0') { inst_args[a++] = arg_list[al++] = un; }
	arg_list[al++] = ')';
	un++;
	break;
      case E:
       	inst_args[a++] = 'E';
	arg_list[al++] = 'e', arg_list[al++] = '(', arg_list[al++] = 'E';
	if (e != '0') { inst_args[a++] = arg_list[al++] = e; }
	arg_list[al++] = ')';
	e++;
	break;
      default:
	fprintf(stderr,
		"! Type %ld in instruction %s... cannot handle\n",
		inst_table[inst][i], (char *)inst_table[inst][0]);
	break;
      }
    if (i == INST_LAST_OP_NUM || inst_table[inst][i+1] == X) break;
    else {
      if (inst_table[inst][i] != P &&
	  inst_table[inst][i] != PP && inst_table[inst][i] != PPP)
	inst_args[a++] = ',';
      arg_list[al++] = ',', arg_list[al++] = ' ';
    }
  }
}

/*----------------------------------------------------------------------*/

#ifdef BITS64
#define sizeof_bytecode_inst(I)		sizeof_inst(I)/2
#else
#define sizeof_bytecode_inst(I)		sizeof_inst(I)
#endif

/*----------------------------------------------------------------------*/
/* The following function should be used to return the size of the
** instruction in the .O file.  This instruction size should be
** independent of the size of the machine word, or of whether threaded
** code is enabled or not.
*/

static int sizeof_inst_in_O_file(byte inst)
{
  int size;

  size = sizeof_bytecode_inst(inst);
#ifdef THREADED
  size -= sizeof(void *);
#endif
  return size;
}

/*----------------------------------------------------------------------*/

static void fgen_inst_sizes(FILE *fd)
{
  int  inst;
  char *inst_name;

  fprintf(fd, "#ifdef THREADED\n\n");
  fprintf(fd,"/* ===== Table with instructions' sizes in .O file ===== */\n");
  fprintf(fd,"\n");
  fprintf(fd,"static unsigned char inst_sizeof_in_O_file[] =\n");
  /* Following line assumes that at least instruction with bytecode 0 exists */
  fprintf(fd,"  {%d", sizeof_inst_in_O_file(0));
  for (inst = 1; inst < INST_TABLE_SIZE; inst++) {
    /* after each 20 instructions change line */
    if (inst % 20 == 0) fprintf(fd, "\n  ");
    if ((inst_name = (char *)inst_table[inst][0]) != NULL) {
      fprintf(fd, ",%d", sizeof_inst_in_O_file(inst));
    } else fprintf(fd, ",0");
  }
  fprintf(fd,"};\n\n");
  fprintf(fd,"#endif\n");
}

/*----------------------------------------------------------------------*/

static void fgen_asm_insts(FILE *fd)
{
  int  inst, len;
  char *inst_name;

  fprintf(fd, "/* The following 3 are only used in index block */\n");
  fprintf(fd, "asm_inst(arglabel(T,I,L),\t\t 9,  -3,  [t(T), n(I), l(L)]).\n");
  fprintf(fd, "asm_inst(arglabel(T,L),\t\t\t 5,  -3,  [t(T), l(L)]).\n");
  fprintf(fd, "asm_inst(label(_T,_X),\t\t\t 0,  -2,  []).\n\n");

  /*
  ** FIRST_DYNAMIC_INST is the instruction number
  ** above which instructions are dynamically generated
  */
  for (inst = 0; inst < FIRST_DYNAMIC_INST; inst++) {
    if ((inst_name = (char *)inst_table[inst][0]) != NULL) {
      construct_inst_args(inst);
      if (inst_args[0] == '\0') { /* inst has no arguments */
	fprintf(fd, "asm_inst(%s,", inst_name);
	len = strlen(inst_name);
      }
      else {
	fprintf(fd, "asm_inst(%s(%s),", inst_name, inst_args);
	len = strlen(inst_name) + strlen(inst_args) + 2;
      }
      if (len < 6) fprintf(fd, "\t\t\t\t");
      else if (len < 14) fprintf(fd, "\t\t\t");
      else if (len < 22) fprintf(fd, "\t\t");
      else if (len < 30) fprintf(fd, "\t");
      fprintf(fd, "%2d,%4d,  [%s]).\n",
	      sizeof_inst_in_O_file(inst), inst, arg_list);
    }
  }

  fprintf(fd, "\n");
  fprintf(fd, "/* executed is an alias for the jump instruction */\n");
  fprintf(fd, "asm_inst(executed(E),\t\t\t 8,  80,  [ppp, e(E)]	).\n");
  fprintf(fd, "/* finally, the following is used for the index segment */\n");
  fprintf(fd, "asm_inst(cases(A,I,N),\t\t\t10, 255,  [a(A), n(I), n(N)]).\n");
  fprintf(fd, "\nforce_asm_inst_loaded.\n");
}

/*----------------------------------------------------------------------*/

int main(void)
{
  FILE *fd;
  char filename[20];

  init_inst_table();

  strcpy(filename, "xxx_inst_defs.h");
  fd = fopen(filename, "w");
  fgen_header(fd, filename);
  fgen_insts_sizes(fd);
  fgen_inst_arg_offsets(fd);
  fclose(fd);

  strcpy(filename, "xxx_inst_sizes.h");
  fd = fopen(filename, "w");
  fgen_header(fd, filename);
  fgen_inst_sizes(fd);
  fclose(fd);

  strcpy(filename, "xxx_inst_labels.h");
  fd = fopen(filename, "w");
  fgen_header(fd, filename);
  fgen_inst_lab_tab(fd);
  fclose(fd);

  strcpy(filename, "asm_inst.P");
  fd = fopen(filename, "w");
  fgen_header(fd, filename);
  fgen_asm_insts(fd);
  fclose(fd);

  return 0;
}

/*----------------------------------------------------------------------*/
