/* File:      inst_xsb.c
** Author(s): Warren, Xu, Swift, Sagonas, Rao
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** Copyright (C) ECRC, Germany, 1990
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: inst_xsb.c,v 1.15 2006/03/20 15:37:58 kostis Exp $
** 
*/


#include "xxx_config.h"

#include "auxlry.h"
#include "cell_xsb.h"
#include "inst_xsb.h"

#ifdef PROFILE
Cell inst_table[INST_TABLE_SIZE][INST_LAST_OP_NUM+2];

#define set_inst(inst, instr, op1type, op2type, op3type, op4type) \
        inst_table[inst][0] = (Cell)(instr); \
	inst_table[inst][1] = op1type; \
	inst_table[inst][2] = op2type; \
	inst_table[inst][3] = op3type; \
	inst_table[inst][4] = op4type; \
	inst_table[inst][5] = 0

#define profile_inst(inst) \
        inst_table[inst][5] = inst_table[inst][5] + 1

#else
Cell inst_table[INST_TABLE_SIZE][INST_LAST_OP_NUM+1];

#define set_inst(inst, instr, op1type, op2type, op3type, op4type) \
        inst_table[inst][0] = (Cell)(instr); \
	inst_table[inst][1] = op1type; \
	inst_table[inst][2] = op2type; \
	inst_table[inst][3] = op3type; \
	inst_table[inst][4] = op4type
#endif

/*----------------------------------------------------------------------*/

static void init_compiler_instructions(void)
{
  set_inst(getpvar, "getpvar",                                 P,  V, R, X);
  set_inst(getpval, "getpval",                                 P,  V, R, X);
  set_inst(getstrv, "getstrv",                                 PP, V, S, X);
  set_inst(gettval, "gettval",                                 P,  R, R, X);
  set_inst(getcon, "getcon",                                   PP, R, C, X);
  set_inst(getnil, "getnil",                                   PP, R, X, X);
  set_inst(getstr, "getstr",                                   PP, R, S, X);
  set_inst(getlist, "getlist",                                 PP, R, X, X);
  set_inst(unipvar, "unipvar",                                 PP, V, X, X);
  set_inst(unipval, "unipval",                                 PP, V, X, X);
  set_inst(unitvar, "unitvar",                                 PP, R, X, X);
  set_inst(unitval, "unitval",                                 PP, R, X, X);
  set_inst(unicon, "unicon",                                   PPP,C, X, X);
  set_inst(uninil, "uninil",                                   PPP,X, X, X);
  set_inst(getnumcon, "getnumcon",                             PP, R, N, X);
  set_inst(putnumcon, "putnumcon",                             PP, R, N, X);
  set_inst(putpvar, "putpvar",                                 P,  V, R, X);
  set_inst(putpval, "putpval",                                 P,  V, R, X);
  set_inst(puttvar, "puttvar",                                 P,  R, R, X);
  set_inst(putstrv, "putstrv",                                 PP, V, S, X);
  set_inst(putcon, "putcon",                                   PP, R, C, X);
  set_inst(putnil, "putnil",                                   PP, R, X, X);
  set_inst(putstr, "putstr",                                   PP, R, S, X);
  set_inst(putlist, "putlist",                                 PP, R, X, X);
  set_inst(bldpvar, "bldpvar",                                 PP, V, X, X);
  set_inst(bldpval, "bldpval",                                 PP, V, X, X);
  set_inst(bldtvar, "bldtvar",                                 PP, R, X, X);
  set_inst(bldtval, "bldtval",                                 PP, R, X, X);
  set_inst(bldcon, "bldcon",                                   PPP,C, X, X);
  set_inst(bldnil, "bldnil",                                   PPP,X, X, X);
  set_inst(uninumcon, "uninumcon",                             PPP,N, X, X);
  set_inst(bldnumcon, "bldnumcon",                             PPP,N, X, X);
  set_inst(getfloat, "getfloat",                               PP, R, F, X);
  set_inst(putfloat, "putfloat",                               PP, R, F, X);
  set_inst(unifloat, "unifloat",                               PPP,F, X, X);
  set_inst(bldfloat, "bldfloat",                               PPP,F, X, X);
  set_inst(unilist, "unilist",                                 PPP,X, X, X);
  set_inst(univoid, "univoid",                                 PPP,X, X, X);
  set_inst(bldvoid, "bldvoid",                                 PPP,X, X, X);
  set_inst(bld_n_voids, "bld_n_voids",                         PP, A, X, X);
  set_inst(putdval, "putdval",                                 P,  V, R, X);
  set_inst(putuval, "putuval",                                 P,  V, R, X);
  set_inst(putvoid, "putvoid",                                 PP, R, X, X);
  set_inst(getpbreg, "getpbreg",                               PP, V, X, X);
  set_inst(gettbreg, "gettbreg",                               PP, R, X, X);
  set_inst(putpbreg, "putpbreg",                               PP, V, X, X);
  set_inst(puttbreg, "puttbreg",                               PP, R, X, X);
  set_inst(jumptbreg, "jumptbreg",                             PP, R, L, X);
  set_inst(getVn, "getVn",                                     PP, V, X, X);
  set_inst(movreg, "movreg",                                   P,  R, R, X);
  set_inst(negate, "negate",                                   PP, R, X, X);
  set_inst(and, "and",                                         P,  R, R, X);
  set_inst(or, "or",                                           P,  R, R, X);
  set_inst(logshiftl, "logshiftl",                             P,  R, R, X);
  set_inst(logshiftr, "logshiftr",                             P,  R, R, X);
  set_inst(addreg, "addreg",                                   P,  R, R, X);
  set_inst(subreg, "subreg",                                   P,  R, R, X);
  set_inst(mulreg, "mulreg",                                   P,  R, R, X);
  set_inst(divreg, "divreg",                                   P,  R, R, X);
  set_inst(idivreg, "idivreg",                                 P,  R, R, X);
  set_inst(remreg, "remreg",                                   P,  R, R, X);
  set_inst(modreg, "modreg",  	   		               P,  R, R, X);
  set_inst(add_byte, "add_byte",                               P,  A, R, X);
  set_inst(sub_byte, "sub_byte",                               P,  A, R, X);
  /* Instructions obtained through instruction merging */
  set_inst(uni_tvar_tvar, "uni_tvar_tvar",                     P,  R, R, X);
  set_inst(uni_tvar_tval, "uni_tvar_tval",                     P,  R, R, X);
  set_inst(uni_tval_tvar, "uni_tval_tvar",                     P,  R, R, X);
  set_inst(uni_tval_tval, "uni_tval_tval",                     P,  R, R, X);
  set_inst(uni_n_voids, "uni_n_voids",	                       PP, A, X, X);
  set_inst(uni_void_tvar, "uni_void_tvar",                     PP, R, X, X);
  set_inst(uni_tvar_void, "uni_tvar_void",                     PP, R, X, X);
  set_inst(getlist_tval_tvar, "getlist_tval_tvar",             RRR,X, X, X);
  set_inst(getlist_tvar_tvar, "getlist_tvar_tvar",             RRR,X, X, X);
  set_inst(getlist_void_tvar, "getlist_void_tvar",             P,  R, R, X);
  set_inst(getlist_tvar_void, "getlist_tvar_void",             P,  R, R, X);
  set_inst(getlist_tval_void, "getlist_tval_void",             P,  R, R, X);
  set_inst(getlist_tvar_pvar, "getlist_tvar_pvar",             RRV,X, X, X);
  set_inst(getstr_tvar, "getstr_tvar",                         P,  R, R, S);
  set_inst(getstr_tvar_tvar, "getstr_tvar_tvar",               RRR,S, X, X);
  set_inst(getstr_tvar_tvar_tvar, "getstr_tvar_tvar_tvar",     RRR,S, X, X);
  /* Branching instructions */
  set_inst(jump, "jump",                                       PPP,L, X, X);
  set_inst(jumpz, "jumpz",                                     PP, R, L, X);
  set_inst(jumpnz, "jumpnz",                                   PP, R, L, X);
  set_inst(jumplt, "jumplt",                                   PP, R, L, X);
  set_inst(jumple, "jumple",                                   PP, R, L, X);
  set_inst(jumpgt, "jumpgt",                                   PP, R, L, X);
  set_inst(jumpge, "jumpge",                                   PP, R, L, X);
  /* Procedure instructions */
  set_inst(allocate_gc, "allocate_gc",                         P,  A, A, X);
  set_inst(call, "call",                                       PP, A, S, X);
  set_inst(calld, "calld",                                     PP, A, E, X);
  set_inst(execute, "execute",                                 PPP,S, X, X);
  set_inst(proceed, "proceed",                                 PPP,X, X, X);
  set_inst(deallocate, "deallocate",                           PPP,X, X, X);
  set_inst(dealloc_executed, "dealloc_executed",               PPP,E, X, X);
  set_inst(dealloc_execute, "dealloc_execute",                 PPP,S, X, X);
  set_inst(dealloc_proceed, "dealloc_proceed",                 PPP,X, X, X);
  /* Non-determinism instructions */
  set_inst(trymeelse, "trymeelse",                             PP, A, L, X);
  set_inst(retrymeelse, "retrymeelse",                         PP, A, L, X);
  set_inst(trustmeelsefail, "trustmeelsefail",                 PP, A, X, X);
  set_inst(try, "try",                                         PP, A, L, X);
  set_inst(retry, "retry",                                     PP, A, L, X);
  set_inst(trust, "trust",                                     PP, A, L, X);
  set_inst(trymeorelse, "trymeorelse",                         PP, A, L, X);
  set_inst(retrymeorelse, "retrymeorelse",                     PP, A, L, X);
  set_inst(trustmeorelsefail, "trustmeorelsefail",             PP, A, X, X);
  /* Non-determinism instructions */
  set_inst(switchonterm, "switchonterm",                       PPR,L, L, X);
  set_inst(switchonlist, "switchonlist",                       PPR,L, L, X);
  set_inst(hash_on_bound, "hash_on_bound",                     PPR,I, I, X);
  set_inst(hash_on_atomic, "hash_on_atomic",                   PPR,I, I, X);
  set_inst(hash_on_struct, "hash_on_struct",                   PPR,I, I, X);
  /* Miscellaneous instructions */
  set_inst(test_heap, "test_heap",                             PP, A, UN,X);
  set_inst(builtin, "builtin",                                 PP, A, X, X);
  set_inst(unifunc, "unifunc",	                               P,  A, R, X);
  set_inst(fail, "fail",                                       PPP,X, X, X);
  set_inst(halt, "halt",                                       PPP,X, X, X);
  set_inst(term_comp, "term_comp",                             RRR,X, X, X);
  set_inst(int_test_eq, "int_test_eq",                         PP, R, UN,L);
  set_inst(int_test_neq, "int_test_neq",                       PP, R, UN,L);
  set_inst(atom_test_eq, "atom_test_eq",                       PP, R, C, L);
  set_inst(atom_test_neq, "atom_test_neq",                     PP, R, C, L);

  /* Just-In-Time Indexing instructions */
  set_inst(jiti_on_bound, "jiti_on_bound",                     PPR,I, I, X);
  set_inst(jiti_on_atomic, "jiti_on_atomic",                   PPR,I, I, X);

  /* Tabling instructions */
  set_inst(check_interrupt, "check_interrupt",                 PP, A, S, X);
  set_inst(new_answer, "new_answer",                           P,  A, A, X);
  set_inst(tabletry, "tabletry",                               PP, A, L, T); 
  set_inst(tableretry, "tableretry",                           PP, A, L, X);
  set_inst(tabletrust, "tabletrust",                           PP, A, L, X); 
  set_inst(tabletrysingle, "tabletrysingle",                   PP, A, L, T);
  /*
  ** the following tabling instructions are generated dynamically but
  ** since they are related to the above ones, they appear here.
  */
  set_inst(check_complete, "check_complete",                   XXX,X, X, X); 
  set_inst(answer_return, "answer_return",                     XXX,X, X, X); 
  set_inst(resume_compl_suspension, "resume_compl_suspension", XXX,X, X, X); 
}

/* 
 * This function inits trie-instructions; they are generated dynamically.
 */
static void init_trie_instructions(void)
{
  set_inst(trie_no_cp_fail, "trie_no_cp_fail",                 PPP,X, X, X);
  set_inst(trie_try_fail, "trie_try_fail",                     PPP,X, X, X);
  set_inst(trie_retry_fail, "trie_retry_fail",                 PPP,X, X, X);
  set_inst(trie_trust_fail, "trie_trust_fail",                 PPP,X, X, X);
#ifdef ATTR_VARS
  set_inst(trie_no_cp_attv, "trie_no_cp_attv",                 PPP,X, X, X);
  set_inst(trie_try_attv, "trie_try_attv",                     PPP,X, X, X);
  set_inst(trie_retry_attv, "trie_retry_attv",                 PPP,X, X, X);
  set_inst(trie_trust_attv, "trie_trust_attv",                 PPP,X, X, X);
#endif
  set_inst(trie_no_cp_str, "trie_no_cp_str",                   PPP,X, X, X);
  set_inst(trie_try_str, "trie_try_str",                       PPP,X, X, X);
  set_inst(trie_retry_str, "trie_retry_str",                   PPP,X, X, X);
  set_inst(trie_trust_str, "trie_trust_str",                   PPP,X, X, X);
  set_inst(trie_no_cp_list, "trie_no_cp_list",                 PPP,X, X, X);
  set_inst(trie_try_list, "trie_try_list",                     PPP,X, X, X);
  set_inst(trie_retry_list, "trie_retry_list",                 PPP,X, X, X);
  set_inst(trie_trust_list, "trie_trust_list",                 PPP,X, X, X);
  set_inst(trie_no_cp_var, "trie_no_cp_var",                   PPP,X, X, X);
  set_inst(trie_try_var, "trie_try_var",                       PPP,X, X, X);
  set_inst(trie_retry_var, "trie_retry_var",                   PPP,X, X, X);
  set_inst(trie_trust_var, "trie_trust_var",                   PPP,X, X, X);
  set_inst(trie_no_cp_val, "trie_no_cp_val",                   PPP,X, X, X);
  set_inst(trie_try_val, "trie_try_val",                       PPP,X, X, X);
  set_inst(trie_retry_val, "trie_retry_val",                   PPP,X, X, X);
  set_inst(trie_trust_val, "trie_trust_val",                   PPP,X, X, X);
  set_inst(trie_no_cp_numcon, "trie_no_cp_numcon",             PPP,X, X, X);
  set_inst(trie_try_numcon, "trie_try_numcon",                 PPP,X, X, X);
  set_inst(trie_retry_numcon, "trie_retry_numcon",             PPP,X, X, X);
  set_inst(trie_trust_numcon, "trie_trust_numcon",             PPP,X, X, X);
  set_inst(trie_no_cp_numcon_succ, "trie_no_cp_numcon_succ",   PPP,X, X, X);
  set_inst(trie_try_numcon_succ, "trie_try_numcon_succ",       PPP,X, X, X);
  set_inst(trie_retry_numcon_succ, "trie_retry_numcon_succ",   PPP,X, X, X);
  set_inst(trie_trust_numcon_succ, "trie_trust_numcon_succ",   PPP,X, X, X);
  set_inst(trie_proceed, "trie_proceed",                       PPP,X, X, X);
  set_inst(trie_root, "trie_root",                             PPP,X, X, X);
  set_inst(hash_opcode, "hash_opcode",                         PPP,X, X, X);
  set_inst(hash_handle, "hash_handle",                         PPP,X, X, X);
  set_inst(trie_assert_inst, "trie_assert_inst",               PPP,X, X, X);

#if defined(CST_SUPPORT)
  set_inst(cst_continue, "cst_continue",                       PPP, X, X, X);
  set_inst(cst_return, "cst_return",                           PPP, X, X, X);
  set_inst(cst_continue_cp, "cst_continue_cp",                 PPP, X, X, X);
  set_inst(cst_return_cp, "cst_return_cp",                     PPP, X, X, X);
  set_inst(cst_remove_cp, "cst_remove_cp",                     PPP, X, X, X);
#endif
}

static void init_misc_dyn_created_instructions(void)
{
  set_inst(areg_equal, "areg_equal",                           PPR,N, L, X);
  set_inst(load_pred, "load_pred", 	                       XXX,S, X, X);
  set_inst(call_forn, "call_forn",                             XXX,L, X, X);
  set_inst(switchon3bound, "switchon3bound",                   RRR,I, I, X);
  set_inst(dyntrustmeelsefail, "dyntrustmeelsefail",           PP, A, L, X); 
  set_inst(noop, "noop",                                       PP, A, X, X);
#ifdef ATTR_VARS
  set_inst(getattv, "getattv",                                 PP, R, X, X);
  set_inst(putattv, "putattv",                                 PP, R, X, X);
#endif
#ifdef GGC
  set_inst(backtrack, "backtrack",                             XXX,X, X, X);
#endif
  set_inst(reset_bytecode, "reset",                            XXX,X, X, X);
}

/*----------------------------------------------------------------------*/

void init_inst_table(void)
{
  init_compiler_instructions();
  init_trie_instructions();
  init_misc_dyn_created_instructions();
}

/*----------------------------------------------------------------------*/
