/* File:      inst_xsb.h
** Author(s): Warren, Swift, Xu, Sagonas, Freire, Rao
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** Copyright (C) ECRC, Germany, 1990
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: inst_xsb.h,v 1.17 2006/08/23 07:58:18 kostis Exp $
** 
*/


#ifndef __INST_XXX_H__
#define __INST_XXX_H__

#ifdef THREADED
extern void **instr_label_tab;
#endif

extern void init_inst_table(void);

#define INST_LAST_OP_NUM	4

/************************************************************************/
/*	The following are operand types of instructions.		*/
/************************************************************************/

#define XXX  0  /* like PPP, but in threaded code the instruction is
		   stored in the same word as in the bytecode version */
#define A    1	/* a byte of integer (for arity, size, builtin-num, etc) */
#define V    2	/* variable offset */
#define R    3	/* register number */
#define S    4	/* structure symbol */
#define C    5	/* constant symbol - atom (tagged) */
#define L    6	/* label (address) */
#define F    7  /* floating point number */
#define N    8	/* integer number (tagged) */
#define I    9	/* 2nd & 3rd arguments of hash_on_* */
#define X   10	/* not present */
#define P   11	/* pad */
#define PP  12	/* double pad */
#define PPP 13	/* triple pad */
#define PPR 14  /* = PP + R; for switchonterm, hash_on_* and jiti_on_* */
#define T   15  /* tabletry */
#define RRR 16  /* = R + R + R */
#define RRV 17  /* = R + R + V */
#define UN  18  /* untagged number (usually integer) */
#define E   19  /* entry-point type; N-type in the compiler; L-type in emu */

/************************************************************************/
/*	Macros to fetch the instructions/operands.			*/
/************************************************************************/

#define inst_name(INSTR)	((char *)inst_table[INSTR][0])

#define cell_opcode(INSTR)	(*(byte *)(INSTR))

#define cell_operand1(INSTR)	(((byte *)(INSTR))[1])
#define cell_operand2(INSTR)	(((byte *)(INSTR))[2])
#define cell_operand3(INSTR)	(((byte *)(INSTR))[3])
#define cell_operandn(INSTR,N)	(((byte *)(INSTR))[N])


#ifdef THREADED
#define get_bytecode(INSTR)     ((CPtr)(INSTR)+1)
#else   /* IF NOT THREADED */
#define get_bytecode(INSTR)     ((CPtr)(INSTR))
#endif	/* THREADED */

#define get_opcode(INSTR)       (*(byte *)(get_bytecode(INSTR)))
#ifdef THREADED
#define set_opcode(INSTR,NEWOP) { \
	   *(CPtr)(INSTR) = (Cell)instr_label_tab[NEWOP];   \
	   *(byte *)(get_bytecode(INSTR)) = (NEWOP);        \
}
#else   /* IF NOT THREADED */
#define set_opcode(INSTR,NEWOP) \
           *(byte *)(get_bytecode(INSTR)) = (NEWOP)
#endif	/* THREADED */

#define get_cellinst(INSTR)     (*(get_bytecode(INSTR)))

#ifdef THREADED
#define set_cellinst(INSTR,VAL) {    \
   *(get_bytecode(INSTR)) = (VAL);   \
   *(CPtr)(INSTR) = (Cell)instr_label_tab[get_opcode(INSTR)]; \
}
#else	/* IF NOT THREADED */
#define set_cellinst(INSTR,VAL)      \
   *(get_bytecode(INSTR)) = (VAL)
#endif	/* THREADED */

#define inst_operand1(INSTR)	(((byte *)(get_bytecode(INSTR)))[1])
#define inst_operand2(INSTR)	(((byte *)(get_bytecode(INSTR)))[2])
#define inst_operand3(INSTR)	(((byte *)(get_bytecode(INSTR)))[3])
#define inst_operandn(INSTR,N)	(((byte *)(get_bytecode(INSTR)))[N])

/* minimum number of cells occupied by instruction's opcode */
#ifdef THREADED
#define MIN_INSTR_LEN		2 
#else	/* IF NOT THREADED */
#define MIN_INSTR_LEN		1
#endif	/* THREADED */
#define cell_after_opcode(INSTR)	((CPtr)(INSTR)+MIN_INSTR_LEN)

/* Unused right now, but may come handy in future. */

/* bit fields */
/*
#define get_last__8bits(cell)		(((Cell)(cell))&0xff)
#define get_last_16bits(cell)		(((Cell)(cell))&0xffff)
#define get_last_32bits(cell)		(((Cell)(cell))&0xffffffff)

#define get_first__8bits(cell)		(((Cell)(cell))>>(sizeof(Cell)*8-8))
#define get_first_16bits(cell)		(((Cell)(cell))>>(sizeof(Cell)*8-16))
#define get_first_32bits(cell)		(((Cell)(cell))>>(sizeof(Cell)*8-32))

#define get__8bit_field(cell, pos)	\
		(((Cell)(cell))>>((sizeof(Cell)-pos-1)*8)&0xff)
#define get_16bit_field(cell, pos)	\
		(((Cell)(cell))>>((sizeof(Cell)-pos-1)*16)&0xffff)
#define get_32bit_field(cell, pos)	\
		(((Cell)(cell))>>((sizeof(Cell)-pos-1)*32)&0xffffffff)
*/

/* opcode & operand fields */

/*
#ifdef BITS64
#define opquarter0(cell)		(get_first_16bits(cell))
#define opquarter1(cell)		(get_16bit_field(cell,1))
#define opquarter2(cell)		(get_16bit_field(cell,2))
#define opquarter3(cell)		(get_last_16bits(cell))
#define opquartern(cell,N)		(get_16bit_field(cell,N))
#else
#define opquarter0(cell)		(get_first__8bits(cell))
#define opquarter1(cell)		(get__8bit_field(cell,1))
#define opquarter2(cell)		(get__8bit_field(cell,2))
#define opquarter3(cell)		(get_last__8bits(cell))
#define opquartern(cell,N)		(get__8bit_field(cell,N))
#endif
#define opcell(cell)			((Cell)(cell))
#define opint(cell)			((Integer)(cell))

#define opcode(cell)			((short)opquarter0(cell))

#define opregaddr1(cell)		(rreg+opquarter1(cell))
#define opregaddr2(cell)		(rreg+opquarter2(cell))
#define opregaddr3(cell)		(rreg+opquarter3(cell))
#define opregaddrn(cell,N)		(rreg+opquartern(cell,N))

#define opreg1(cell)			(cell(opregaddr1(cell)))
#define opreg2(cell)			(cell(opregaddr2(cell)))
#define opreg3(cell)			(cell(opregaddr3(cell)))
#define opregn(cell)			(cell(opregaddrn(cell,N)))

#define operand1(cell)			((short)opquarter1(cell))
#define operand2(cell)			((short)opquarter2(cell))
#define operand3(cell)			((short)opquarter3(cell))
#define operandn(cell,N)		((short)opquartern(cell,N))
*/


/*
#define opvaraddr(cell)			(ereg+(-(Cell)(cell)))
*/
/* 64 bit could benefit from change in compiler/loader for the V operands */

/*
#define opvaraddr1(cell)	(ereg+(-(unsigned int)(opquarter1(cell))))
#define opvaraddr2(cell)	(ereg+(-(unsigned int)(opquarter2(cell))))
#define opvaraddr3(cell)	(ereg+(-(unsigned int)(opquarter3(cell))))
#define opvaraddrn(cell,N)	(ereg+(-(unsigned int)(opquartern(cell,n))))

#define opvar1(cell)			(cell(opvaraddr1(cell)))
#define opvar2(cell)			(cell(opvaraddr2(cell)))
#define opvar3(cell)			(cell(opvaraddr3(cell)))
#define opvarn(cell,N)			(cell(opvaraddrn(cell,N)))
*/

/************************************************************************/
/*	The following is the set of all instructions.			*/
/************************************************************************/

/*----- Compiler instructions ----*/
#define getpvar                 0x00
#define getpval                 0x01
#define getstrv                 0x02
#define gettval                 0x03
#define getcon                  0x04
#define getnil                  0x05
#define getstr                  0x06
#define getlist                 0x07
#define unipvar                 0x08
#define unipval                 0x09
#define unitvar                 0x0a
#define unitval                 0x0b
#define unicon                  0x0c
#define uninil                  0x0d
#define getnumcon	        0x0e
#define putnumcon	        0x0f
#define putpvar                 0x10
#define putpval                 0x11
#define puttvar                 0x12
#define putstrv                 0x13
#define putcon                  0x14
#define putnil                  0x15
#define putstr                  0x16
#define putlist                 0x17
#define bldpvar                 0x18
#define bldpval                 0x19
#define bldtvar                 0x1a
#define bldtval                 0x1b
#define bldcon                  0x1c
#define bldnil                  0x1d
#define uninumcon	        0x1e
#define bldnumcon	        0x1f
#define getfloat	        0x20
#define putfloat	        0x21
#define unifloat	        0x22
#define bldfloat		0x23
#define unilist			0x24
#define univoid			0x25
#define bldvoid			0x26
#define bld_n_voids		0x27
/*----- Unsafe term instructions -*/
#define putdval			0x28
#define putuval			0x29
#define putvoid			0x2a
/*----- Instructions for cut -----*/
#define getpbreg		0x2b
#define gettbreg		0x2c
#define putpbreg	        0x2d
#define puttbreg	        0x2e
#define jumptbreg	        0x2f	/* dynamically gener; used in assert */
/*----- For tabled predicates ----*/
#define getVn	                0x30
/*----- Arithmetic instructions --*/
#define movreg			0x31
#define negate		        0x32
#define and 		        0x33
#define or 		        0x34
#define logshiftl	        0x35
#define logshiftr	        0x36
#define addreg		        0x37
#define subreg		        0x38
#define mulreg		        0x39
#define divreg		        0x3a
#define idivreg		        0x3b
#define remreg		        0x3c
#define modreg		        0x3d
#define add_byte	        0x3e
#define sub_byte	        0x3f
/*----- Merged instructions ------*/
#define uni_tvar_tvar           0x40
#define uni_tvar_tval           0x41
#define uni_tval_tvar           0x42
#define uni_tval_tval           0x43
#define uni_n_voids		0x44
#define uni_void_tvar		0x45
#define uni_tvar_void		0x46
#define getlist_tval_tvar	0x47
#define getlist_tvar_tvar	0x48
#define getlist_void_tvar	0x49
#define getlist_tvar_void	0x4a
#define getlist_tval_void	0x4b
#define getlist_tvar_pvar       0x4c
#define getstr_tvar		0x4d
#define getstr_tvar_tvar       	0x4e
#define getstr_tvar_tvar_tvar 	0x4f
/*----- Branching instructions ---*/
#define jump                    0x50
#define jumpz                   0x51
#define jumpnz                  0x52
#define jumplt                  0x53
#define jumple                  0x54
#define jumpgt                  0x55
#define jumpge                  0x56
/*----- Procedure instructions ---*/
#define allocate_gc 	 	0x57
#define call 	 	 	0x58
#define calld  		 	0x59
#define execute 	 	0x5a
#define proceed 	 	0x5b
#define deallocate 	 	0x5c
#define dealloc_executed 	0x5d
#define dealloc_execute 	0x5e
#define dealloc_proceed 	0x5f
/*----- Non-determinism instructions */
#define trymeelse 	 	0x60
#define retrymeelse 	 	0x61
#define trustmeelsefail 	0x62
#define try 	 	 	0x63
#define retry 	 		0x64
#define trust 	 		0x65
#define trymeorelse		0x66
#define retrymeorelse		0x67
#define trustmeorelsefail	0x68
/*----- Indexing instructions ----*/
#define switchonterm		0x69
#define switchonlist		0x6a
#define hash_on_bound		0x6b
#define hash_on_atomic		0x6c
#define hash_on_struct		0x6d
/*----- Miscellaneous instructions */
#define test_heap		0x6e
#define builtin			0x6f
#define unifunc		        0x70
#define fail			0x71
#define halt			0x72
/*----- Term comparison instructions */
#define term_comp		0x73
#define int_test_eq		0x74
#define int_test_neq		0x75
#define atom_test_eq		0x76
#define atom_test_neq		0x77

/*----- JIT Indexing instructions ----*/
#define jiti_on_bound		0x7f
#define jiti_on_atomic		0x80

/*----- Tabling instructions -----*/
#define check_interrupt		0x87
#define new_answer		0x88
#define tabletry		0x89
#define tableretry		0x8a
#define tabletrust		0x8b
#define tabletrysingle		0x8c

/*----- From here on, instructions are generated dynamically ----------*/

#define check_complete		0x8d	/* generated dynamically ! */
#define answer_return		0x8e	/* generated dynamically ! */
#define resume_compl_suspension 0x8f	/* generated dynamically ! */

/*----- Instructions for tries as code (Do NOT change the numbers) -----*/

#define trie_no_cp_str		0x90	/* Structs */
#define trie_trust_str		0x91
#define trie_try_str		0x92
#define trie_retry_str		0x93
#define trie_no_cp_list		0x94	/* Lists */
#define trie_trust_list		0x95
#define trie_try_list		0x96
#define trie_retry_list		0x97
#define trie_no_cp_var		0x98	/* Vars */
#define trie_trust_var		0x99
#define trie_try_var		0x9a
#define trie_retry_var		0x9b
#define trie_no_cp_val		0x9c	/* Vals */
#define trie_trust_val		0x9d
#define trie_try_val		0x9e
#define trie_retry_val		0x9f
#define trie_no_cp_numcon	0xa0	/* Constants */
#define trie_trust_numcon	0xa1
#define trie_try_numcon		0xa2
#define trie_retry_numcon	0xa3
#define trie_no_cp_numcon_succ	0xa4	/* Constants in trie leaves */
#define trie_trust_numcon_succ	0xa5
#define trie_try_numcon_succ	0xa6
#define trie_retry_numcon_succ	0xa7
#define trie_no_cp_fail         0xa8	/* to reclaim deleted returns */
#define trie_trust_fail         0xa9
#define trie_try_fail           0xaa
#define trie_retry_fail         0xab
#define trie_proceed		0xac	/* Just a leaf */
#define trie_root		0xad
#define hash_opcode 		0xae
#define hash_handle 		0xaf
#ifdef ATTR_VARS
#define trie_no_cp_attv		0xb0
#define trie_trust_attv		0xb1
#define trie_try_attv		0xb2
#define trie_retry_attv		0xb3
#endif
#define trie_assert_inst	0xb4

/*----------------------------------------------------------------------*/

#define areg_equal              0xb5	/* generated by the loader */
#define load_pred		0xb6
#define call_forn               0xb7
#define switchon3bound	        0xb8	/* for multi-argument indexing */
#define dyntrustmeelsefail	0xb9	/* dynamic trust instruction */
#define noop                    0xba	/* used in retract */
/*---- Miscellaneous instructions depending on the configuration */
#ifdef ATTR_VARS
#define getattv		        0xbb	/* used in assert ? */
#define putattv		        0xbc	/* NOT used ? */
#endif
#ifdef GGC
#define backtrack               0xbe	/* needs to be taken out */
#endif
#if defined(CST_SUPPORT)
#define cst_continue            0xc0
#define cst_return              0xc1
#define cst_continue_cp         0xc3
#define cst_return_cp           0xc4
#define cst_remove_cp           0xc5
#endif
/* NOTE: The following should be the last instruction;
 * the size of the inst table depends on it; see below */
#define reset_bytecode		0xc6

/*					Cannot be used:
#define cases			0xff	reserved in .O files for indexing
*/
/*----------------------------------------------------------------------*/

#define FIRST_DYNAMIC_INST	check_complete
#define INST_TABLE_SIZE		reset_bytecode+1
#define BUILTIN_TABLE_SIZE	256

#ifdef PROFILE
extern Cell inst_table[INST_TABLE_SIZE][6];
#else
extern Cell inst_table[INST_TABLE_SIZE][5];
#endif

extern Cell builtin_table[BUILTIN_TABLE_SIZE][2];

#endif
