
#include "xxx_config.h"

#include <stdio.h>

#include "auxlry.h"
#include "psc_xsb.h"
#include "psc_defs.h"
#include "cell_xsb.h"
#include "heap_xsb.h"
#include "flags_xsb.h"
#include "macro_xsb.h"
#include "loader_xsb.h"
#include "loader_defs.h"
#include "error_xsb.h"
#include "inst_xsb.h"
#include "xxx_inst_defs.h"
#include "xxx_inst_sizes.h"
#include "memory_xsb.h"
#include "register.h"
#include "varstring.h"
#include "dis.h"

#define TRY_RETRY_TRUST_SIZE try_sizeof

#include "jiti.h"

/* #define DEBUG_JITI */

#ifdef DEBUG_JITI
#define DBG_INST(msg,x)   fprintf(stderr,"%s - INST %p %s\n",msg,x,(char *)inst_table[get_opcode(x)][0]);
#else
#define DBG_INST(msg,x)
#endif

/*-----------------------------------------------------------------------*/

void fprint_hashtable(FILE *pFile, const byte *ptrInitHTable, Integer iSize)
{
  CPtr    ptrInstAddr;
  Integer iIndex = 0;
  CPtr    ptrHTable = (CPtr)ptrInitHTable;

  fprintf(pFile, "PRINT HASHTABLE ADDRESS(%p) SIZE(%d)\n",
	  (void*)ptrInitHTable,iSize);

  /* iterate all entries */
  for(iIndex=0;iIndex<iSize;iIndex++) 
  {
    if (*(ptrHTable)==(Cell)&fail_inst)
    {
      fprintf(pFile, "%d %p:    &fail\n",iIndex,(void*)ptrHTable);
    }
    else
    {
      fprintf(pFile, "%d %p:  0x%lx\n",iIndex,(void*)ptrHTable,*(ptrHTable)); 
      /* check if it's a bucket with more entries */
      ptrInstAddr = (CPtr)*ptrHTable;

      if((get_opcode(ptrInstAddr)==try) ||
 	  (get_opcode(ptrInstAddr)== areg_equal) )
     {
        fprintf(pFile, "      BUCKET %p\n", (void*)ptrHTable);
	while ((get_opcode(ptrInstAddr)!=trust) &&
		(get_opcode(ptrInstAddr)!=fail) )
	{
          fprintf(pFile, "      ");
	  ptrInstAddr = (CPtr)fprint_inst(pFile, (byte *)ptrInstAddr);
	}
      }
        fprintf(pFile, "      ");
      fprint_inst(pFile, (byte *)ptrInstAddr);

    }
    ptrHTable++;
  }
  fprintf(pFile, "END PRINT HASHTABLE ADDRESS(%p-%p) SIZE(%d)\n",
	  (void*)ptrInitHTable,(void*)ptrHTable,iSize);
}

/*-----------------------------------------------------------------------*/

/************************************************************************/
/*									*/
/* 									*/
/* 									*/
/* 									*/
/* 									*/
/*									*/
/************************************************************************/

byte * jiti_create_hash_table(CPtr ptrJitiInstruction,
			      Integer iArg, 
			      Integer iArity, 
			      Integer iClauseNo)
{
  byte * pResult;
  byte * pInstAddress;
  byte * pClauseAddr;
  Cell iValue;
  byte opCode;
  struct hash_space * pHashSpace;

  Integer iClauseIndex;
  pInstAddress = (byte *)ptrJitiInstruction;
#ifdef DEBUG_JITI
  fprintf(stderr,"JITI_CREATE_HASH_TABLE %p Arg=%d Arity=%d ClauseNo=%d\n",(void *)pInstAddress,iArg,iArity,iClauseNo);
#endif	
	
  /* create a hash table to build the index */
  pHashSpace = alloc_hspace(iClauseNo);
     
#ifdef DEBUG_JITI
  fprint_inst(stderr, (byte *)pInstAddress);	  
#endif
  
  /* skip hash_ or jiti_ instructions */
  /* it's assumed that jiti_on_atomic and hash_on_atomic have the same size */
  pInstAddress += hash_on_atomic_sizeof; 
  opCode=get_opcode(pInstAddress);
  while((opCode==jiti_on_atomic)||(opCode==hash_on_atomic))
  {
    pInstAddress+=hash_on_atomic_sizeof;
    opCode=get_opcode(pInstAddress);
  }	
  /* skip section try-retry-trust */
  pInstAddress += (iClauseNo*try_sizeof);

#ifdef DEBUG_JITI
  DBG_INST("try-retry-trust section jumped",pInstAddress)
#endif

  /* NOTE: the following code assumes that there is a instruction per parameter */
  /* DECODE INSTRUCTIONS */
  /* iterate clauses (in memory) to get the  n-parameter of clause */
  iClauseIndex = 0;
  /* get group address */
  pClauseAddr = pInstAddress;

#ifdef DEBUG_JITI
  fprintf(stderr,"BLOCK ADDRESS %d %p\n",iClauseIndex,(void *)pClauseAddr);
#endif
  while(iClauseIndex<iClauseNo)
  {
    Integer iRegNum;
    /* parse instruction, search instructions with get Register-(iArg) */
    opCode=get_opcode(pInstAddress);

#ifdef DEBUG_JITI
    fprint_inst(stderr,pInstAddress); 
#endif
    
    if((opCode==getnumcon)||(opCode==getcon))
    {
      iRegNum=*(pInstAddress+getnumcon_arg_1);
      iValue=*(CPtr)(pInstAddress+getnumcon_arg_2);
      if(iRegNum==iArg)
      {
	/* insert value and address in hash table */
	inserth(pHashSpace,(CPtr)pClauseAddr, iValue);
#ifdef DEBUG_JITI
	fprintf(stderr,"MATCH GET_INST %p %d %ld\n",(void *)pClauseAddr,iRegNum,iValue); 
#endif
        }
    }
    else if(opCode==proceed)
    {
      /* next block */
      pClauseAddr = pInstAddress+proceed_sizeof;	
      iClauseIndex++;
#ifdef DEBUG_JITI
      fprintf(stderr,"BLOCK ADDRESS ClauseIndex=%d ClauseNo=%d %p\n",iClauseIndex,iClauseNo,(void *)pClauseAddr); 
#endif
    }
    /* next instruction */
    pInstAddress += inst_sizeof_in_O_file[opCode]; /*size instruction */
#ifdef THREADED
    pInstAddress += 4;
#endif
  }

  /* generate code to the hash table */
  gen_index(iClauseNo, (void*)(&pResult), iArity, pHashSpace);

#ifdef DEBUG_JITI
  fprint_hashtable(stderr,pResult,pHashSpace->size);
#endif
  /* release resources */
  free_hspace(pHashSpace);
  return pResult;
}



