/* File:      macro_xsb.h
** Author(s): Swift, Sagonas, Rao, Freire
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: macro_xsb.h,v 1.8 2005/10/13 08:20:12 kostis Exp $
** 
*/

#ifndef __MACRO_XSB_H__
#define __MACRO_XSB_H__

/*
 * Need to be here due to the include of "trie_internals.h".
 */
typedef struct subgoal_frame *SGFrame;
typedef struct completion_stack_frame *ComplStackFrame;

#include "trie_internals.h"
#include "table_status_defs.h"

/*===========================================================================*/

/*
 *                         Table Information Frame
 *                         =======================
 *
 *  Table Information Frames are created for each tabled predicate,
 *  allowing access to its calls and their associated answers.
 */

typedef struct Table_Info_Frame *TIFptr;
typedef struct Table_Info_Frame {
  TIFptr next_tif;	/* pointer to next table info frame */
  BTNptr call_trie;	/* pointer to the root of the call trie */
  Psc  psc_ptr;		/* pointer to the PSC record of the subgoal */
} TableInfoFrame;

#define TIF_NextTIF(pTIF)	   ( (pTIF)->next_tif )
#define TIF_CallTrie(pTIF)	   ( (pTIF)->call_trie )
#define TIF_PSC(pTIF)		   ( (pTIF)->psc_ptr )

#define New_TIF(pTIF,pPSC) {						\
   pTIF = malloc(sizeof(TableInfoFrame));				\
   if ( IsNULL(pTIF) )							\
     xxx_abort("Ran out of memory in allocation of TableInfoFrame");	\
   TIF_NextTIF(pTIF) = NULL;						\
   TIF_CallTrie(pTIF) = NULL;						\
   TIF_PSC(pTIF) = pPSC;						\
 }

/*----------------------------------------------------------------------*/
/*  Completion Stack Structure (ASCC node structure).			*/
/*----------------------------------------------------------------------*/

struct completion_stack_frame {
  SGFrame subgoal_ptr;
  int     level_num;
  CPtr    hreg;	          /* for accessing the substitution factor */
  CPtr    pdreg;
  CPtr    ptcp;
  CPtr    cons_copy_list; /* Pointer to a list of consumer copy frames */
  ALNptr  del_ret_list;   /* to reclaim deleted returns */
};

#define COMPLFRAMESIZE	(sizeof(struct completion_stack_frame)/sizeof(CPtr))

#define compl_subgoal_ptr(b)	((ComplStackFrame)(b))->subgoal_ptr
#define compl_level(b)		((ComplStackFrame)(b))->level_num
#define compl_hreg(b)		((ComplStackFrame)(b))->hreg
#define compl_pdreg(b)		((ComplStackFrame)(b))->pdreg
#define compl_ptcp(b)		((ComplStackFrame)(b))->ptcp
#define compl_cons_copy_list(b)	((ComplStackFrame)(b))->cons_copy_list
#define compl_del_ret_list(b)	((ComplStackFrame)(b))->del_ret_list

#define prev_compl_frame(b)	(((ComplStackFrame)(b))+1)
#define next_compl_frame(b)	(((ComplStackFrame)(b))-1)

#define prev_CSF(b)		((b)++)
#define next_CSF(b)		((b)--)

/*
 *  The overflow test MUST be placed after the initialization of the
 *  ComplStackFrame in the current implementation.  This is so that the
 *  corresponding subgoal which points to this frame can be found and its
 *  link can be updated if an expansion is required.  This was the simplest
 *  solution to not leaving any dangling pointers to the old area.
 */
#define	push_completion_frame(subgoal)	\
  level_num++; \
  next_CSF(openreg); \
  compl_subgoal_ptr(openreg) = subgoal; \
  compl_level(openreg) = level_num; \
  compl_hreg(openreg) = hreg - 1; /* so that it points to something useful */ \
  compl_pdreg(openreg) = delayreg; \
  compl_ptcp(openreg) = ptcpreg; \
  compl_cons_copy_list(openreg) = NULL; \
  compl_del_ret_list(openreg) = NULL; \
  check_completion_stack_overflow

/*----------------------------------------------------------------------*/

#include "cst.h"
#include "slgdelay.h"

/*----------------------------------------------------------------------*/
/*  Subgoal (Call) Structure.						*/
/*----------------------------------------------------------------------*/

     /* should not change the order - unless tables.P is updated accordingly
      * also, if adding fields, add to the end! 
      */
struct subgoal_frame {
  SGFrame next_subgoal;
  BTNptr  ans_root_ptr;	/* Root of the return trie */
  TIFptr  tif_ptr;	/* Used for residual program and for debugging */
  ComplStackFrame compl_stack_ptr;	/* Pointer to subgoal's completion stack frame */
  CPtr compl_suspens_ptr; /* pointer to CHAT area; type is chat_init_pheader */
  ALNptr  ans_list_ptr;	/* Pointer to the list of returns in the ret trie */
  SGFrame prev_subgoal;
  BTNptr  leaf_ptr;	/* Used in deletion and for residual programs */
  CPtr    cp_ptr;       /* Pointer to the Generator CP */
  ALNptr ans_list_tail; /* pointer to the tail of the answer list */
  CPtr    compl_flag;   /* indicates whether subgoal is completed */
  PNDE    nde;		/* pointer to a negative DE */
#if defined(CST_SUPPORT)
  CST     constraints;  /* constraints for this call */
#endif
};

#define CALLSTRUCTSIZE	(sizeof(struct subgoal_frame)/sizeof(CPtr))

#define subg_next_subgoal(SUBG)		((SGFrame)(SUBG))->next_subgoal
#define subg_prev_subgoal(SUBG)		((SGFrame)(SUBG))->prev_subgoal
#define subg_ans_root_ptr(SUBG)		((SGFrame)(SUBG))->ans_root_ptr
#define subg_tif_ptr(SUBG)		((SGFrame)(SUBG))->tif_ptr
#define subg_leaf_ptr(SUBG)		((SGFrame)(SUBG))->leaf_ptr
/* use this for mark as completed == 0 */
#define subg_compl_stack_ptr(SUBG)	((SGFrame)(SUBG))->compl_stack_ptr
#define subg_compl_susp_ptr(SUBG)	((SGFrame)(SUBG))->compl_suspens_ptr
#define subg_ans_list_ptr(SUBG)		((SGFrame)(SUBG))->ans_list_ptr
#define subg_cp_ptr(SUBG)		((SGFrame)(SUBG))->cp_ptr
#define subg_ans_list_tail(SUBG)	((SGFrame)(SUBG))->ans_list_tail
#define subg_compl_flag(SUBG)		((SGFrame)(SUBG))->compl_flag
#define subg_nde(SUBG)			((SGFrame)(SUBG))->nde
#if defined(CST_SUPPORT)
#define subg_cst(SUBG)                  ((SGFrame)(SUBG))->constraints
#endif

/* beginning of REAL answers in the answer list */
#define subg_answers(SUBG)		aln_next_aln(subg_ans_list_ptr(SUBG))

#define subg_in_cp_stack(SUBG)		((Integer)subg_cp_ptr(SUBG) != 0)

extern ALNptr empty_return(void);


/* Appending to the Answer List of a SF
   ------------------------------------ */
#define SF_AppendNewAnswerList(pSF,pAnsList) {	\
						\
   ALNptr pLast;				\
						\
   pLast = pAnsList;				\
   while ( IsNonNULL(ALN_Next(pLast)) )		\
     pLast = ALN_Next(pLast);			\
   SF_AppendToAnswerList(pSF,pAnsList,pLast);	\
 }

#define SF_AppendNewAnswer(pSF,pAns)	SF_AppendToAnswerList(pSF,pAns,pAns)

#define SF_AppendToAnswerList(pSF,pHead,pTail) {			\
   if ( has_answers(pSF) )						\
     /*
      *  Insert new answer at the end of the answer list.
      */								\
     ALN_Next(subg_ans_list_tail(pSF)) = pHead; 			\
   else									\
     /*
      * The dummy answer list node is the only node currently in the list.
      * It's pointed to by the head ptr, but the tail ptr is NULL.
      */								\
     ALN_Next(subg_ans_list_ptr(pSF)) = pHead;				\
   subg_ans_list_tail(pSF) = pTail;					\
 }


/* Global Structure Management
   --------------------------- */
#define SUBGOAL_FRAMES_PER_BLOCK    16

extern struct Structure_Manager smSF;


/* Subgoal Frames (De)Allocation
   ----------------------------- */
/*
 * Allocated Subgoal Frames are maintained on a list to facilitate
 * deallocation of specific calls and answer sets.
 */
#define SF_AddNewToAllocList(SF)	\
   SM_AddToAllocList_DL(smSF,SF,subg_prev_subgoal,subg_next_subgoal)

#define SF_RemoveFromAllocList(SF)	\
   SM_RemoveFromAllocList_DL(smSF,SF,subg_prev_subgoal,subg_next_subgoal)

/*
 * Allocates and initializes a subgoal frame for a producer subgoal.  It
 * is inserted at the head of the global subgoal list, located in the
 * structure manager, smSF.  The TIP field is initialized, fields of the
 * call trie leaf and this subgoal are set to point to one another, while
 * the completion stack frame pointer is set to the next available
 * location (frame) on the stack (but the space is not yet allocated from
 * the stack).  Also, an answer-list node is allocated for pointing to a
 * dummy answer node and inserted into the answer list.  Note that answer
 * sets (answer tries, roots) are lazily created -- not until an answer
 * is generated.  Therefore this field may potentially be NULL, as it is
 * initialized here.  memset() is used so that the remaining fields are
 * initialized to 0/NULL so, in some sense making this macro independent
 * of the number of fields.
 */

#if defined(CST_SUPPORT)
#define NewProducerSF(SF,CSTLeaf,Leaf,TableInfo) {              \
                                                                \
   SGFrame pNewSF;                                              \
                                                                \
   SM_AllocateStruct(smSF,pNewSF);                              \
   pNewSF = memset(pNewSF,0,sizeof(struct subgoal_frame));      \
   subg_tif_ptr(pNewSF) = TableInfo;                            \
   subg_leaf_ptr(pNewSF) = Leaf;                                \
   subg_cst(pNewSF) = CSTLeaf;                                  \
   CST_set_SGF(CSTLeaf, pNewSF);                                \
   subg_ans_list_ptr(pNewSF) = empty_return();                  \
   subg_compl_stack_ptr(pNewSF) = next_compl_frame(openreg);    \
   SF_AddNewToAllocList(pNewSF);                                \
   SF = pNewSF;                                                 \
}
#else
#define NewProducerSF(SF,Leaf,TableInfo) {			\
								\
   SGFrame pNewSF;						\
								\
   SM_AllocateStruct(smSF,pNewSF);				\
   pNewSF = memset(pNewSF,0,sizeof(struct subgoal_frame));	\
   subg_tif_ptr(pNewSF) = TableInfo;				\
   subg_leaf_ptr(pNewSF) = Leaf;				\
   CallTrieLeaf_SetSF(Leaf,pNewSF);				\
   subg_ans_list_ptr(pNewSF) = empty_return();			\
   subg_compl_stack_ptr(pNewSF) = next_compl_frame(openreg);	\
   SF_AddNewToAllocList(pNewSF);				\
   SF = pNewSF;							\
}
#endif

#define FreeProducerSF(SF)			\
   SF_RemoveFromAllocList(SF);			\
   SM_DeallocateStruct(smSF,SF)

/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/

#define tab_level(SUBG_PTR)     \
        compl_level((subg_compl_stack_ptr(SUBG_PTR)))
#define next_tab_level(CSF_PTR) \
        compl_level(prev_compl_frame(CSF_PTR))

#define is_leader(CSF_PTR)	\
	((prev_compl_frame(CSF_PTR) >= COMPLSTACKBOTTOM) || \
         (next_tab_level(CSF_PTR) < compl_level(CSF_PTR)))

/*----------------------------------------------------------------------*/
/* Codes for completed subgoals (assigned to subg_answers)              */
/*----------------------------------------------------------------------*/

#define NO_ANSWERS	(ALNptr)0
#define UNCOND_ANSWERS	(ALNptr)1
#define COND_ANSWERS	(ALNptr)2

/*----------------------------------------------------------------------*/
/* The following 2 macros are to be used for incomplete subgoals.	*/
/*----------------------------------------------------------------------*/

#define has_answers(SUBG_PTR)	    IsNonNULL(subg_answers(SUBG_PTR))
#define has_no_answers(SUBG_PTR)    IsNULL(subg_answers(SUBG_PTR))

/*----------------------------------------------------------------------*/
/* The following 5 macros should be used only for completed subgoals.	*/
/*----------------------------------------------------------------------*/

/*
 * These defs depend on when the root of an answer trie is created.
 * Currently, this is when the first answer is added to the set.  They
 * are also dependent upon representation of the truth of ground goals.
 * Currently, true subgoals have an ESCAPE node placed below the root,
 * while false goals have no root nor leaves (since no answer was ever
 * inserted, there was no opportunity to create a root).
 */
#define has_answer_code(SUBG_PTR)				\
	( IsNonNULL(subg_ans_root_ptr(SUBG_PTR)) )

#define subgoal_fails(SUBG_PTR)			\
	( ! has_answer_code(SUBG_PTR) )

/* should only be used on ground subgoals (is for escape node inspection) */
#if defined(CST_SUPPORT)
#define subgoal_unconditionally_succeeds(SUBG_PTR)			    \
        ( has_answer_code(SUBG_PTR) &&					    \
	  is_unconditional_answer(BTN_CST(                                  \
				      BTN_Child(                            \
					  subg_ans_root_ptr(SUBG_PTR)))) )
#else
#define subgoal_unconditionally_succeeds(SUBG_PTR)			    \
        ( has_answer_code(SUBG_PTR) &&					    \
	  is_unconditional_answer(BTN_Child(subg_ans_root_ptr(SUBG_PTR))) )
#endif

#define mark_subgoal_failed(SUBG_PTR)	\
	(subg_ans_root_ptr(SUBG_PTR) = NULL)

#define neg_simplif_possible(SUBG_PTR)	\
	((subgoal_fails(SUBG_PTR)) && (subg_nde(SUBG_PTR) != NULL))

/*----------------------------------------------------------------------*/

/* #ifndef DEBUG */
#define OLD
/* #endif */

#ifdef OLD
#define is_completed(SUBG_PTR)	((Integer) subg_compl_flag(SUBG_PTR) < 0)

#define mark_as_completed(SUBG_PTR) \
        if (!is_completed(SUBG_PTR)) {					\
          /* the following is used to maintain invariants; */		\
          /* ideally the completion stack should be compacted */	\
          /* and completed subgoals should be removed instead */	\
          compl_pdreg(subg_compl_stack_ptr(SUBG_PTR)) = NULL;		\
          reclaim_del_ret_list(SUBG_PTR);				\
          if (neg_simplif_possible(SUBG_PTR))				\
            simplify_neg_fails(SUBG_PTR);				\
          subg_compl_flag(SUBG_PTR) = (CPtr) -1;			\
        } 
#else
#define is_completed(SUBG_PTR)	((Integer)subg_cp_ptr(SUBG_PTR) < 0)

#define mark_as_completed(SUBG_PTR) \
        if (!is_completed(SUBG_PTR)) {					\
          /* the following is used to maintain invariants; */		\
          /* ideally the completion stack should be compacted */	\
          /* and completed subgoals should be removed instead */	\
          compl_pdreg(subg_compl_stack_ptr(SUBG_PTR)) = NULL;		\
          reclaim_del_ret_list(SUBG_PTR);				\
          if (neg_simplif_possible(SUBG_PTR))				\
            simplify_neg_fails(SUBG_PTR);				\
          subg_cp_ptr(SUBG_PTR) = (CPtr) -1; 				\
        } 
#endif

/*
 * do not put TRUE but some !0 value that GC can recognize as int tagged
 */
#define mark_neg_loop(susp)     csf_neg_loop(susp) = INT;

/*----------------------------------------------------------------------*/
/* The following macro might be called more than once for some subgoal. */
/* Functions or macros that it calls are robust under repeated uses.    */
/*----------------------------------------------------------------------*/

#define reclaim_incomplete_table_structs(SUBG_PTR,CS_FRAME) {	\
     chat_free_cons_chat_areas(CS_FRAME);		\
     table_complete_entry(SUBG_PTR);			\
 }

/*----------------------------------------------------------------------*/

#define adjust_level(CS_FRAME) { \
  register ComplStackFrame xt1;                         \
  register int level;                                   \
    level = compl_level(CS_FRAME);	   	        \
    if (level < compl_level(openreg)) {                 \
      for (xt1 = CS_FRAME;			        \
	   compl_level(xt1) >= level && xt1 >= openreg; \
	   next_CSF(xt1)) {	                        \
	compl_level(xt1) = level;                       \
      }						        \
    }                                                   \
  }

/*----------------------------------------------------------------------*/

#define reset_slg_registers \
    level_num = 0; \
    ptcpreg = NULL

/*----------------------------------------------------------------------*/

#define pdlpush(cell)	*(pdlreg) = cell;  pdlreg--

#define pdlpop		*(++pdlreg)

#define pdlempty	(pdlreg == (CPtr)(pdl.high) - 1)

#define resetpdl \
   if (pdlreg < (CPtr)pdl.low) \
     xxx_exit("pdlreg grew too much"); \
   else (pdlreg = (CPtr)(pdl.high) - 1)

#define remove_open_tables_loop(Endpoint) remove_open_tries(Endpoint)

#define remove_open_tables() remove_open_tries(COMPLSTACKBOTTOM)

/*----------------------------------------------------------------------*/

/* #ifdef ATTR_VARS */
#define get_var_and_attv_nums(var_num, attv_num, tmp_int)	\
  var_num = tmp_int & 0xffff;					\
  attv_num = tmp_int >> 16
/* #endif */

/*----------------------------------------------------------------------*/

#endif
