/* File:      residual.c
** Author(s): Kostis Sagonas, Baoqiu Cui
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: residual.c,v 1.9 2004/06/09 09:21:43 kostis Exp $
** 
*/


#include <stdio.h>
#include <stdlib.h>

#include "xxx_config.h"

#include "auxlry.h"
#include "cell_xsb.h"
#include "memory_xsb.h"
#include "psc_xsb.h"
#include "heap_xsb.h"
#include "macro_xsb.h"
#include "register.h"
#include "binding.h"
#include "error_xsb.h"
#include "io_builtins_xsb.h"
#include "debug.h"
#include "residual.h"

/*----------------------------------------------------------------------*/

#ifdef DEBUG_DELAYVAR
#define print_trie_atom(X) {\
 if (isatom(X)) \
   printf("atom(%s)",atom_val(X));\
 else if (isstruct(X)) \
   printf("atom(%s/%d)",get_name((Psc)dec_addr(X)),\
                        get_arity((Psc)dec_addr(X)));\
 else if (isinteger(X)) \
   printf("atom(%d)",int_val(X));\
 else if (islist(X))\
   printf("./2");\
 else\
  printf("Unk(%x)",(int)X);\
 }
#endif

/*----------------------------------------------------------------------*/

static Cell cell_array[500];

/*----------------------------------------------------------------------*/

#if defined(CST_SUPPORT)
#define build_subgoal_args(SUBG)	\
    load_solution_trie(arity, 0, &cell_array[arity-1], subg_cst(SUBG))
#else
#define build_subgoal_args(SUBG)	\
    load_solution_trie(arity, 0, &cell_array[arity-1], subg_leaf_ptr(SUBG))
#endif
static CPtr *copy_of_var_addr;
static int  copy_of_num_heap_term_vars;

/*----------------------------------------------------------------------*/

/*
 * Function build_delay_list() is called by builtin #143 GET_DELAY_LISTS
 * to construct on the heap the delay list pointed by `de'.  Since XSB
 * 1.8.1, this function is changed to handle variables in delay list.
 * Basically, to construct a delayed subgoal, we have to go through three 
 * tries one by one: the call trie, answer trie, and delay trie of the
 * delayed subgoal.  Delay trie contains the substitution factor of the
 * answer for the delayed subgoal.
 */

static void build_delay_list(CPtr delay_list, DE de)
{
  Psc  psc;
  int  i, j, arity;
  CPtr head, tail;
  SGFrame subg;
#if defined(CST_SUPPORT)
  CST
#else
  BTNptr
#endif
      ans_subst;
#ifdef DEBUG_DELAYVAR
  BTNptr subs_factp;
#endif
  CPtr *tmp_var_addr;
  CPtr oldhreg = hreg;

#ifdef HEAP_DEBUG
  xsb_dbgmsg("delay_list starts at %p", hreg);
#endif
  i = 0;
  if (de != NULL && ((Cell)de != makenil)) {
    tail = hreg+1;
    bind_list(delay_list, hreg);
    hreg = hreg + 3; 
    build_delay_list(tail, de_next(de)); /* recursive call */
    head = hreg;
    subg = de_subgoal(de);
    psc = TIF_PSC(subg_tif_ptr(subg));
    arity = get_arity(psc);
    if ((ans_subst = de_ans_subst(de)) == NULL) { /* Negative DE */
      follow(oldhreg) = makestruct(hreg);
      new_heap_functor(head, tnot_psc);
      if (arity == 0) {
	bind_atom(head, get_name(psc));
	hreg += 3;
      } else {
	CPtr s_reg = head+1;
	follow(head++) = makestruct(s_reg);
	hreg += arity+4; /* need arity(tnot)+2+arity(psc)+1 new cells */
	new_heap_functor(s_reg, psc);
	for (j = 1; j <= arity; j++) {
	  new_heap_free(s_reg);
	  cell_array[arity-j] = cell(s_reg-1);
	}
	build_subgoal_args(subg);
      }
    } else {					/* Positive DE */
      if (arity == 0) {
	new_heap_atom(oldhreg, get_name(psc));
      } else {
	CPtr s_reg;
#ifdef DEBUG_DELAYVAR
	/*
	 * de_subs_fact(de) is the root of de's delay trie -- the saved
	 * substitution factor of the answer to the subgoal call of
	 * this delayed element.
	 */
	subs_factp = de_subs_fact(de);
#endif
	s_reg = head;
	follow(oldhreg) = makestruct(head);
	hreg += arity+1;
	new_heap_functor(s_reg, psc);
	for (j = 1; j <= arity; j++) {
	  new_heap_free(s_reg);
	  cell_array[arity-j] = cell(s_reg-1);
	}
	
#ifdef DEBUG_DELAYVAR
	xsb_dbgmsg(">>>> (before build_subgoal_args) num_heap_term_vars = %d",
		   num_heap_term_vars);
#endif

	/*
	 * Function build_subgoal_args() goes through the subgoal trie
	 * and binds all the arguments of the subgoal skeleton, which
	 * were built in the heap (like q(_,_)), to the actual call
	 * arguments (like X and f(Y,g(Z)) in q(X, f(Y,g(Z)))).
	 *
	 * After build_subgoal_args, var_addr[] contains all the
	 * variables in the subgoal call.
	 */
	build_subgoal_args(subg);
	
#ifdef DEBUG_DELAYVAR
	xsb_dbgmsg(">>>> (after build_subgoal_args) num_heap_term_vars = %d",
		   num_heap_term_vars);
#endif
	
	for (i = 0, j = num_heap_term_vars-1; j >= 0; j--) {
	  cell_array[i++] = (Cell)var_addr[j];
#ifdef DEBUG_DELAYVAR
	  xsb_dbgmsg(">>>> var_addr[%x] = %x", j, (int)var_addr[j]);
#endif
	}
	
	/*
	 * Function load_solution_trie() goes through the answer trie
	 * and binds the call substitution factor.  The substitution
	 * factor of the answer is left in var_addr[].
	 */
	load_solution_trie(i, 0, &cell_array[i-1], ans_subst);
	
#ifdef DEBUG_DELAYVAR
	xsb_dbgmsg(">>>> (after load_solution_trie) num_heap_term_vars = %d",
		   num_heap_term_vars);
#endif
	
	for (i = 0, j = num_heap_term_vars-1; j >= 0; j--) {
	  cell_array[i++] = (Cell)var_addr[j];
#ifdef DEBUG_DELAYVAR
	  xsb_dbgmsg(">>>> var_addr[%x] = %x", j, (int)var_addr[j]);
#endif
	}
	
	tmp_var_addr = var_addr;
	
	/*
	 * Restore var_addr[] to copy_of_var_addr[], which contains all
	 * the variables in the _head_ predicate after get_returns is
	 * called (see bineg_xsb_i.h).
	 *
	 * The content of copy_of_var_addr[] is used in
	 * load_delay_trie(), and might be changed by each delay
	 * element.  But during the whole process for one paticular
	 * delay list, copy_of_var_addr[] will be shared by all the
	 * delay elements.
	 *
	 * It is not necessary to restore copy_of_num_heap_term_vars
	 * each time before load_delay_trie() is called for a delay
	 * element.  Each variable in the trie can find its binding in
	 * copy_of_var_addr[] without num_heap_term_vars.
	 */
	var_addr = copy_of_var_addr; /* variables left in the head */
	num_heap_term_vars = copy_of_num_heap_term_vars;
	
	
#ifdef DEBUG_DELAYVAR
	xsb_dbgmsg(">>>> NOW copy_of_num_heap_term_vars = %d",
		   copy_of_num_heap_term_vars);
	{
	  int i;
	  for(i = 0; i < num_heap_term_vars; i++)
	    xsb_dbgmsg(">>>> var_addr[%d] = %x",i, (int)var_addr[i]);
	  
	  fprintf(stddbg, "Stored Subs Fact: <");
	  {
	    BTNptr x = subs_factp;
	    if (x == NULL)
	      xsb_dbgmsg(">>>> subs_factp is NULL");
	    while(x != NULL){
	      print_trie_atom(Atom(x));
	      if(Sibl(x) != NULL) 
		fprintf(stddbg, "!");
	      x = Child(x);
	      
	    }
	  }
	  fprintf(stddbg, ">\n");
	}
	xsb_dbgmsg(">>>> num_heap_term_vars is %d before calling load_delay_trie",
		   num_heap_term_vars);
#endif /* DEBUG_DELAYVAR */
	
#ifndef IGNORE_DELAYVAR
	load_delay_trie(i, &cell_array[i-1], de_subs_fact_leaf(de));
#endif
	
#ifdef DEBUG_DELAYVAR
	xsb_dbgmsg(">>>> num_heap_term_vars becomes %d",
		   num_heap_term_vars);
	for (i = 0; i < num_heap_term_vars; i++)
	  xsb_dbgmsg(">>>> var_addr[%d] = %x",i, (int)var_addr[i]);
#endif
	var_addr = tmp_var_addr;
      }
    }
    hreg++;
  } else {
    bind_nil(delay_list);
  }
}

/*----------------------------------------------------------------------*/
/*
 * When GET_DELAY_LISTS is called, we can assume that the
 * corresponding tabled subgoal call has been completed and so trie
 * code will be used to return the answer (see
 * trie_get_returns_for_call()).  After the execution of trie code,
 * var_regs[] contains the substitution factor of the _answer_ to the
 * call.
 */

void get_delay_lists(
#if defined(CST_SUPPORT)
    CST as_leaf,
#else
    BTNptr as_leaf, 
#endif
    Cell delay_lists)
{
  DL dl;
  DE de;
  CPtr dls_head, dls_tail = NULL;

#ifdef DEBUG_DELAYVAR
  xsb_mesg(">>>> (at the beginning of GET_DELAY_LISTS");
  xsb_mesg(">>>> global_num_vars = %d)",global_num_vars);
  
  {
    int i;
    for (i = 0; i <= global_num_vars; i++) {
      Cell x;
      fprintf(stddbg, ">>>> var_regs[%d] =",i);
      x = (Cell)var_regs[i];
      deref(x);
      printterm(x,1,25);
      fprintf(stddbg, "\n");
    }
  }
#endif /* DEBUG_DELAYVAR */

  if (is_conditional_answer(as_leaf)) {
    bind_list((CPtr)delay_lists, hreg);
    { /*
       * Make copy of var_regs & global_num_vars (after get_returns,
       * which calls trie_get_returns_for_call).  (global_num_vars +
       * 1) is the number of variables left in the answer
       * (substitution factor of the answer)
       *
       * So, copy_of_var_addr[] is the substitution factor of the
       * answer for the head predicate.
       */
      int i;
      copy_of_var_addr = calloc(var_addr_arraysz, sizeof(CPtr));
      if (copy_of_var_addr == NULL) {
	xxx_exit("Not enough memory to calloc copy_of_var_addr!\nBye");
      }
      for( i = 0; i <= global_num_vars; i++)
	copy_of_var_addr[i] = var_regs[i];
      
      copy_of_num_heap_term_vars = global_num_vars + 1;
    }

    for (dl = asi_dl_list((ASI) Delay(as_leaf)); dl != NULL; ) {
      dls_head = hreg;
      dls_tail = hreg+1;
      new_heap_free(hreg);
      new_heap_free(hreg);
      de = dl_de_list(dl);
      /*
       * This answer may have more than one delay list.  We have to
       * restore copy_of_num_heap_term_vars for each of them.  But,
       * among delay elements of each delay list, it is not necessary
       * to restore this value.
       *
       * Note that global_num_vars is always set back to
       * copy_of_num_heap_term_vars at the end of build_delay_list().
       */
      copy_of_num_heap_term_vars = global_num_vars + 1;
      build_delay_list(dls_head, de);
      if ((dl = dl_next(dl)) != NULL) {
	bind_list(dls_tail, hreg);
      }
    }
    bind_nil(dls_tail);
    free(copy_of_var_addr);
  } else {
    bind_nil((CPtr)delay_lists);
  }
}

/*---------------------- end of file residual.c ------------------------*/
