/* File:    term_builtins.c
** Author:  Kostis Sagonas
** Contact: kostis@csd.uu.se
** 
** Copyright (C) Kostis Sagonas, 2000
** 
** XXX is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XXX is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XXX; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: term_builtins.c,v 1.10 2006/03/14 21:57:17 kostis Exp $
** 
*/


#include "xxx_config.h"

#include <stdio.h>

#include "auxlry.h"
#include "cell_xsb.h"
#include "psc_xsb.h"
#include "deref.h"
#include "heap_xsb.h"
#include "macro_xsb.h"
#include "memory_xsb.h"
#include "register.h"
#include "binding.h"
#include "subp.h"
#include "flags_xsb.h"
#include "error_xsb.h"
#include "sp_unify_xsb_i.h"
#include "term_builtins.h"

/*----------------------------------------------------------------------*/

#include "ptoc_tag_xsb_i.h"

/*----------------------------------------------------------------------*/

bool ground(CPtr term)
{
  cptr_deref2(term, { return FALSE; });
  if (isstruct(term))
    {
      int j, arity, flag=1;
      arity = get_arity(get_str_psc(term));
      for (j=1; j <= arity; j++) 
	flag = flag & ground(clref_val(term)+j);
      return flag;
    }
  if (islist(term))
    {
      int flag;
      flag = ground(clref_val(term));
      flag = flag & ground(clref_val(term)+1);
      return flag;
    }
  return TRUE;
}

/*----------------------------------------------------------------------*/
/* rec_variant() implements a simultaneous numbervars operation but a
 * hack is used: instead of binding the variables to terms of the form
 * $VAR(N), N = 1,2,... variables are bound to *atoms* 1,2,...  This
 * is safe to do as these values do not appear as pointers to regular
 * atoms anyway (the var_num values are shifted so that they do not
 * clash with the tagging scheme which assumes that atom pointers are
 * word-aligned), and the bindings are undone when variant is
 * finished. */
/*----------------------------------------------------------------------*/

static int var_num;

static bool rec_variant(CPtr term1, CPtr term2)
{
  cptr_deref(term1);
  cptr_deref(term2);
  if (term1 == term2) return TRUE; /* catches the same term */
  if (isref(term1))
    {
      Cell atmp;
      var_num++;
      atmp = (var_num << 2); /* make it look like an atom pointer */
      bind_atom(term1,atmp); /* a hack! */
      return atom_unify((Cell)term1,(Cell)term2);
    }
  if (isstruct(term1) && isstruct(term2) &&
      (get_str_psc(term1) == get_str_psc(term2)))
    {
      int arity = get_arity(get_str_psc(term1));
      if (arity == get_arity(get_str_psc(term2)))
	{
	  int j, flag = 1;
	  for (j=1; j <= arity; j++) 
	    flag &= rec_variant(clref_val(term1)+j,clref_val(term2)+j);
	  return (flag != 0);
	}
      else return FALSE;
    }
  if (islist(term1) && islist(term2))
    {
      if (rec_variant(clref_val(term1),clref_val(term2)))
	return rec_variant(clref_val(term1)+1,clref_val(term2)+1);
      else return FALSE;
    }
  return (term1 == term2);
}

bool variant_builtin(Cell term1, Cell term2)
{
  var_num = 0;
  return rec_variant((CPtr)term1, (CPtr)term2);
}

/*----------------------------------------------------------------------*/

static int num_vars;
static Psc num_vars_psc;

static void rec_numbervars(CPtr term)
{
#if (!defined(ATTR_VARS))
  cptr_deref2(term, { bind_struct(term, hreg);
                      new_heap_functor(hreg, num_vars_psc);
		      new_heap_num(hreg, num_vars);
		      num_vars++;
		      return; });
#else
  cptr_deref(term);
  if (isref(term)) {
    /* fprintf(stderr, "%s: In is_ref case... binding to $VAR(%d)\n",
       __FUNCTION__, num_vars); */
    bind_struct(term, hreg);
    new_heap_functor(hreg, num_vars_psc);
    new_heap_num(hreg, num_vars);
    num_vars++;
    return;
  }
  if (isattv(term)) {
    /* fprintf(stderr, "%s: In is_attv case... binding to $VAR(%d)\n",
       __FUNCTION__, num_vars); */
    /* bind the attributed variable to a $VAR(N) term without
       triggering an attv interrupt -- if the binding is postponed,
       non-linearity of attributed variables cannot be handled */
    bind_copy((CPtr)dec_addr(term), makestruct(hreg));
    new_heap_functor(hreg, num_vars_psc);
    new_heap_num(hreg, num_vars);
    num_vars++;
    return;
  }
#endif
  if (isstruct(term))
    {
      byte j, arity;
      arity = get_arity(get_str_psc(term));
      for (j=1; j <= arity; j++) rec_numbervars(clref_val(term)+j);
      return;
    }
  if (islist(term))
    {
      rec_numbervars(clref_val(term));
      rec_numbervars(clref_val(term)+1);
      return;
    }
}

/*----------------------------------------------------------------------*/

int numbervars(CPtr term, int num)
{
  bool is_new;

  num_vars = num;
  if (! num_vars_psc) {
    Pair pair = insert("$VAR", 1, (Psc)flags[CURRENT_MODULE], &is_new);
    num_vars_psc = pair_psc(pair);
  }
  rec_numbervars(term);
  return num_vars;
}

/*----------------------------------------------------------------------*/
