/* File:      tr_utils.c
** Author(s): Prasad Rao, Kostis Sagonas
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: tr_utils.c,v 1.23 2006/07/23 13:50:00 kostis Exp $
** 
*/

#include "xxx_config.h"

#include <stdio.h>
#include <stdlib.h>

#include "auxlry.h"
#include "cell_xsb.h"
#include "psc_xsb.h"
#include "cinterf.h"
#include "memory_xsb.h"
#include "deref.h"
#include "heap_xsb.h"
#include "flags_xsb.h"
#include "tst_aux.h"
#if (!defined(WAM_TRAIL))
#include "trail_defs.h"
#endif
#include "macro_xsb.h"
#include "register.h"
#include "binding.h"
#include "bldval.h"
#include "choice.h"
#include "inst_xsb.h"
#include "error_xsb.h"
#include "biassert.h"
#include "chat.h"
#include "emuloop.h"   /* for unify() */
#include "tr_utils.h"

/*----------------------------------------------------------------------*/

#define MAX_VAR_SIZE	200

/*----------------------------------------------------------------------*/

#include "ptoc_tag_xsb_i.h"
#include "term_psc_xsb_i.h"

/*----------------------------------------------------------------------*/

bool has_unconditional_answers(SGFrame subg)
{
  ALNptr node_ptr = subg_answers(subg);
 
  /* Either subgoal has no answers or it is completed */
  /* and its answer list has already been reclaimed. */
  /* In either case, the result is immediately obtained. */
 
  if (node_ptr <= COND_ANSWERS) return (node_ptr == UNCOND_ANSWERS);
 
  /* If the subgoal has not been completed, or is early completed but its */
  /* answer list has not been reclaimed yet, check each of its nodes. */
 
  while (node_ptr) {
    if (
#if defined(CST_SUPPORT)
	is_unconditional_answer(ALN_Constraint(node_ptr))
#else
	is_unconditional_answer(aln_answer_ptr(node_ptr))
#endif
	) return TRUE;
    node_ptr = aln_next_aln(node_ptr);
  }
  return FALSE;
}

/*----------------------------------------------------------------------*/

/*
 * Given a term as an arity and array of subterms, determines whether
 * this term is present in the given trie.  If an array is supplied,
 * the variables in the term are copied into it, with the 0th element
 * containing the count.  The leaf representing the term is returned
 * if present, or NULL otherwise.
 */

#if defined(CST_SUPPORT)
btn_cst_tuple
#else
BTNptr 
#endif
variant_trie_lookup(int nTerms, CPtr termVector, BTNptr trieRoot,
			   Cell varArray[])
{
  BTNptr trieNode;      /* Used for stepping down through the trie */
  Cell symbol;		/* Trie representation of current heap symbol,
			   used for matching/inserting into a TSTN */
  Cell subterm;		/* Used for stepping through the term */
  int std_var_num;	/* Next available TrieVar index; for standardizing
			   variables when interned */
#if defined(CST_SUPPORT)
  CL   constraints;
  int  var_ctr, attv_ctr;
  btn_cst_tuple tup;

  tup.btn = NULL;
  tup.cst = NULL;
  constraints = NULL;
  var_ctr = 0;
  attv_ctr = 0;
#endif

  if ( IsNULL(trieRoot) || IsNULL(BTN_Child(trieRoot)) )
#if defined(CST_SUPPORT)
    return tup;
#else
    return NULL;
#endif
  else if ( nTerms > 0) {
    trieNode = trieRoot;
    std_var_num = 0;

    Trail_ResetTOS;
    TermStack_ResetTOS;
    TermStack_PushLowToHighVector(termVector,nTerms);

    while ( ! TermStack_IsEmpty && IsNonNULL(trieNode) ) {
      subterm = TermStack_Pop;
      deref(subterm);
      switch (cell_tag(subterm)) {

      case REF:
      case REF1:
	if ( ! IsStandardizedVariable(subterm) ) {
	  StandardizeVariable(subterm, std_var_num);
	  Trail_Push(subterm);
	  symbol = EncodeNewTrieVar(std_var_num);
	  std_var_num++;
	}
	else
	  symbol = EncodeTrieVar(IndexOfStdVar(subterm));
	break;

      case ATM:
      case INT:
      case FLOAT:
	symbol = EncodeTrieConstant(subterm);
	break;

      case STRUCT:
	symbol = EncodeTrieFunctor(subterm);
	TermStack_PushFunctorArgs(subterm);
	break;

      case LIST:
	symbol = EncodeTrieList(subterm);
	TermStack_PushListArgs(subterm);
	break;

#ifdef ATTR_VARS
      case ATTV:
#if defined(CST_SUPPORT)
	  constraints = 
	      CL_insert_constraints(subterm, 
				    constraints, 
				    &var_ctr, 
				    &attv_ctr, 
				    &(tstTrail.top),
				    1);
#endif
	  symbol = EncodeNewTrieAttv(subterm);
	  std_var_num++;
	  break;
#endif

      default: 
	/* Bad tag Error */
	trieNode = NULL;
	continue;
      }

      if ( IsHashHeader(BTN_Child(trieNode)) ) {
	BTHTptr ht = BTN_GetHashHdr(trieNode);
	trieNode = *CalculateBucketForSymbol(ht,symbol);
      }
      else
	trieNode = BTN_Child(trieNode);
      {
	int chain_length;
	SearchChainForSymbol(trieNode,symbol,chain_length);
      }
    }

    if ( IsNonNULL(varArray) ) {
      unsigned int i;

      varArray[0] = tstTrail.top - tstTrail.base;
      for (i = 1; i <= varArray[0]; i++)
	varArray[i] = (Cell)tstTrail.base[i - 1];
    }

#if defined(CST_SUPPORT)
    tup.btn = trieNode;
    if(trieNode == NULL)
    {
	tup.btn = NULL;
	tup.cst = NULL;
    }else{
	tup.cst = CST_find_node(BTN_CST(tup.btn), 
				CL_remove_duplicates(constraints));
	if(!is_start_node(tup.cst))
	{
	    tup.btn = NULL;
	    tup.cst = NULL;
	}
    }
#endif
    
    Trail_Unwind_All;
    
#if defined(CST_SUPPORT)
    return tup;
#else
    return trieNode;
#endif
  }

  else if ( IsEscapeNode(BTN_Child(trieRoot)) ) {
    if ( IsNonNULL(varArray) )
      varArray[0] = 0;
#if defined(CST_SUPPORT)
    tup.btn = BTN_Child(trieRoot);
    tup.cst = CST_find_node(BTN_CST(tup.btn), 
			    (CL) NULL);
    return tup;
#else
    return BTN_Child(trieRoot);
#endif
  }

  else { /* Error conditions: Malformed trie or nTerms is wrong. */
#if defined(CST_SUPPORT)
    return tup;
#else
    return NULL;
#endif
  }
}

/*----------------------------------------------------------------------*/

SGFrame get_subgoal_ptr(Cell callTerm, TIFptr pTIF)
{
  int arity;
#if defined(CST_SUPPORT)
  btn_cst_tuple leaf;
#else
  BTNptr call_trie_leaf;
#endif

  arity = get_arity(term_psc(callTerm));
#if defined(CST_SUPPORT)
  leaf = variant_trie_lookup(arity, clref_val(callTerm) + 1,
			     TIF_CallTrie(pTIF), NULL);

  if ( leaf.cst != NULL)
    return CST_get_SGF(leaf.cst);
  else
    return NULL;
#else
  call_trie_leaf = variant_trie_lookup(arity, clref_val(callTerm) + 1,
				       TIF_CallTrie(pTIF), NULL);
  
  if ( IsNonNULL(call_trie_leaf) )
    return CallTrieLeaf_GetSF(call_trie_leaf);
  else
    return NULL;
#endif
}

/*----------------------------------------------------------------------*/

/*
 * Given a vector of terms and their number, N, builds a ret/N structure
 * on the heap containing those terms.  Returns this constructed term.
 */

Cell build_ret_term(int arity, Cell termVector[])
{
  Pair sym;
  CPtr ret_term;
  int  i;
  bool is_new;

  if (arity == 0)
    return makeatom(ret_psc[0]);  /* return as a term */
  else {
    ret_term = hreg;  /* pointer to where ret(..) will be built */
    sym = insert("ret", (byte)arity, (Psc)flags[CURRENT_MODULE], &is_new);
    new_heap_functor(hreg, pair_psc(sym));
    for ( i = 0; i < arity; i++ )
      bld_val(termVector[i]);
    return makestruct(ret_term);  /* return as a term */
  }
}

/*----------------------------------------------------------------------*/

/*
 * Given a term representing a tabled call, determine whether it is
 * recorded in the Call Table.  If it is, then return a pointer to its
 * subgoal frame and construct on the heap the answer template required
 * to retrieve answers for this call.  Place a reference to this term in
 * the location pointed to by the second argument.
 */

SGFrame get_call(Cell callTerm, Cell *retTerm)
{
  Psc  psc;
  TIFptr tif;
  int arity;
#if defined(CST_SUPPORT)
  btn_cst_tuple tup;
#else
  BTNptr call_trie_leaf;
#endif
  Cell callVars[MAX_VAR_SIZE + 1];

  psc = term_psc(callTerm);
  if ( IsNULL(psc) ) {
    xxx_err_handle(TYPE_ERROR, 1, "get_call", 3, "callable term", callTerm);
    return NULL;
  }
  tif = get_tip(psc);
  arity = get_arity(psc);
  if ( IsNULL(tif) )
    xxx_abort("Predicate %s/%d is not tabled", get_name(psc), arity);

#if defined(CST_SUPPORT)
  tup
#else
  call_trie_leaf
#endif
      = variant_trie_lookup(arity, clref_val(callTerm) + 1,
			    TIF_CallTrie(tif), callVars);
  if ( 
#if defined(CST_SUPPORT)
      tup.cst == NULL
#else
      IsNULL(call_trie_leaf)
#endif
      )
    return NULL;
  else {
    SGFrame sf = 
#if defined(CST_SUPPORT)
	CST_get_SGF(tup.cst);
#else
	CallTrieLeaf_GetSF(call_trie_leaf);
#endif
    *retTerm = build_ret_term(callVars[0],&callVars[1]);
    return sf;
  }
}

/*======================================================================*/

/*
 *                     D E L E T I N G   T R I E S
 *                     ===========================
 */


/* Stack for top-down traversing and freeing components of a trie
   -------------------------------------------------------------- */
struct freeing_stack_node {
  BTNptr item;
  struct freeing_stack_node *next;
};

#define push_node(node){\
  struct freeing_stack_node *temp;\
  temp = (struct freeing_stack_node *)malloc(sizeof(struct freeing_stack_node));\
  if (temp == NULL){\
    xxx_exit("Out of Memory");\
  } else {\
    temp->next   = node_stk_top;\
    temp->item   = node;\
    node_stk_top = temp;\
  }\
}

#define pop_node(node) \
{\
  struct freeing_stack_node *temp;\
  if (node_stk_top == NULL) {\
    xxx_dbgmsg("DELETE_PREDICATE_TABLE: pop attempted from NULL");\
    return;\
  }\
  node         = node_stk_top->item;\
  temp         = node_stk_top;\
  node_stk_top = node_stk_top->next;\
  free(temp);\
}


static void free_trie_ht(BTHTptr ht)
{
  TrieHT_RemoveFromAllocList(*smBTHT,ht);
  free(BTHT_BucketArray(ht));
  SM_DeallocateStruct(*smBTHT,ht);
}


static void delete_variant_table(BTNptr x)
{
  struct freeing_stack_node *node_stk_top = 0, *call_nodes_top = 0;
  BTNptr node, rnod, *Bkp; 
  BTHTptr ht;

  if ( IsNULL(x) )
    return;

  push_node(x);
  while (node_stk_top != 0) {
    pop_node(node);
    if ( IsHashHeader(node) ) {
      ht = (BTHTptr) node;
      for (Bkp = BTHT_BucketArray(ht);
	   Bkp < BTHT_BucketArray(ht) + BTHT_NumBuckets(ht);
	   Bkp++) {
	if ( IsNonNULL(*Bkp) )
	  push_node(*Bkp);
      }
      free_trie_ht(ht);
    }
    else {
      if ( IsNonNULL(Sibl(node)) )
	push_node(Sibl(node));
      if ( IsNonNULL(BTN_Child(node)) ) {
	if ( IsLeafNode(node) ) {
#if defined(CST_SUPPORT)
	    /*
	     * Remove the CST
	     */
	    CST cst, root;
	    
	    cst = root = BTN_CST(node);
	    
	    while(cst->left != NULL)
		cst = cst->left;
	    
	    while(cst != NULL)
	    {
		if(CST_get_SGF(cst) != NULL)
		{
		    SGFrame pSF = CST_get_SGF(cst);
#else
		    SGFrame pSF = CallTrieLeaf_GetSF(node);
#endif
		    /*
		     * Remove the subgoal frame and its dependent structures
		     */
		    if ( subg_answers(pSF) == COND_ANSWERS ) {
			xxx_warn("abolish_table_pred/1 is deleting a table with conditional\
				  answers: delay dependencies may be corrupted.\n");
		    }
		    if ( IsNonNULL(subg_ans_root_ptr(pSF)) ) {
			call_nodes_top = node_stk_top;
			push_node((BTNptr)subg_ans_root_ptr(pSF));
			while (node_stk_top != call_nodes_top) {
			    pop_node(rnod);
			    if ( IsHashHeader(rnod) ) {
				ht = (BTHTptr) rnod;
				for (Bkp = BTHT_BucketArray(ht);
				     Bkp < 
					 BTHT_BucketArray(ht) + 
					 BTHT_NumBuckets(ht);
				     Bkp++) {
				    if ( IsNonNULL(*Bkp) )
					push_node(*Bkp);
				}
				free_trie_ht(ht);
			    }
			    else {
				if (Sibl(rnod)) 
				    push_node(Sibl(rnod));
				if ( ! IsLeafNode(rnod) )
				{
				    push_node(Child(rnod));
				}
#if defined(CST_SUPPORT)
				else {
				    CST_release_tree(BTN_CST(rnod));
				}
#endif
				SM_DeallocateStruct(*smBTN,rnod);
			    }
			}
		    } /* free answer trie */
		    free_answer_list(pSF);
		    FreeProducerSF(pSF);
#if defined(CST_SUPPORT)
		}
		if(CST_is_root_node(cst))
		    cst = NULL;
		else
		    if(cst->parent->left == cst && cst->parent->right != NULL)
		    {
			cst = cst->parent->right;
			while(cst->left != NULL)
			    cst = cst->left;
		    }else
			cst = cst->parent;
	    }
	    CST_release_tree(root);
#endif
	} /* is leaf */
	else 
	  push_node(Child(node));
      } /* there is a child of "node" */
      SM_DeallocateStruct(*smBTN,node);
    }
  }
}

/*----------------------------------------------------------------------*/
/* Delete the table for a given tabled predicate, specified as a TIF    */
/*----------------------------------------------------------------------*/

void delete_predicate_table(TIFptr tif)
{
  delete_variant_table(TIF_CallTrie(tif));
  TIF_CallTrie(tif) = NULL;
}

/*----------------------------------------------------------------------*/

inline static int is_hash(BTNptr x) 
{
  if (x == NULL)
    return 0;
  else
    return( IsHashHeader(x) );
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
 * Set values for "parent" -- the parent node of "current" -- and
 * "cur_hook" -- an address containing a pointer into "current"'s level
 * in the trie.  If there is no parent node, use the value of
 * "root_hook" to find the level.  If the hook is actually contained in
 * the parent of current (as its child field), then we've ascended as
 * far as we need to go.  Set parent to NULL to indicate this.
 */

static void set_parent_and_node_hook(BTNptr current, BTNptr *root_hook,
				     BTNptr *parent, BTNptr **cur_hook)
{
  BTNptr par;

  if (IsTrieRoot(current))  /* defend against root having a set parent field */
    par = NULL;
  else {
    par = BTN_Parent(current);
    if ( IsNonNULL(par) && (root_hook == &BTN_Child(par)) )
      par = NULL;    /* stop ascent when hooking node is reached */
  }
  if ( IsNULL(par) )
    *cur_hook = root_hook;
  else
    *cur_hook = &BTN_Child(par);
  *parent = par;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
 * Given some non-root node which is *not the first* (or only) sibling,
 * find the node which precedes it in the chain.  Should ONLY be used
 * when deleting trie components.  If a hash table is encountered, then
 * its number of contents is decremented.
 */
static BTNptr get_prev_sibl(BTNptr node)
{
  BTNptr sibling_chain;

  sibling_chain = BTN_Child(BTN_Parent(node));
  if (IsHashHeader(sibling_chain)) {
    BTHTptr ht = (BTHTptr)sibling_chain;
    BTHT_NumContents(ht)--;
    sibling_chain = *CalculateBucketForSymbol(ht,BTN_Symbol(node));
  }
  while (sibling_chain != NULL) {
    if (BTN_Sibling(sibling_chain) == node)
      return(sibling_chain);
    sibling_chain = BTN_Sibling(sibling_chain);
  }  
  xxx_abort("Error in %s()", __FUNCTION__);
  return NULL;
}

/*----------------------------------------------------------------------*/
/* deletes and reclaims a whole branch in the return trie               */
/*----------------------------------------------------------------------*/

/*
 * Delete a branch in the trie down from node `lowest_node_in_branch'
 * up to the level pointed to by the hook location, as pointed to by
 * `hook'.  Under normal use, the "hook" is either for the root of the
 * trie, or for the first level of the trie (is a pointer to the child
 * field of the root).
 */

void delete_branch(BTNptr lowest_node_in_branch, BTNptr *hook)
{
  int num_left_in_hash;
  BTNptr prev, parent_ptr, *y1, *z;

  while ( IsNonNULL(lowest_node_in_branch) && 
	  ( Contains_NOCP_Instr(lowest_node_in_branch) ||
	    IsTrieRoot(lowest_node_in_branch) ) ) {
    /*
     *  Walk up a path with no branches, i.e., the nodes along this path
     *  have no siblings.  We know this because the instruction in the
     *  node is of the no_cp variety.
     */
    set_parent_and_node_hook(lowest_node_in_branch,hook,&parent_ptr,&y1);
    if (is_hash(*y1)) {
      z = CalculateBucketForSymbol((BTHTptr)(*y1),
				   BTN_Symbol(lowest_node_in_branch));
      if ( *z != lowest_node_in_branch )
	xxx_dbgmsg("DELETE_BRANCH: trie node not found in hash table");
      *z = NULL;
      num_left_in_hash = --BTHT_NumContents((BTHTptr)*y1);
      if (num_left_in_hash  > 0) {
	/*
	 * `lowest_node_in_branch' has siblings, even though they are not in
	 * the same chain.  Therefore we cannot delete the parent, and so
	 * we're done.
	 */
	SM_DeallocateStruct(*smBTN,lowest_node_in_branch);
	return;
      }
      else
	free_trie_ht((BTHTptr)(*y1));
    }
    /*
     *  Remove this node and continue.
     */
    SM_DeallocateStruct(*smBTN,lowest_node_in_branch);
    lowest_node_in_branch = parent_ptr;
  }
  if (lowest_node_in_branch == NULL)
    *hook = 0;
  else {
    if (Contains_TRY_Instr(lowest_node_in_branch)) {
      /* Alter sibling's instruction:  trust -> no_cp  retry -> try */
      set_opcode(Sibl(lowest_node_in_branch), Instr(Sibl(lowest_node_in_branch)) -1);
      y1 = &BTN_Child(BTN_Parent(lowest_node_in_branch));
      if (is_hash(*y1)) {
	z = CalculateBucketForSymbol((BTHTptr)(*y1),
				     BTN_Symbol(lowest_node_in_branch));
	num_left_in_hash = --BTHT_NumContents((BTHTptr)*y1);
      }
      else
	z = y1;
      *z = Sibl(lowest_node_in_branch);      
    }
    else { /* not the first in the sibling chain */
      prev = get_prev_sibl(lowest_node_in_branch);      
      Sibl(prev) = Sibl(lowest_node_in_branch);
      if (Contains_TRUST_Instr(lowest_node_in_branch))
	set_opcode(prev, Instr(prev) - 2); /* retry -> trust ; try -> nocp */
    }
    SM_DeallocateStruct(*smBTN,lowest_node_in_branch);
  }
}

/*----------------------------------------------------------------------*/

void safe_delete_branch(BTNptr lowest_node_in_branch)
{
  byte choicepttype;

  MakeStatusDeleted(lowest_node_in_branch);
  choicepttype = 0x3 & Instr(lowest_node_in_branch);
  set_opcode(lowest_node_in_branch, choicepttype | trie_no_cp_fail);
}

void undelete_branch(BTNptr lowest_node_in_branch)
{
  byte choicepttype; 
  byte typeofinstr;

  if( IsDeletedNode(lowest_node_in_branch) ){
    choicepttype = 0x3 &  Instr(lowest_node_in_branch);
    typeofinstr = (~0x3) & BTN_Status(lowest_node_in_branch);
    
    set_opcode(lowest_node_in_branch, choicepttype | typeofinstr);
    MakeStatusValid(lowest_node_in_branch);
  }
  else
    xxx_dbgmsg("Attempt to undelete a node that is not deleted");
}

/*----------------------------------------------------------------------*/

#define DELETE_TRIE_STACK_INIT 100
#define MAX_DELETE_TRIE_STACK_SIZE 1000
#define DT_NODE 0
#define DT_DS 1
#define DT_HT 2

static char *delete_trie_op = NULL;
static BTNptr *delete_trie_node = NULL;
static BTHTptr *delete_trie_hh = NULL;

static int trie_op_size, trie_node_size, trie_hh_size;

#define push_delete_trie_node(node,op) {\
  trie_op_top++;\
  if (trie_op_top >= trie_op_size) {\
    trie_op_size = 2*trie_op_size;\
    delete_trie_op = (char *)realloc(delete_trie_op,trie_op_size*sizeof(char));\
    if (!delete_trie_op) xxx_exit("out of space for deleting trie");\
    /*printf("realloc delete_trie_op to %d\n",trie_op_size);*/\
  }\
  delete_trie_op[trie_op_top] = op;\
  trie_node_top++;\
  if (trie_node_top >= trie_node_size) {\
    trie_node_size = 2*trie_node_size;\
    delete_trie_node = (BTNptr *)realloc(delete_trie_node,trie_node_size*sizeof(BTNptr));\
    if (!delete_trie_node) xxx_exit("out of space for deleting trie");\
    /*printf("realloc delete_trie_node to %d\n",trie_node_size);*/\
  }\
  delete_trie_node[trie_node_top] = node;\
}  

#define push_delete_trie_hh(hh) {\
  trie_op_top++;\
  if (trie_op_top >= trie_op_size) {\
    trie_op_size = 2*trie_op_size;\
    delete_trie_op = (char *)realloc(delete_trie_op,trie_op_size*sizeof(char));\
    if (!delete_trie_op) xxx_exit("out of space for deleting trie");\
    /*printf("realloc delete_trie_op to %d\n",trie_op_size);*/\
  }\
  delete_trie_op[trie_op_top] = DT_HT;\
  trie_hh_top++;\
  if (trie_hh_top >= trie_hh_size) {\
    trie_hh_size = 2*trie_hh_size;\
    delete_trie_hh = (BTHTptr *)realloc(delete_trie_hh,trie_hh_size*sizeof(BTHTptr));\
    if (!delete_trie_hh) xxx_exit("out of space for deleting trie");\
    /*printf("realloc delete_trie_hh to %d\n",trie_hh_size);*/\
  }\
  delete_trie_hh[trie_hh_top] = hh;\
}  

static void delete_trie(BTNptr iroot)
{
  BTNptr root, sib, chil;  
  int trie_op_top = 0;
  int trie_node_top = 0;
  int trie_hh_top = -1;

  if (!delete_trie_op) {
    delete_trie_op = (char *)malloc(DELETE_TRIE_STACK_INIT*sizeof(char));
    delete_trie_node = (BTNptr *)malloc(DELETE_TRIE_STACK_INIT*sizeof(BTNptr));
    delete_trie_hh = (BTHTptr *)malloc(DELETE_TRIE_STACK_INIT*sizeof(BTHTptr));
    trie_op_size = trie_node_size = trie_hh_size = DELETE_TRIE_STACK_INIT;
  }

  delete_trie_op[0] = 0;
  delete_trie_node[0] = iroot;
  while (trie_op_top >= 0) {
    /*    printf("top %d %d %d %p\n",trie_op_top,trie_hh_top,
	  delete_trie_op[trie_op_top],delete_trie_node[trie_node_top]); */
    switch (delete_trie_op[trie_op_top--]) {
    case DT_DS:
      root = delete_trie_node[trie_node_top--];
      SM_DeallocateStruct(*smBTN,root);
      break;
    case DT_HT:
      free_trie_ht(delete_trie_hh[trie_hh_top--]);
      break;
    case DT_NODE:
      root = delete_trie_node[trie_node_top--];
      if ( IsNonNULL(root) ) {
	if ( IsHashHeader(root) ) {
	  BTHTptr hhdr;
	  BTNptr *base, *cur;
	  hhdr = (BTHTptr)root;
	  base = BTHT_BucketArray(hhdr);
	  push_delete_trie_hh(hhdr);
	  for ( cur = base; cur < base + BTHT_NumBuckets(hhdr); cur++ ) {
	    if (IsNonNULL(*cur)) {
	      push_delete_trie_node(*cur,DT_NODE);
	    }
	  }
	}
	else {
	  sib  = Sibl(root);
	  chil = Child(root);      
	  /* Child nodes == NULL is not the correct test*/
	  if (IsLeafNode(root)) {
	    if (IsNonNULL(chil))
#if defined(CST_SUPPORT)
		/* WARNING: THIS WILL ONLY WORK WITH ANSWER
                   TRIES/CSTS. USING THIS WITH A CALL TRIE WILL CAUSE
                   A MEMORY LEAK! */
		CST_release_tree(BTN_CST(root));
#else
	      xxx_exit("Anomaly in delete_trie !");
#endif
	    push_delete_trie_node(root,DT_DS);
	    if (IsNonNULL(sib)) {
	      push_delete_trie_node(sib,DT_NODE);
	    }
	  }
	  else {
	    push_delete_trie_node(root,DT_DS);
	    if (IsNonNULL(sib)) {
	      push_delete_trie_node(sib,DT_NODE);
	    }
	    if (IsNonNULL(chil)) {
	      push_delete_trie_node(chil,DT_NODE);
	    }
	  }
	}
      } else printf("null node");
      break;
    }
  }
  if (trie_op_size > MAX_DELETE_TRIE_STACK_SIZE) {
    free(delete_trie_op); delete_trie_op = NULL;
    free(delete_trie_node); delete_trie_node = NULL;
    free(delete_trie_hh); delete_trie_hh = NULL;
    trie_op_size = 0; 
  }
}

/*----------------------------------------------------------------------*/

/*
 * This does not reclaim space for deleted nodes, only marks the node
 * as deleted and changes the try instruction to fail.  The deleted
 * node is then linked into the del_nodes_list in the completion
 * stack.
 */
void delete_return(
#if defined(CST_SUPPORT)
    CST leaf,
#else
    BTNptr leaf, 
#endif
    SGFrame sg_frame) 
{
  int is_groundcall = FALSE;
  ALNptr a, n, next;
#if defined(CST_SUPPORT)
  int answer_is_conditional = is_conditional_answer(leaf);
  ASI leaf_asi = Delay(leaf);
#endif

  /* deleting an answer makes it false, so we have to deal with delay lists */
  if (is_conditional_answer(leaf)) {
    ASI asi = Delay(leaf);
    release_all_dls(asi);
    if(
#if defined(CST_SUPPORT)
	leaf == BTN_CST(BTN_Child(subg_ans_root_ptr(sg_frame))) && 
	IsEscapeNode(BTN_Child(subg_ans_root_ptr(sg_frame)))
#else
	leaf == BTN_Child(subg_ans_root_ptr(sg_frame)) && IsEscapeNode(leaf)
#endif
	)
      is_groundcall = TRUE; /* do it here, when l is still valid */
  }

  if (is_completed(sg_frame)) 
#if defined(CST_SUPPORT)
      CST_safe_delete_branch(leaf);
#else
      safe_delete_branch(leaf);
#endif
  else {
#if defined(CST_SUPPORT)
    CST_delete_branch(leaf,&subg_ans_root_ptr(sg_frame));
#else
    delete_branch(leaf,&subg_ans_root_ptr(sg_frame));
#endif
    n = subg_ans_list_ptr(sg_frame);
    while (
#if defined(CST_SUPPORT)
	ALN_Constraint(aln_next_aln(n)) != leaf
#else
	aln_answer_ptr(aln_next_aln(n)) != leaf
#endif
	) {
      n = aln_next_aln(n); /* if a is not in that list a core dump ? */
    }
    if (n == NULL) { xxx_exit("Error in delete_return()"); return; }
    a               = aln_next_aln(n);
    next            = aln_next_aln(a);
    aln_next_aln(a) = compl_del_ret_list(subg_compl_stack_ptr(sg_frame));
    compl_del_ret_list(subg_compl_stack_ptr(sg_frame)) = a;
    aln_next_aln(n) = next;
    if (next == NULL) { subg_ans_list_tail(sg_frame) = n; /* last answer */ }      
    /* Make consumed answer field of consumers point to
       previous sibling if they point to a deleted answer */
    {
      CPtr c;
      chat_init_pheader chat_ptr;

      chat_ptr =
       (chat_init_pheader)compl_cons_copy_list(subg_compl_stack_ptr(sg_frame));
      while (chat_ptr != NULL) {
	NLChoice c = (NLChoice)(&chat_get_cons_start(chat_ptr));
	if (nlcp_trie_return(c) == a) { nlcp_trie_return(c) = n; }
	chat_ptr = (chat_init_pheader)nlcp_prevlookup(c);
      }
      /* run also the CP stack for any consumers that are in both places */
      for (c = breg; c <= subg_cp_ptr(sg_frame); c = nlcp_prevbreg(c)) {
	if (is_consumer_choicepoint(c) && nlcp_trie_return(c) == a) {
	  nlcp_trie_return(c) = n;
	}
      }
    }    
  }

  if (
#if defined(CST_SUPPORT)
      answer_is_conditional
#else
      is_conditional_answer(leaf)
#endif
      ) {

    simplify_pos_unsupported(
#if defined(CST_SUPPORT)
	leaf_asi
#else
	leaf
#endif
	);
    if (is_groundcall) {
      mark_subgoal_failed(sg_frame);
      simplify_neg_fails(sg_frame);
    }
  }
}

/*----------------------------------------------------------------------*/
/* Given a tabled subgoal, go through its list of deleted nodes (in the
 * completion stack), and reclaim the leaves and corresponding branches
 *----------------------------------------------------------------------*/

void reclaim_del_ret_list(SGFrame sg_frame)
{
  ALNptr x, y;
  
  x = compl_del_ret_list(subg_compl_stack_ptr(sg_frame));
  while (x != NULL) {
    y = x;
    x = aln_next_aln(x);
    SM_DeallocateStruct(smALN,y);
  }
}
 
/*----------------------------------------------------------------------*/

/*
**  Used in aggregs.P to implement aggregates.
**  Returns: Subgoal skeleton (i.e., ret(X,Y,Z), where X,Y,Z are all the 
**      	               distinct variables in the subgoal);
**           Pointer to the subgoal.
**  NOTE: Works only for incomplete subgoals
*/

void incomplete_retskel_subgoal(void)
{
  SGFrame sg_frame;
  CPtr    where;
  Integer Nvars;
  
  sg_frame = (SGFrame)(tcp_subgoal_ptr(breg));
  where = compl_hreg(subg_compl_stack_ptr(sg_frame));
  Nvars = int_val(cell(where)) & 0xffff; /* See get_var_and_attv_nums() */
  if (Nvars == 0) {
    ctop_string(1, (char *) ret_psc[0]);
  } else {
    Pair sym;
    Cell term;
    int  i;
    bool is_new;

    bind_struct((CPtr)ptoc_tag(1), hreg);
    sym = insert("ret", (byte)Nvars, (Psc)flags[CURRENT_MODULE], &is_new);
    new_heap_functor(hreg, sym->psc_ptr);
    where = where - Nvars - 1;
    for (i = Nvars; i > 0; i--) {
      term = *(where+i);
      bld_val(term);
    }
  }
  ctop_addr(2, sg_frame);
}

/*======================================================================*/

/*
 *                    I N T E R N E D   T R I E S
 *                    ===========================
 */

#define ADJUST_SIZE 100

BTNptr *Set_ArrayPtr = NULL;
/*
 * first_free_set is the index of the first deleted set.  The deleted
 * tries are deleted in builtin DELETE_TRIE, and the corresponding
 * elements in Set_ArrayPtr are linked to form a list.  So
 * Set_ArrayPtr[first_free_set] contains the index of the next deleted
 * set, ..., the last one contains 0.  If first_free_set == 0, that
 * means no free set available.
 */
static int first_free_set = 0;
static int Set_ArraySz = 100;
/*
 * num_sets is the number of sets have been used (including the fixed
 * trie, Set_ArrayPtr[0] (see trie_intern/3)).  It is also the index for
 * the next element to use when no free element is available.
 */
static int num_sets = 1;

/*----------------------------------------------------------------------*/

/* Allocate an array of handles to interned tries. */

void init_newtrie(void)
{
  Set_ArrayPtr = (BTNptr *) calloc(Set_ArraySz,sizeof(BTNptr));
}

/*----------------------------------------------------------------------*/

/* Returns a handle to an unused interned trie. */

long new_trie(void)
{
  long i;
  
  if (first_free_set != 0) {	/* a free set is available */
    i = first_free_set;		/* save it in i */
    first_free_set = (long) Set_ArrayPtr[first_free_set] >> 2;
    Set_ArrayPtr[i] = NULL;	/* must be reset to NULL */
  }
  else {
    if (num_sets == Set_ArraySz) { /* run out of elements */
      BTNptr *temp_arrayptr;

      temp_arrayptr = Set_ArrayPtr;
      Set_ArraySz += ADJUST_SIZE;  /* adjust the array size */
      Set_ArrayPtr = (BTNptr *) calloc(Set_ArraySz ,sizeof(BTNptr));
      if (! Set_ArrayPtr) xxx_exit("Out of memory in new_trie/1");
      for (i = 0; i < num_sets; i++)
	Set_ArrayPtr[i] = temp_arrayptr[i];
      free(temp_arrayptr);
    }
    i = num_sets;
    num_sets++;
  }
  return i;
}

/*----------------------------------------------------------------------*/

bool trie_intern(Cell term, int RootIndex, Cell ret_skeleton)
{
  int flag;
  BTNptr Leaf;

  switch_to_trie_assert;
  Leaf = whole_term_chk_ins(term,&(Set_ArrayPtr[RootIndex]),&flag);
  switch_from_trie_assert;
  
  ctop_addr(3,Leaf);
  ctop_int(4,flag);
  /* imitates get_lastnode_cs_retskel() */
  return unify(ret_skeleton, build_ret_term(global_num_vars+1,(void *)var_regs));
}

/*----------------------------------------------------------------------*/

bool trie_interned(Cell trie_term, int rootidx, Cell leaf_term, Cell ret_skeleton)
{
  /*
   * Only if Set_ArrayPtr[RootIndex] is a valid BTNptr can we run this
   * builtin.  That means Set_ArrayPtr[RootIndex] can neither be NULL,
   * nor a deleted set (deleted by builtin delete_trie/1).
   */
  if ((Set_ArrayPtr[rootidx] != NULL) &&
      (! ((long)Set_ArrayPtr[rootidx] & 0x3))) {
    deref(leaf_term);
    if (isref(leaf_term)) {  
      deref(trie_term);
      reg_arrayptr = reg_array-1;
      num_vars_in_var_regs = -1;
      pushreg(trie_term);
      /* imitates get_lastnode_cs_retskel() */
      ctop_addr(3, Last_Nod_Sav);  /* bind leaf_term */
      if (unify(ret_skeleton, build_ret_term(global_num_vars+1,(void *)var_regs))) {
	pcreg = (byte *)Set_ArrayPtr[rootidx];
	return TRUE;
      }
    } else {
    if (isinteger(leaf_term)) { /* encoded address of a trie leaf node */
      BTNptr leaf = (BTNptr)int_val(leaf_term);

      /* fprintf(stderr, "In %s/4: integer leaf found...\n", __FUNCTION__); */
      if (IsDeletedNode(leaf)) return FALSE;
      return bottom_up_unify(leaf, trie_term, ret_skeleton);
    } else
      xxx_err_handle(TYPE_ERROR, 3, __FUNCTION__, 4,
		     "variable or integer", leaf_term);
    }
  }
  return FALSE;
}

/*----------------------------------------------------------------------*/

/*
 * This is builtin #162: TRIE_DISPOSE(+ROOT, +LEAF), to dispose a branch
 * of the trie rooted at Set_ArrayPtr[ROOT].
 */

void trie_dispose(int rootidx, BTNptr leaf)
{
  switch_to_trie_assert;
  delete_branch(leaf, &(Set_ArrayPtr[rootidx]));
  switch_from_trie_assert;
}

/*----------------------------------------------------------------------*/

#define DELETED_SET 1

void delete_interned_trie(int rootidx)
{
  /*
   * We can only delete a valid BTNptr, so that only those sets
   * that were used before can be put into the free set list.
   */
  if ((Set_ArrayPtr[rootidx] != NULL) &&
      (! ((long)Set_ArrayPtr[rootidx] & 0x3))) {
    switch_to_trie_assert;
    delete_trie(Set_ArrayPtr[rootidx]);
    switch_from_trie_assert;
    /*
     * Save the value of first_free_set into Set_ArrayPtr[tmpval].
     * Some simple encoding is needed, because in trie_interned/4 we
     * have to know this set is already deleted.
     */
    Set_ArrayPtr[rootidx] = (BTNptr) ((long)first_free_set << 2 | DELETED_SET);
    first_free_set = rootidx;
  }
}

/*----------------------------------------------------------------------*/
