/* File:      tst_utils.c
** Author(s): Ernie Johnson
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: tst_utils.c,v 1.7 2003/04/07 15:32:09 kostis Exp $
** 
*/


#include <stdio.h>
#include <stdlib.h>

#include "xxx_config.h"

#include "auxlry.h"
#include "cell_xsb.h"
#include "psc_xsb.h"
#include "deref.h"
#include "tst_aux.h"
#include "macro_xsb.h"
#include "register.h"
#include "memory_xsb.h"
#include "binding.h"
#include "choice.h"
#include "tst_utils.h"
#include "error_xsb.h"

/* ====================================================================== */

/*
 *            Stacks needed by TST-Subsumptive Operations
 *            ===========================================
 */


struct tstTermStack      tstTermStack;
struct tstTrail          tstTrail;


void tstInitDataStructs(void)
{
  TermStack_Init;
  Trail_Init;
}

/* ======================================================================= */

#ifdef DEBUG

#include "debug.h"
#include "inst_xsb.h"  /* for inst_name() */

#ifdef BITS64
#define IntegerFormatString	"%ld"
#else
#define IntegerFormatString	"%d"
#endif

/* ======================================================================= */

/*
 *		    D E B U G G I N G   R O U T I N E S
 *	            ===================================
 */

void printNodeType(byte fieldNodeType)
{
  switch (fieldNodeType) {
  case TRIE_ROOT_NT:
    printf("TRIE_ROOT_NT");
    break;
  case HASH_HEADER_NT:
    printf("HASH_HEADER_NT");
    break;
  case LEAF_NT:
    printf("LEAF_NT");
    break;
  case HASHED_LEAF_NT:
    printf("HASHED_LEAF_NT");
    break;
  case INTERRIOR_NT:
    printf("INTERRIOR_NT");
    break;
  case HASHED_INTERRIOR_NT:
    printf("HASHED_INTERRIOR_NT");
    break;
  default:
    printf("Bad NodeType (%c)", fieldNodeType);
    break;
  }
}


void printTrieType(byte fieldTrieType)
{
  switch (fieldTrieType) {
  case CALL_TRIE_TT:
    printf("CALL_TRIE_TT");
    break;
  case ANSWER_TRIE_TT:
    printf("ANSWER_TRIE_TT");
    break;
  case DELAY_TRIE_TT:
    printf("DELAY_TRIE_TT");
    break;
  case ASSERT_TRIE_TT:
    printf("ASSERT_TRIE_TT");
    break;
  case INTERN_TRIE_TT:
    printf("INTERN_TRIE_TT");
    break;
  default:
    printf("Bad TrieType (%c)", fieldTrieType);
    break;
  }
}


void printTrieSymbol(Cell symbol)
{
  if ( symbol == ESCAPE_NODE_SYMBOL )
    printf("%lu [ESCAPE_NODE_SYMBOL]", ESCAPE_NODE_SYMBOL);
  else {
    switch(TrieSymbolType(symbol)) {
    case INT:
      printf(IntegerFormatString, int_val(symbol));
      break;
    case ATM:
      printf("%s", atom_val(symbol));
      break;
    case FLOAT:
      printf("%f", float_val(symbol));
      break;
    case TrieVar:
      printf("V" IntegerFormatString, DecodeTrieVar(symbol));
      break;
    case STRUCT:
      {
	Psc psc = DecodeTrieFunctor(symbol);
	printf("%s/%d", get_name(psc), get_arity(psc));
      }
      break;
    case LIST:
      printf("LIST");
      break;
    default:
      printf("Unknown symbol (tag = %ld)", cell_tag(symbol));
      break;
    }
  }
}

/* ========== CURRENTLY NOWHERE USED ===========
void printTrieNode(BTNptr pTN)
{
  printf("Trie Node: Addr(0x%p)", (void*)pTN);
  if ( IsDeletedNode(pTN) )
    printf("  (DELETED)");
  printf("\n\tInstr(%s), NodeType(", inst_name(TN_Instr(pTN)));
  printNodeType(TN_NodeType(pTN));
  printf(")\n\tTrieType(");
  printTrieType(TN_TrieType(pTN));
  printf("), Symbol(");
  printTrieSymbol(TN_Symbol(pTN));
  printf(")");
  printf("\n\tParent(0x%p), Child(0x%p), Sibling(0x%p)\n",
	 (void*)TN_Parent(pTN), (void*)TN_Child(pTN), (void*)TN_Sibling(pTN));
}
 */

/* printterm() writes to stdout, therefore, so should this. */
/* ========== CURRENTLY NOWHERE USED ===========
void printTabledCall(TabledCallInfo callInfo)
{
  int arity, i;
  Psc pPSC;
  
  pPSC = TIF_PSC(CallInfo_TableInfo(callInfo));
  printf( "%s(", get_name(pPSC) );
  arity = CallInfo_CallArity(callInfo);
  for (i = 1; i <= arity; i++) {
    printterm( (Cell)(CallInfo_Arguments(callInfo)+i), 1, 8 );
    if (i < arity)
      printf( "," );
  }
  printf( ")" );
}
 */
/*-----------------------------------------------------------------------*/


/* Printing the Answer Template
 * ----------------------------
 *  The answer template is assumed to be stored in high-to-low memory
 *  vector fashion, with the incoming template pointer pointing to the
 *  first term Cell (highest memory Cell of the vector).
 */

void printAnswerTemplate(CPtr pAnsTmplt, int size)
{
  int i;

  printf("Answer Template:  (");
  if (size > 0) {
    for (i = 1; i < size; i++) {
      printterm(*pAnsTmplt--,1,10);
      printf(",");
    }
    printterm(*pAnsTmplt,1,10);
  }
  printf(")\n");
}

#endif /* DEBUG */
