/* File:      eval.P
** Author(s): David S. Warren, Kostis F. Sagonas
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1993-1998
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: eval.P,v 1.5 2003/11/23 23:00:52 kostis Exp $
** 
*/

:- compiler_options([xpp_on]).

#include "celltags_xsb.h"

%-------------------------------------------------------------------------
% NOTE:	Predicate eval/2 is the one who is responsible for error checking
%	of arithmetic comparisons and evaluations. According to the Prolog
%	standard, arithmetic comparisons and evaluations can cause
%	instantiation errors ONLY. Everything else should simply fail as
%	far as evaluation is concerned.
%-------------------------------------------------------------------------

eval(Arg, Res) :- 
	term_type(Arg, T), 
	( (T =:= FREE; T =:= ATTV) ->
	    abort('Unbound variable in arithmetic expression')
	; T =:= INT -> Res=Arg
	; T =:= FLOAT -> Res=Arg
	; eval0(Arg, Res)	/* system defined function */
	).

eval0('+'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 + Res2.
eval0('-'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 - Res2.
eval0('*'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 * Res2.
eval0('/'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 / Res2.
eval0('//'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 // Res2.
eval0(rem(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 rem Res2.
eval0(mod(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 mod Res2.
eval0('/\'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 /\ Res2.
eval0('\/'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 \/ Res2.
eval0('<<'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 << Res2.
eval0('>>'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 >> Res2.

eval0('-'(X), Res) :- eval(X, Res1), Res is -(Res1).
eval0('\'(X), Res) :- eval(X, Res1), Res is '\'(Res1).
eval0(float(X), Res) :- eval(X, Res1), Res is float(Res1).
eval0(floor(X), Res) :- eval(X, Res1), Res is floor(Res1).
eval0(sin(X), Res) :- eval(X, Res1), Res is sin(Res1).
eval0(cos(X), Res) :- eval(X, Res1), Res is cos(Res1).
eval0(tan(X), Res) :- eval(X, Res1), Res is tan(Res1).
eval0(asin(X), Res) :- eval(X, Res1), Res is asin(Res1).
eval0(acos(X), Res) :- eval(X, Res1), Res is acos(Res1).
eval0(atan(X), Res) :- eval(X, Res1), Res is atan(Res1).
eval0(exp(X), Res) :- eval(X, Res1), Res is exp(Res1).
eval0(log(X), Res) :- eval(X, Res1), Res is log(Res1).
eval0(log10(X), Res) :- eval(X, Res1), Res is log10(Res1).
eval0(sqrt(X), Res) :- eval(X, Res1), Res is sqrt(Res1).
eval0(abs(X), Res) :- eval(X, Res1), Res is abs(Res1).
eval0(truncate(X), Res) :- eval(X, Res1), Res is truncate(Res1).
eval0(round(X), Res) :- eval(X, Res1), Res is round(Res1).
eval0(ceiling(X), Res) :- eval(X, Res1), Res is ceiling(Res1).

/* ---------------------- end of file eval.P --------------------------	*/
