/* File:      file_io.P
** Author(s): kifer
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1999
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: file_io.P,v 1.1.1.1 2001/06/13 16:41:16 kostis Exp $
** 
*/


:- compiler_options([xpp_on]).
#include "io_defs_xsb.h"
#include "standard.h"
#include "flag_defs_xsb.h"


	       /********************************/
	       /*  Formatted I/O    	       */
	       /********************************/

fmt_write(Fmt, Args) :-
       stat_flag(CURRENT_OUTPUT, IOport),   % Current output
	fmt_write(IOport, Fmt, Args).

fmt_write(IOport, Fmt, Args) :-
	formatted_io(FMT_WRITE, IOport, Fmt, Args, _Ret).

fmt_write_string(Str, Fmt, Args) :-
	formatted_io(FMT_WRITE_STRING, Str, Fmt, Args, _Ret).


fmt_read(Fmt,Result,Status) :-
        stat_flag(CURRENT_INPUT, IOport),  % Current input
	fmt_read(IOport, Fmt,Result,Status).

fmt_read(IOport, Fmt, Result, Status) :-
	formatted_io(FMT_READ, IOport, Fmt, Result, Status).


/***************************************************************/
/*    	      	      	      	      	      	      	       */
/*       file_flush, file_seek, file_truncate, file_pos, etc.  */
/*    	      	      	      	      	      	      	       */
/***************************************************************/

file_flush(IOport, Ret) :-
	file_function(FILE_FLUSH, IOport, Ret, _D, _E, _,_).
file_seek(IOport, Offset, Place, Ret) :-
	file_function(FILE_SEEK, IOport, Offset, Place, Ret, _,_).
file_truncate(IOport, Length, Ret) :-
	file_function(FILE_TRUNCATE, IOport, Length, Ret, _E, _,_).
file_pos(IOport, Pos) :-
	file_function(FILE_POS, IOport, Pos, _Ret, _E, _,_).
file_open(FileName, Mode, IOport) :-
	file_function(FILE_OPEN, FileName, Mode, IOport, _E, _,_).
file_close(IOport) :-
	file_function(FILE_CLOSE,IOport,_X,_Y,_Z, _,_).
file_get(IOport,Char) :-
	file_function(FILE_GET, IOport, Char, _X, _Y, _,_).
file_put(IOport,Char) :-
	file_function(FILE_PUT, IOport, Char, _X, _Y, _,_).
file_clearerr(IOport) :-
	file_function(FILE_CLEARERR, IOport, _, _, _, _, _).
tmpfile_open(IOport) :-
	file_function(TMPFILE_OPEN, IOport, _, _, _, _, _).

%% These are lower level than file_read_line_atom/list
%% Therefore, they don''t check their arguments.
%% file_getbuf doesn''t intern ToString.
file_getbuf(IOport, ByteCount, ToString, BytesRead) :-
	file_function(FILE_GETBUF, IOport,ByteCount,ToString,BytesRead,_,_).

file_getbuf_atom(IOport, ByteCount, ToAtom, BytesRead) :-
	(var(ToAtom), var(BytesRead) ->
	    file_function(FILE_GETBUF, IOport,ByteCount,TmpStr,BytesRead,_,_),
	    intern_string(TmpStr, ToAtom)
	;
	 abort('Usage: file_getbuf_atom([+IOport,] +ByteCount, -ToAtom, -BytesRead)')
        ).

file_getbuf_atom(ByteCount, ToAtom, BytesRead) :-
	stat_flag(CURRENT_INPUT, IOport),
	(var(ToAtom), var(BytesRead) ->
	    file_function(FILE_GETBUF, IOport,ByteCount,TmpStr,BytesRead,_,_),
	    intern_string(TmpStr, ToAtom)
	;
	 abort('Usage: file_getbuf_atom([+IOport,] +ByteCount, -ToAtom, -BytesRead)')
        ).

file_getbuf_list(IOport, ByteCount, ToList, BytesRead) :-
	(var(ToList), var(BytesRead) ->
	    file_function(FILE_GETBUF, IOport,ByteCount,TmpStr,BytesRead,_,_),
	    atom_codes(TmpStr, ToList)
	;
	abort('Usage: file_getbuf_list([+IOport,] +ByteCount, -ToList, -BytesRead)')).

file_getbuf_list(ByteCount, ToList, BytesRead) :-
	stat_flag(CURRENT_INPUT, IOport),
	(var(ToList), var(BytesRead) ->
	    file_function(FILE_GETBUF, IOport,ByteCount,TmpStr,BytesRead,_,_),
	    atom_codes(TmpStr, ToList)
	;
	abort('Usage: file_getbuf_list([+IOport,] +ByteCount, -ToList, -BytesRead)')).

file_putbuf(IOport, ByteCount, FromString, Offset, BytesWritten) :-
	(var(BytesWritten) ->
	    file_function(FILE_PUTBUF, IOport,ByteCount,FromString,Offset,BytesWritten,_)
	;
	abort('Usage: file_putbuf([+IOport,] +ByteCount, +FromString, +Offset, -BytesWritten)')
    ).

file_putbuf(ByteCount, FromString, Offset, BytesWritten) :-
	stat_flag(CURRENT_OUTPUT, IOport),
	(var(BytesWritten) ->
	    file_function(FILE_PUTBUF, IOport,ByteCount,FromString,Offset,BytesWritten,_)
	;
	abort('Usage: file_putbuf([+IOport,] +ByteCount, +FromString, +Offset, -BytesWritten)')
	).

%% This doesn''t intern Str
file_read_line(IOport, Str, IsFullLine) :-
	( var(Str), var(IsFullLine) ->
	    file_function(FILE_READ_LINE, IOport, Str, IsFullLine, _, _,_)
	    ;
	    abort('Usage: file_read_line([+IOport,] -InputStr, -IsFullLine)')
        ).

%% Like file_read_line, but Str becomes an interned atom
file_read_line_atom(IOport, Str, IsFullLine) :-
	( var(Str), var(IsFullLine) ->
	    file_function(FILE_READ_LINE, IOport, UnStr, IsFullLine, _, _,_),
	    intern_string(UnStr,Str)
	    ;
	    abort('Usage: file_read_line_atom([+IOport,] -InputStr, -IsFullLine)')
        ).

%% Read line from current input stream (previously set by 'see')
%% Intern string
file_read_line_atom(Str, IsFullLine) :-
	stat_flag(CURRENT_INPUT, IOport),  % Current input
	file_read_line_atom(IOport, Str, IsFullLine).

%% Like file_read_line_atom, but returns list of characters
file_read_line_list(IOport, CharLst, IsFullLine) :-
	( var(Str), var(IsFullLine) ->
	    file_function(FILE_READ_LINE, IOport, Str, IsFullLine, _, _,_),
	    atom_codes(Str,CharLst)
	    ;
	    abort('Usage: file_read_line_list([+IOport,] -InputStr, -IsFullLine)')
        ).

%% Read line from current input stream (previously set by 'see')
%% Return list of chars
file_read_line_list(Lst, IsFullLine) :-
	stat_flag(CURRENT_INPUT, IOport),  % Current input
	file_read_line_list(IOport, Lst, IsFullLine).

%% Write Str out to IOport. Str must be atom or list of chars.
%% This is checked in C, by the builtin.
file_write_line(IOport, Str, Offset) :-
	file_function(FILE_WRITE_LINE, IOport, Str, Offset, _,_,_).

%% Same as file_write_line/3, but write to current output
file_write_line(Str, Offset) :-
	stat_flag(CURRENT_OUTPUT, IOport),  % Current output
	file_function(FILE_WRITE_LINE, IOport, Str, Offset, _,_,_).


/* Fancy stuff */
%% Like freopen
file_reopen(FileName, Mode, IOport,ErrCode) :-
	file_function(FILE_REOPEN, FileName,  Mode, IOport, ErrCode,_,_).

%% Like dup/dup2: make DestIOport into the same file descriptor as SrcIOport
%% If DestIOport is var, then behaves like dup; else like dup2 
%% (and then expects DestIOport to be a file descriptor).
%% ErrCode=0 - ok
file_clone(SrcIOport, DestIOport, ErrorCode) :-
	file_function(FILE_CLONE, SrcIOport, DestIOport, ErrorCode,_,_,_).

%% Open pipe. ReadPipe is for reading from pipe,
%% WritePipe is for writing to the pipe.
pipe_open(ReadPipe, WritePipe) :-
	file_function(PIPE_OPEN, ReadPipe, WritePipe, _,_,_,_).

%% Convert pipe into an XSB file descriptor.
%% This should be done on both ends: by the parent XSB process and by 
%% the child XSB process
fd2ioport(Pipe, IOport) :-
	file_function(FD2IOPORT, Pipe, IOport,_,_,_,_).


/***************************************************************/
/*    	      	      	      	      	      	      	       */
/*       file_stat, file_size 	      	   	     	       */
/*    	      	      	      	      	      	      	       */
/***************************************************************/

file_time(FileName, Time) :- file_stat(FileName,FILE_STAT_TIME,Time).
file_size(FileName, Size) :- file_stat(FileName,FILE_STAT_SIZE,Size).

/***************************************************************
**
** print_file_contents/1
**
****************************************************************/

print_file_contents(Filename) :-
	stat_flag(CURRENT_OUTPUT, ToIOport),  % current output
	print_file_contents(Filename, ToIOport).

print_file_contents(FileName, ToIOport) :-
	file_open(FileName, r, FromIOport),
	copyIOport(FromIOport, ToIOport),
	file_close(FromIOport).
copyIOport(FromIOport, ToIOport) :-
	(file_read_line(FromIOport, Line, _) -> file_write(ToIOport, Line)),
	copyIOport(FromIOport, ToIOport),
	fail.        	 %% so as to not leave choice points around
copyIOport(_,_).


